/*===========================*/
/*    1-D FFT   routines     */
/*===========================*/

/*-Title: 1-D FFT ROUTINES
-end*/

/*-Intro:
These are a small set of 'C' based routines for calculating complex or real
1-dimensional Fast Fourier Transforms (forward or inverse). They are only
for transforms where the number of points is a power of two. The code is
based around that described in the book 'Numerical Recipes in Fortran'.
In the current implementation the weighting factors are calculated only
once (or when the number of points or direction of the FFT is changed) 
providing increased performance when a series of 1-dimensional FFTs are
to be computed for the same no. of points and same direction. Fortran
interfaces are provided for the user callable routines.
-end*/

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include <stdio.h>
#include <math.h>

#include "fft_1d_systyp.h"
#include "fft_1d.h"

/* Define Fortran binding names */

#if    LINKTYP == 1

#define fftf_1d_cmplx fftf_1d_cmplx_
#define fftf_1d_real fftf_1d_real_

#endif

#if  LINKTYP == 2

#define fftf_1d_cmplx FFTF_1D_CMPLX
#define fftf_1d_real FFTF_1D_REAL

#endif

void fftf_1d_cmplx();
void fftf_1d_real();


/* Declarations for service routines */
int fft_1d_set_wtab (int npow, int isgn);
int fft_1d_set_wtab2 (int npow, int isgn);




/*-Section: Calculate Complex and Real FFT's
Routines are available to calculate 1-dimensional forward and reverse
Fourier transforms for complex and real data.
-end*/

int npow_tab=0;
int isgn_tab=1;
float *w_real;
float *w_imag;

int npow_tab2=0;
int isgn_tab2=1;
float *w_real_tab2;
float *w_imag_tab2;

int npow_tab3=0;
int isgn_tab3=1;
float *w_real_tab3;
float *w_imag_tab3;

/*-Routine: Complex 1-D FFT  - fft_1d_cmplx
The routine fft_1d_cmplx (fftf_1d_cmplx) is used to calculate a forward or 
reverse FFT for an array of complex data (stored in a real array). The number
of points must be a power of two. The reverse transform points are scaled
by npow**2 (the number of complex points to be transformed).
-end*/

/*                                            
************************
**    fft_1d_cmplx    **
************************

Purpose: 1-D complex FFT (power of two)

Author: John W. Campbell, January 1997

CD-Fortran:
         CALL FFTF_1D_CMPLX  (X, NPOW, ISIGN, IERR)
CD-end:
*/

/*-Parameters:
X()          r  (R/W)  Array of data to be transformed (2*NPOW**2
                       points). The real and imaginary components
                       of the data are stored in adjacant elements
                       in the real array.
NPOW         i  (R)    The power of two. There are 2**NPOW points in the 
                       transform
ISIGN        i  (R)    Flag =1  forward transform
                            =-1 reverse transform
IERR         i  (W)    Error flag =0 OK
                                  =1 Error in parameters
                                  =2 Cannot allocate required memory
-end*/

/*-C:*/
int fft_1d_cmplx (float *x, int npow, int isgn)
/*end*/

/*-Parameters:
float *x   Array of data to be transformed (2*npow**2
           points). The real and imaginary components
           of the data are stored in adjacant elements
           in the real array. (M)
int npow   The power of two. There are 2**npow points in the 
           transform. (R)
int isgn   Flag =1  forward transform, =-1 reverse transform
   
Return:  Error flag =0, OK
                    =1, Error in parameters
                    =2, Cannot allocate required memory
-end*/

{
   int n;
   int m;
   int nm;
   int i;
   int j;
   int ipow;
   int indx;
   int err;
   float tempr;
   float tempi;

   if (npow<=0) return 1;
   if (isgn!=1&&isgn!=-1) return 1;

/* Set up weight factor tables if not already done */

   if (npow!=npow_tab||isgn!=isgn_tab) 
   {
      err = fft_1d_set_wtab (npow, isgn);
      if (err!=0) return 2;
   }

/* Bit reversal */

   n = 1<<npow;
   j = 0;
   for (i=0;i<2*n;i+=2)
   {
      if (j>i)
      {
         tempr = x[j];
         tempi = x[j+1];
         x[j] = x[i];
         x[j+1] = x[i+1];
         x[i] = tempr;
         x[i+1] = tempi;
      }    
      m = n;
      while (m>=2&&j>=m)
      {
        j -=m;
        m = m/2;
      }
      j+=m;
   }

/* Perform the transform */

   nm = 2;
   indx = 0;
   for (ipow=0;ipow<npow;++ipow)
   {
      for (m=0;m<nm;m+=2)
      {
         for (i=m;i<2*n;i+=2*nm)
         {
            j = i + nm;
            tempr = w_real[indx]*x[j] - w_imag[indx]*x[j+1];
            tempi = w_real[indx]*x[j+1] + w_imag[indx]*x[j];
            x[j] = x[i] - tempr;
            x[j+1] = x[i+1] - tempi;
            x[i] += tempr;
            x[i+1] += tempi;
         }
         indx++;
      }
      nm *=2;
   }
   return 0;
}

/* Fortran binding: fftf_1d_cmplx */

void  fftf_1d_cmplx (float * x, int * npow, int *isign, int *ierr)
{
   *ierr = fft_1d_cmplx (x, *npow, *isign);
   return;
}

/*-Routine: Real 1-D FFT  - fft_1d_real
The routine fft_1d_real (fftf_1d_real) is used to calculate a forward or 
reverse real FFT for an array of real data. The number of points must be a 
power of two. The reverse transform points are scaled by npow**2 (the number 
of points to be transformed). The result is the complex transform with
the real and imaginary components in adjacant locations.
-end*/

/*                                            
************************
**    fft_1d_real     **
************************

Purpose: 1-D real FFT (power of two)

Author: John W. Campbell, January 1997

CD-Fortran:
         CALL FFTF_1D_REAL  (X, NPOW, ISIGN, IERR)
CD-end:
*/

/*-Parameters:
X()          r  (R/W)  Array of real data to be transformed (2*NPOW**2 
                       words in length). On input values are stored in the
                       first 2**NPOW elements. On output the complex
                       data are stored in 2**NPOW/2 or 2**NPOW  pairs of 
                       data values with the real and imaginary components 
                       in adjacant locations in the real array.
NPOW         i  (R)    The power of two. There are 2**NPOW points in the 
                       transform. NPOW must be > 1.
ISIGN        i  (R)    Flag =1  forward transform -> half of complex 
                                transform 2**NPOW/2 real/imag pairs
                            =2  forward transform -> full complex
                                transform 2**NPOW real/imag pairs
                            =-1 reverse transform (from half complex
                                transform)
                            =-2 reverse transform (from full complex
                                transform)
IERR         i  (W)    Error flag =0 OK
                                  =1 Error in parameters
                                  =2 Cannot allocate required memory
-end*/

/*-C:*/
int fft_1d_real (float *x, int npow, int isgn)
/*end*/

/*-Parameters:
float *x   Array of real data to be transformed (2*npow**2
           words in length). On input values are stored in the
           first 2**npow elements. On output the complex
           data are stored in 2**npow pairs of data values
           with the real and imaginary components in
           adjacant locations in the real array. (M)
int npow   The power of two. There are 2**npow points in the 
           transform. npow must be > 1. (R)
int isgn   Flag =1  forward transform -> half of complex 
                    transform 2**NPOW/2 real/imag pairs
                =2  forward transform -> full complex
                    transform 2**NPOW real/imag pairs
                =-1 reverse transform (from half complex
                    transform)
                =-2 reverse transform (from full complex
                    transform) (R)
   
Return:  Error flag =0, OK
                    =1, Error in parameters
                    =2, Cannot allocate required memory
-end*/

{
   int n;
   int n2p1;
   int i;
   int i1;
   int i2;
   int i3;
   int i4;
   int indx;
   int err;
   int ksgn;
   float c1;
   float c2;
   float h1_real;
   float h1_imag;
   float h2_real;
   float h2_imag;

   if (npow<=1) return 1;
   if (isgn>2||isgn<-2||isgn==0) return 1;
   ksgn = 1;
   if (isgn<0) ksgn = -1;

/* Set up weight factor tables if not already done */

   if (npow!=npow_tab2||ksgn!=isgn_tab2) 
   {
      err = fft_1d_set_wtab2 (npow, ksgn);
      if (err!=0) return 2;
   }
   c1 = 0.5;
   n = 1<<(npow-1);
   n2p1 = 2*n + 1;
   if (isgn>0)
   {
      c2 = -0.5;
      err = fft_1d_cmplx (x, npow-1, ksgn);
      if (err!=0) return err;  
   }
   else
   {
      c2 = 0.5;
      if (isgn==-2)
      {
         x[1] = x[2*n];
      }
   }

   indx = 0;;
   for (i=2;i<=n/2;++i)
   {
     i1 = 2*i - 2;
     i2 = i1 + 1;
     i3 = n2p1 - i2;
     i4 = i3 + 1;
     h1_real = c1*(x[i1]+x[i3]);
     h1_imag = c1*(x[i2]-x[i4]);
     h2_real = -c2*(x[i2]+x[i4]);
     h2_imag = c2*(x[i1]-x[i3]);
     x[i1]= h1_real + w_real_tab2[indx]*h2_real - w_imag_tab2[indx]*h2_imag;
     x[i2]= h1_imag + w_real_tab2[indx]*h2_imag + w_imag_tab2[indx]*h2_real;
     x[i3]= h1_real - w_real_tab2[indx]*h2_real + w_imag_tab2[indx]*h2_imag;
     x[i4]= -h1_imag + w_real_tab2[indx]*h2_imag + w_imag_tab2[indx]*h2_real;
     indx++;
   }

   if (isgn>0)
   {
      h1_real = x[0];
      x[0] = h1_real + x[1];
      x[1] = h1_real - x[1];
   }
   else
   {
      h1_real = x[0];
      x[0] = c1*(h1_real + x[1]);
      x[1] = c1*(h1_real - x[1]); 
      err = fft_1d_cmplx (x, npow-1, ksgn);
      if (err!=0) return err;
   }

   if (isgn==2)
   {
      for (i=0;i<n;++i)
      {
         i1 = 2*n + 2*i;
         i2 = 2*n - 2*i;
         x[i1] = x[i2];
         x[i1+1] = -x[i2+1];
      }
      x[2*n] = x[1];
      x[2*n+1] = 0.0;
      x[1] = 0.0;
   }
     
   return 0;
}

/* Fortran binding: fftf_1d_real */

void  fftf_1d_real (float * x, int * npow, int *isign, int *ierr)
{
   *ierr = fft_1d_real (x, *npow, *isign);
   return;
}


/*                                            
****************************
**    fft_1d_set_wtab     **
****************************

Purpose: Set weight table for fft_1d_cmplx routine

Author:  John W. Campbell, January 1997

**** SERVICE ROUTINE ONLY ****

*/

int fft_1d_set_wtab (int npow, int isgn)
{
   int n;
   int m;
   int nm;
   int ipow;
   int indx;
   double wc_real;
   double wc_imag;
   double wm1_real;
   double wm1_imag;
   double w_temp;
   double theta;
   double sht;
   double twopi = 6.2831853;

   if (npow==npow_tab&&isgn==isgn_tab) return 0;

   if (npow!=npow_tab)
   {      
      if (npow_tab!=0) {free(w_real); free(w_imag);} 
      n = 1<<npow;
      w_real = (float *) malloc (n*sizeof(float));
      if (w_real==0) return 1;
      w_imag = (float *) malloc (n*sizeof(float));
      if (w_imag==0) {free(w_real); return 1;}
   }
   npow_tab = npow;
   isgn_tab = isgn;
   nm = 2;
   indx = 0;
   for (ipow=0;ipow<npow;++ipow)
   {
      theta = twopi/(nm*isgn_tab);
      sht = sin (0.5*theta);
      wm1_real = -2.0*sht*sht;
      wm1_imag = sin(theta);
      wc_real = 1.0;
      wc_imag = 0.0;
      for (m=0;m<nm;m+=2)
      {
         w_real[indx] = wc_real;
         w_imag[indx] = wc_imag;
         w_temp = wc_real;
         wc_real += wc_real*wm1_real - wc_imag*wm1_imag; 
         wc_imag += wc_imag*wm1_real + w_temp*wm1_imag; 
         indx++;
      }
      nm *= 2;
   }
   return 0;
}

/*                                            
****************************
**    fft_1d_set_wtab2    **
****************************

Purpose: Set weight table for fft_1d_real routine

Author:  John W. Campbell, January 1997

**** SERVICE ROUTINE ONLY ****

*/

int fft_1d_set_wtab2 (int npow, int isgn)
{
   int n;
   int i;
   int indx;
   double wc_real;
   double wc_imag;
   double wm1_real;
   double wm1_imag;
   double w_temp;
   double theta;
   double sht;
   double pi = 3.1415927;

   if (npow==npow_tab2&&isgn==isgn_tab2) return 0;

   n = 1<<(npow-1);
   if (npow!=npow_tab2)
   {      
      if (npow_tab2!=0) {free(w_real_tab2); free(w_imag_tab2);} 
      w_real_tab2 = (float *) malloc ((n/2)*sizeof(float));
      if (w_real_tab2==0) return 1;
      w_imag_tab2 = (float *) malloc ((n/2)*sizeof(float));
      if (w_imag_tab2==0) {free(w_real_tab2); return 1;}
   }
   npow_tab2 = npow;
   isgn_tab2 = isgn;

   indx = 0;
   theta = pi/n;
   if (isgn<0) theta = -theta;
   sht = sin (0.5*theta);
   wm1_real = -2.0*sht*sht;
   wm1_imag = sin(theta);
   wc_real = 1.0 + wm1_real;
   wc_imag = wm1_imag;
   for (i=2;i<=n/2;++i)
   {
      w_real_tab2[indx] = wc_real;
      w_imag_tab2[indx] = wc_imag;
      w_temp = wc_real;
      wc_real += wc_real*wm1_real - wc_imag*wm1_imag; 
      wc_imag += wc_imag*wm1_real + w_temp*wm1_imag; 
      indx++;
   }
   return 0;
}
