/*======================================================*/
/* Some PX X-Ray Diffraction Image Handling Routines    */
/*======================================================*/

/*-Title: IMAGE HANDLING ROUTINES - IMGLIB
-end*/

/*-Intro:
This file contains a series of routines concerned with the handling
of single crystal X-ray diffraction images and, in particular
Protein Crystallographic rotation images. The routines are written
in 'C' with Fortran interfaces provided. They are designed to handle
image data in any of the eight possible orders with various formats
allowed for the stored pixels data.

-end*/


#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include <stdio.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <math.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "imglib_systyp.h"
#include "imglib.h"
#include "dml.h"

/* Define Fortran binding names */

#if    LINKTYP == 1

#define imgf_readfile imgf_readfile_
#define imgf_getrect imgf_getrect_
#define imgf_ax1line imgf_ax1line_
#define imgf_ax2line imgf_ax2line_
#define imgf_getpix imgf_getpix_
#define imgf_filename imgf_filename_
#define imgf_axord imgf_axord_
#define imgf_bgcalc imgf_bgcalc_
#define imgf_findspots imgf_findspots_
#define imgf_spot_details imgf_spot_details_
#define imgstr imgstr_

#endif

#if  LINKTYP == 2

#define imgf_readfile IMGF_READFILE
#define imgf_getrect IMGF_GETRECT
#define imgf_ax1line IMGF_AX1LINE
#define imgf_ax2line IMGF_AX2LINE
#define imgf_getpix IMGF_GETPIX
#define imgf_filename IMGF_FILENAME
#define imgf_axord IMGF_AXORD
#define imgf_bgcalc IMGF_BGCALC
#define imgf_findspots IMGF_FINDSPOTS
#define imgf_spot_details IMGF_SPOT_DETAILS
#define imgstr IMGSTR

#endif

void imgf_readfile();
void imgf_getrect();
void imgf_ax1line();
void imgf_ax2line();
void imgf_getpix();
void imgf_filename();
void imgf_axord();
void imgf_bgcalc();
void imgf_findspots();
void imgf_spot_details();
int imgstr();


/* Define Integer & Integer*2 data types */

typedef int Intpix;              /* 4 byte signed integer */
typedef short I2pix;             /* 2 byte signed integer */
typedef unsigned short I2uns;    /* 2 byte unsigned integer */
typedef short I2squash;          /* 2 byte squashed integer */

/* Define nearest integer macro */

#define nint(x)  x>=0.0?(int)(x+0.5):(int)(x-0.5)
#define min(i1,i2) i1<=i2?i1:i2
#define max(i1,i2) i1>=i2?i1:i2

/* Pointers list for IMGSTR */

#define MAXPOINTERS 20       /* Maximum no. of pointers for IMGSTR transfers */

char * IMG_pointers[MAXPOINTERS];          /* List of string pointers */
int    IMG_pointers_index = MAXPOINTERS;   /* Last index to pointers currently
                                              used (list used in a circular 
                                              manner) */
/* Structures for finding spots on an image */

typedef struct{
   float sumxi;   /* Sum of intensity*(current x c_of_g) in rasters*/
   float sumyi;   /* Sum of intensity*(current y c_of_g) in rasters*/
   float sumi;    /* Sum of intensities */
   int ixcent;    /* Current x centre position (rasters) */
   int iycent;    /* Current y centre position (rasters) */
   int npixcent;  /* Current no. of pixels in spot */
   int iybeg;     /* Start y pixel of spot */
   int iyend;     /* End y pixel of spot */
   int ixmin;     /* Minimum 'x' extent of spot */
   int ixmax;     /* Maximum 'x' extent of spot */
   int iymin;     /* Minimum 'y' extent of spot */
   int iymax;     /* Maximum 'y''extent of spot */
   int iflag;     /* Flags spots at limits of image/RMAX etc. */
}Active_Spot;

typedef struct{
   float ax1_pos;      /* Spot position along 1'st local axis in rasters */
   float ax2_pos;      /* Spot position along 2'nd local axis in rasters */
   float sumint;       /* Spot intensity */
   int npix;           /* No. of pixels in spot */
   int siz_ax1;        /* Spot size along 1'st local axis in rasters*/
   int siz_ax2;        /* Spot size along 2'nd local axis in rasters*/
}Found_Spot;

/* Define Parameter Constants */

#define MAX_ACT_START 1000        /* Start storage for active spots in
                                     spot finding routine */
#define MAX_ACT_INC 500           /* Increment to storage for active spots
                                     in spot finding routine */
#define INIT_SPOT_RECS 500        /* Initial allocation for found spot 
                                     records */
#define INC_SPOT_RECS 500         /* Incremental allocation for found spot 
                                     records */

/* Define service routine types */
long img_desq();
long img_desq2();
long img_desq3();
long img_desq4();
int img_unsq1();
int img_unsq2();
int img_unsq3();
int img_unsq4();
int img_unsq5();
int img_byte_log ();
I2pix img_sqsh1();
I2pix img_sqsh2();
I2pix img_sqsh3();
I2pix img_sqsh4();
I2pix img_sqsh5();
void set_mar_ovpix (void *img, int * ovpix, int iord, 
                   int nxrasts, int nyrasts, int out_typ, int * ovfl);
void img_copy_chars (char * string_in, int lin, 
                     char * string_out, int max_len);
int img_blank_string (char * str, int lens);

/*-Section: Reading an image
This contains a general routine to read an Protein Crystallographic
X-ray diffraction images.  A lower level 'C' routine which may be useful 
in reading the image data section of an image file is also provided.
-end*/

/*-Routine: Read an image file - img_readfile
The routine img_readfile (imgf_readfile) is a general purpose routine to
read an image file into memory. The images in question are those used for 
collection protein crystallographic X-ray diffraction data. New specific 
image types will be added as needed. A special option enables image
parameters to be retrieved from an image file (if stored) without
actually reading the image data; This may be useful so that an appropriate
sized array me be allocated in which to return the image data on a 
subsequent call to the routine.
-end*/

/*
****************************
**     img_readfile       **
****************************

Purpose: Read an image data file

Author:  John W. Campbell, August 1997

CD-Fortran:
      SUBROUTINE IMGF_READFILE (IMGSTR(FILNAM), IFTYPE, IOUT_TYP, LHEAD,
                               ITYP, IORDER, IROTN, ISWAP,
                               NAX1_RASTS, NAX2_RASTS, LREC, PROGRESS,
                               IMG_FULL, MINVAL, MAXVAL, IOVFL, MAX_OK,
                               IFAIL_SOFT, IORD_USED, 
                               NAX1_RASTS_USED, NAX2_RASTS_USED, IERR)
CD-end
*/
/*-Parameters:

FILNAM          c (R) The name of the image file to be read
                      ** Pass address using the IMGSTR routine **
IFTYPE          i (R) File type 
                        = 0 'raw' General image. 
                            All input items must be defined.
                        = 1 'byte' (unsigned byte data - no header)
                            Ignores ITYP, ISWAP, LHEAD, LREC
                        = 2 'i2' (unsigned two byte data - no header)
                            Ignores ITYP, LHEAD, LREC
                        = 3 'mar' (MAR 'image' format)
                            Ignores ITYP, IORDER, ISWAP, LHEAD, 
                            NAX1_RASTS, NAX2_RASTS, LREC. 
                            (Soft check on IORDER, NAX1_RASTS, 
                             NAX2_RASTS)
IOUT_TYP        i (R) Format required for returned data
                      1 = unsigned byte data
                      2 = unsigned two-byte data (i2)
                      3 = signed integer data (i4)
                      4 = 'squashed i2' data (if stored I<0
                          ipix = -i*8
                      5 = 'squashed i2' data if stored I<0
                          ipix = (I+32768)*8
                      6 = 'squashed i2' data if stored I<0
                          ipix = (I+32768)*32
                      7 = 'squashed i2' data if stored I<0
                          ipix = -(I+1)*256 + 32768

                      0 = do not return image data (may be used to
                          for example to get nos. of rasters from
                          a MAR file so that the required amount
                          of memory can be allocated for the image
                          prior to another call to IMGF_READFILE
                          with a value of IOUT_TYP>0)

                      Notes: IOUT_TYP==1 may only be used if IFTYPE==1
                             or (IFTYPE==0 and ITYP==1) 

                             If ITYP (see below) is 3 and IOUT_typ is 
                             not 3 then any negative pixel values will 
                             be set to 0.
LHEAD           i (R) No. of header records or -(number of bytes of
                      header info). (For IFTYPE 0 only)
ITYP            i (R) Format of stored (input) pixels data 
                      (for filtyp==0 - otherwise ignored)
                      1 = unsigned byte data
                      2 = unsigned two-byte data (i2)
                      3 = signed integer data (i4)
                      4 = 'squashed i2' data (if stored I<0
                          ipix = -i*8
                      5 = 'squashed i2' data if stored I<0
                          ipix = (I+32768)*8
                      6 = 'squashed i2' data if stored I<0
                          ipix = (I+32768)*32
                      7 = 'squashed i2' data if stored I<0
                          ipix = -(I+1)*256 + 32768
                      8 = Photon factory byte data on a log scale.
IORDER          i (R) Order of the data in the input image 
                      with respect to two local axes ax1, ax2 e.g.
                      (xf, yf) as a number from 1 to 8.
                      1   +ax1 slow   +ax2 fast    (+xf, +yf)
                      2   +ax1 slow   -ax2 fast    (+xf, -yf)
                      3   -ax1 slow   +ax2 fast    (-xf, +yf)
                      4   -ax1 slow   -ax2 fast    (-xf, -yf)
                      5   +ax2 slow   +ax1 fast    (+yf, +xf)
                      6   +ax2 slow   -ax1 fast    (+yf, -xf)
                      7   -ax2 slow   +ax1 fast    (-yf, +xf)
                      8   -ax2 slow   -ax1 fast    (-yf, -xf)

                      If a value of 0 is used, the data will be
                      returned in its input order; otherwise
                      the data will be returned with the first
                      local axis (ax1) as the slow axis and the second
                      local axis (ax2) as the fast axis in the stored image.

                      For file types >2 the order is not needed, the
                      data will be returned in the standard order
                      with +ax1 slow and +ax2 fast. As the internally
                      derived order flag may depend on the physical
                      set up of the detector and coordinate axes
                      required, the 'irotn' parameter must be set
                      in such cases.
IROTN           i (R) Orientation of rotation axis flag (ignored for 
                      IFTYPE 0,1,2)
                      =1 to right looking along the beam
                      =2 to left looking along the beam
                      =3 up looking along the beam
                      =4 down looking along the beam
ISWAP           i (R) Byte swap data =1 yes, =0 no
                      (for IFTYPE==0 or IFTYPE==2 - otherwise ignored)
NAX1_RASTS      i (R) Number of rasters along the first local axis 
                      (for IFTYPE 0,1,2 - soft check for others)
                      Must be >0 if used - otherwise may be 0 
NAX2_RASTS      i (R) Number of rasters along the second local axis
                      (for IFTYPE 0,1,2 - soft check for others)
                      Must be >0 if used - otherwise may be 0
LREC            i (R) Length of records in bytes (for IFTYPE 0 only) -
                      may give zero if record length can be calculated 
                      internally from the numbers of rasters parameters 
                      and data type assuming no padding.
PROGRESS        s (R) Subroutine which will be called at intervals 
                      during the reading of the image which may be used 
                      to display its progress if required (See below for 
                      its arguments). Note that the routine must be 
                      declared as EXTERNAL in the calling program.

                      CALL PROGRESS (IOP, NRECS, IREC)

                      IOP = -1 uncompressing a packed image
                          =  0 Start reading image data
                          =  1 Reading image data in progress
                          =  2 Image data reading completed

                      NRECS = total no. of records to read (if IOP>=0)

                      IREC = no. of the last record read (if IOP>=0)
IMG_FULL        i (W) The integer array in which the (packed) image 
                      data will be returned in the requested format.
                      (May be dummy if IOUT_TYP = 0)
MINVAL          i (W) Returns the minimum pixel value found
MAXVAL          i (W) Returns the maximum pixel value found
IOVFL           i (W) Returns overflow flag
                      = 0 data within storage range
                      = 1 data outside storage range found (these 
                          pixels set to maximum storable value)
MAX_OK          i (W) Returns the maximum storable value for iOUT_TYP 
IFAIL_SOFT      i (W) Soft checked parameters set internally differ from
                      those passed = 1 yes, =0 no 
IORDER_USED     i (W) Returns IORDER value used (-IORDER if byte swap set)
NAX1_RASTS_USED i (W) Returns NAX1_RASTS value used
NAX2_RASTS_USED i (W) Returns NAX2_RASTS value used
IERR            i (W) Error flag as returned by img_readfile
      Note: ignored parameters must be in the valid range.
-end*/

/*-C:*/
int img_readfile (char * filnam, int filtyp, int out_typ, int lhead, 
                  int typ, int order, int irotn, int swap, 
                  int nax1_rasts, int nax2_rasts,
                  int lrec, void (* progress)(), void * img_full, 
                  int * minval, int * maxval, int * ovfl, int * max_ok,
                  int * soft_fail, int * order_used, 
                  int * nax1_rasts_used, int * nax2_rasts_used)
/*end*/

/*-Parameters:
char * filnam;       The name of the image file to be read (R)
int filtyp;          File type = 0 'raw' General image. 
                                   All input items must be defined.
                               = 1 'byte' (unsigned byte data - no header)
                                   Ignores typ, swap, lhead, lrec
                               = 2 'i2' (unsigned two byte data - no header)
                                   Ignores typ, lhead, lrec
                               = 3 'mar' (MAR 'image' format)
                                   Ignores typ, order, swap, lhead, 
                                   nax1_rasts, nax2_rasts, lrec. 
                                   (Soft check on order, nax1_rasts, 
                                   nax2_rasts)
                       (R)
int out_typ;           Format required for returned data
                       1 = unsigned byte data
                       2 = unsigned two-byte data (i2)
                       3 = signed integer data (i4)
                       4 = 'squashed i2' data (if stored I<0
                           ipix = -i*8
                       5 = 'squashed i2' data if stored I<0
                           ipix = (I+32768)*8
                       6 = 'squashed i2' data if stored I<0
                           ipix = (I+32768)*32
                       7 = 'squashed i2' data if stored I<0
                           ipix = -(I+1)*256 + 32768

                       0 = do not return image data (may be used to
                           for example to get nos. of rasters from
                           a MAR file so that the required amount
                           of memory can be allocated for the image
                           prior to another call to img_readfile
                           with a value of out_typ>0)

                       Notes: out_typ==1 may only be used if filtyp==1
                              or (filtyp==0 and typ==1) 

                              If typ (see below) is 3 and out_typ is 
                              not 3 then any negative pixel values will 
                              be set to 0.
                       (R)
int lhead;             No. of header records or -(number of bytes of
                       header info). (For filtyp 0 only) (R)
int typ;               Format of stored (input) pixels data 
                       (for filtyp==0 - otherwise ignored)
                       1 = unsigned byte data
                       2 = unsigned two-byte data (i2)
                       3 = signed integer data (i4)
                       4 = 'squashed i2' data (if stored I<0
                           ipix = -i*8
                       5 = 'squashed i2' data if stored I<0
                           ipix = (I+32768)*8
                       6 = 'squashed i2' data if stored I<0
                           ipix = (I+32768)*32
                       7 = 'squashed i2' data if stored I<0
                           ipix = -(I+1)*256 + 32768
                       8 = Photon factory byte data on a log scale.
                       (R)
int order;             Order of the data in the input image 
                       with respect to two local axes ax1, ax2 e.g.
                       (xf, yf) as a number from 1 to 8.
                       1   +ax1 slow   +ax2 fast    (+xf, +yf)
                       2   +ax1 slow   -ax2 fast    (+xf, -yf)
                       3   -ax1 slow   +ax2 fast    (-xf, +yf)
                       4   -ax1 slow   -ax2 fast    (-xf, -yf)
                       5   +ax2 slow   +ax1 fast    (+yf, +xf)
                       6   +ax2 slow   -ax1 fast    (+yf, -xf)
                       7   -ax2 slow   +ax1 fast    (-yf, +xf)
                       8   -ax2 slow   -ax1 fast    (-yf, -xf)

                       If a value of 0 is used, the data will be
                       returned in its input order; otherwise
                       the data will be returned with the first
                       local axis (ax1) as the slow axis and the second
                       local axis (ax2) as the fast axis in the stored image.

                       For file types >2 the order is not needed, the
                       data will be returned in the standard order
                       with +ax1 slow and +ax2 fast. As the internally
                       derived order flag may depend on the physical
                       set up of the detector and coordinate axes
                       required, the 'irotn' parameter must be set
                       in such cases.
                       (R)
int irotn  ;           Orientation of rotation axis flag (ignored for 
                       filetyp 0,1,2)
                       =1 to right looking along the beam
                       =2 to left looking along the beam
                       =3 up looking along the beam
                       =4 down looking along the beam
                       (R)
int swap;              Byte swap data =1 yes, =0 no
                       (for filtyp==0 or filtyp==2 - otherwise ignored)
int nax1_rasts;        Number of rasters along the first local axis 
                       (for filtyp 0,1,2 - soft check for others)
                       Must be >0 if used - otherwise may be 0 (R)
int nax2_rasts;        Number of rasters along the second local axis
                       (for filtyp 0,1,2 - soft check for others)
                       Must be >0 if used - otherwise may be 0 (R)
int lrec;              Length of records in bytes (for filtyp 0 only) -
                       may give zero if record length can be calculated 
                       internally from the numbers of rasters parameters 
                       and data type assuming no padding. (R) 
void (*progress)();    Optional routine to monitor progress - set to 0 
                       if not required.

                       The routine will be called with the following 
                       parameters (Note: pointers used as routine may 
                       be a Fortran subroutine)

                       progress (int *iop, int * nrecs, int *irec)

                       iop = -1 uncompressing a packed image
                           =  0 Start reading image data
                           =  1 Reading image data in progress
                           =  2 Image data reading completed

                       nrecs = total no. of records to read (if iop>=0)

                       irec = no. of the last record read (if iop>=0)
                       (R)
void *img_full;        Array in which to return the image data (W)
                       (May be dummy if out_typ==0)
int * minval;          Returns the minimum pixel value found (W)
int * maxval;          Returns the maximum pixel value found (W)
int * ovfl;            Returns overflow flag
                          = 0 data within storage range
                          = 1 data outside storage range found (these 
                              pixels set to maximum storable value)
                       (W)
int * max_ok;          Returns the maximum storable value for out_typ
int * soft_fail;       Soft checked parameters set internally differ from
                       those passed = 1 yes, =0 no (W)
int * order_used;      'order' value used  (-order if byte swap) (W)
int * nax1_rasts_used; 'nax1_rasts' value used (W)
int * nax2_rasts_used; 'nax2_rasts' value used (W)

      Note: ignored parameters must be in the valid range.

Return = 0   OK
       > 0   Record number where a read error occurred
       =-1   Cannot open file
       =-10  Cannot allocate memory required for temp buffer(s)
       =-20  invalid parameters in call
       =-100 Error reading header
       =-200 Header record swapped or unswapped does not give
             nax1_rasts = 1200, 1600, 1800, 2000, 2300, 2400, 3000, 3450
-end*/

{
   long offset_data;
   long ovfl_data_offset;
   I2pix *img;
   FILE *fp;
   int fd;
   int read_file;
   int old_mar;
   int nrecs;
   int nrasts;
   int iswap;
   int ityp;
   int iord;
   int recl;
   int nbpix;
   int nxrasts;
   int nyrasts;
   int novfl;
   int nov_rec;
   int i, k, l;
   int iret;
   int hdr[6];
   int ovpix[2];

   unsigned char * cbuf;
   unsigned char * ovbuf;
   unsigned char tempc;

   extern void readpack_word_c(short *, char *); /* from CCP4 pack_c.c */

   /* Initialisations */
   *soft_fail = 0;
   read_file = 1;

   /* Check parameters in call */

   if (filtyp<0||filtyp>3) return -20;
   if (out_typ<0||out_typ>7) return -20;
   if (filtyp<=2&&(typ<1||typ>8)) return -20;
   if (order<0||order>8) return -20;
   if (irotn<1||irotn>4) return -20;
   if (swap!=0&&swap!=1) return -20;
   if (filtyp<=2)
   {
      if (nax1_rasts<1) return -20;
      if (nax2_rasts<1) return -20;
   }
   if (lrec<0) return -20;

   /* Open file */

   fp = fopen (filnam, "rb");
   if (fp==0) return -1;
   fd = fileno(fp);

   /* Read header and set flags etc. for return and/or reading the data 
      and set file to start of data records */

   nrecs = nax1_rasts;
   nrasts = nax2_rasts;
   if (order>4)
   {
      nrecs = nax2_rasts;
      nrasts = nax1_rasts;
   }

   iswap = swap;
   recl = 0;

#if 0				/* DL: nbpix undefined -- what's this
                                   doing here?  */
   if (lrec==0) lrec = nrasts*nbpix;
#endif
   switch (filtyp)
   {
      case 0: /* 'raw' */
      ityp = typ;
      iord = order;
      nbpix = 1;
      if (typ>1) nbpix = 2;
      if (typ==3) nbpix = 4;
      recl = nrasts*nbpix;
      if (lrec>recl) recl = lrec;
      if (lhead>=0)
      {
         offset_data = lhead*recl;
      }
      else
      {
         offset_data = -lhead;
      }
      lseek (fd, offset_data, 0);
      nxrasts = nax1_rasts;
      nyrasts = nax2_rasts;
      break;

      case 1: /* 'film' */
      ityp = 1;
      iord = order;
      recl = nrasts;
      lseek (fd, 0L, 0);
      nxrasts = nax1_rasts;
      nyrasts = nax2_rasts;
      break;

      case 2: /* 'i2' */
      ityp = 2;
      iord = order;
      recl = nrasts*2;
      lseek (fd, 0L, 0);
      nxrasts = nax1_rasts;
      nyrasts = nax2_rasts;
      break;

      case 3: /* 'mar' */
      ityp = 2;
      iswap = 0;
      lseek (fd, 0L, 0);
      cbuf = (unsigned char *) hdr;
      i = read (fd, hdr, 24);
      if (i!=24) return -100;
      iswap = 0;
      if (hdr[0]!=1234&&hdr[0]!=1200&&hdr[0]!=1600
          &&hdr[0]!=1800&&hdr[0]!=2000&&hdr[0]!=2300
          &&hdr[0]!=2400&&hdr[0]!=3000&&hdr[0]!=3450) iswap = 1;
      if (iswap)
      {
         l=0;
         for (k=0;k<6;++k)
         {
             tempc = cbuf[l];
             cbuf[l] = cbuf[l+3];
             cbuf[l+3] = tempc;
             tempc = cbuf[l+1];
             cbuf[l+1] = cbuf[l+2];
             cbuf[l+2] = tempc;
             l += 4;
         }
          if (hdr[0]!=1234&&hdr[0]!=1200&&hdr[0]!=1600
          &&hdr[0]!=1800&&hdr[0]!=2000&&hdr[0]!=2300
          &&hdr[0]!=2400&&hdr[0]!=3000&&hdr[0]!=3450) return -200;
      }
      old_mar = 1;
      if (hdr[0]==1234) old_mar = 0;
      if (old_mar)
      {
         if (irotn==1||irotn==2)
         {
            iord = 5;
            if (irotn==2) iord = 8;
            nxrasts = hdr[0];
            nyrasts = hdr[1];
            novfl = hdr[4];
            offset_data = 2*nyrasts;
            ovfl_data_offset = 2*(nxrasts+1)*nyrasts;
         }
         else
	 {
            iord = 3;
            if (irotn==4) iord = 2;
            nyrasts = hdr[0];
            nxrasts = hdr[1];
            novfl = hdr[4];
            offset_data = 2*nxrasts;
            ovfl_data_offset = 2*(nyrasts+1)*nxrasts;
         }
         lseek (fd, offset_data, 0);
      }
      else
      {
         if (irotn==1||irotn==2)
         {
            iord = 3;
            if (irotn==2) iord = 2;
            nxrasts = hdr[1];
            nyrasts = hdr[1];
            novfl = hdr[2];
            nov_rec = (novfl+7)/8;
            if (hdr[3]!=0&&hdr[3]!=1) return -200;
            offset_data = 4096 + 64*nov_rec;
            ovfl_data_offset = 4096;
         }
         else
	 {
            iord = 8;
            if (irotn==4) iord = 5;
            nyrasts = hdr[1];
            nxrasts = hdr[1];
            novfl = hdr[2];
            nov_rec = (novfl+7)/8;
            if (hdr[3]!=0&&hdr[3]!=1) return -200;
            offset_data = 4096 + 64*nov_rec;
            ovfl_data_offset = 4096;
	 }
         lseek (fd, offset_data, 0);
         if (hdr[3]==1&&out_typ!=0)
         {
            read_file = 0;
            img =  (I2pix *) malloc(2*nxrasts*nyrasts);
            if (img==0) return -10;
            readpack_word_c (img, filnam);
         }
      }
      break;
   }

   /* Read/reorder data */

   iret = 0;
   if (out_typ!=0) iret = img_read_data (read_file, fd, (void *) img, 
                         out_typ, ityp, iord, iswap, nxrasts, nyrasts, 
                         recl, progress, img_full, minval, maxval, ovfl, 
                         max_ok);

   /* Process overflow records for MAR */

   if (filtyp==3&&out_typ!=0)
   {
      lseek (fd,  ovfl_data_offset, 0);
      ovbuf = (unsigned char * ) &ovpix;
      for (i=0;i<novfl;++i)
      {
         read(fd,ovbuf,8);
         if (iswap)
	 {
            l=0;
            for (k=0;k<2;++k)
            {
               tempc = ovbuf[l];
               ovbuf[l] = ovbuf[l+3];
               ovbuf[l+3] = tempc;
               tempc = ovbuf[l+1];
               ovbuf[l+1] = ovbuf[l+2];
               ovbuf[l+2] = tempc;
               l += 4;
            }
         }
         set_mar_ovpix (img_full, ovpix, iord, nxrasts, nyrasts, 
                        out_typ, ovfl);
         if (ovpix[1]>*maxval) *maxval = ovpix[1];
      }
   }

   if (read_file==0&&out_typ!=0&&img!=0) free(img);
   *order_used = iord;
   if (iswap==1) *order_used = -iord;
   *nax1_rasts_used = nxrasts;
   *nax2_rasts_used = nyrasts;

   if (iret!=0) return iret;

   return 0;
}
/* Fortran binding: imgf_readfile */

void imgf_readfile (int * filnam, int * iftype, int * iout_typ, int * lhead,
                    int * ityp, int * iorder, int * irotn, int * iswap,
                    int * nax1_rasts, int * nax2_rasts, int * lrec,
                    void (*progress)(), int * img_full, int * minval, 
                    int * maxval, int * iovfl, int * max_ok,
                    int * ifail_soft, int * iord_used, 
                    int * nax1_rasts_used, int * nax2_rasts_used, int * ierr)
{
   *ierr = img_readfile (IMG_pointers[*filnam], *iftype, *iout_typ, *lhead,
                      *ityp, *iorder, *irotn, *iswap,
                      *nax1_rasts, *nax2_rasts, *lrec,
                      progress, img_full, minval, maxval, iovfl, max_ok,
                      ifail_soft, iord_used, nax1_rasts_used, nax2_rasts_used);
   return;
}

/*-Routine: Read image data - img_read_data
The routine img_read_data ('C' routine only available) is used to read
image pixels data from file or a stored image and re-order and re-pack it
as required. When reading from a file, the file must be already open
and positioned at the start of the pixels data.
-end*/

/*
****************************
**     img_read_data      **
****************************

Purpose: Read image pixels data from file or a stored image;
         Re-order & re-pack as required.

Author:  John W. Campbell, August 1997

         NOTE: When reading from a file, the file must be already open
               and positioned at the start of the pixels data.


*/

/*-C:*/
int img_read_data (int read_file, int fd, void * img, int out_typ, int typ, 
                   int order, int swap, int nax1_rasts, int nax2_rasts,
                   int recl, void (* progress)(), void * img_full, 
                   int * minval, int * maxval, int * ovfl, int * max_ok)
/*end*/

/*-Parameters:
int read_file;     Read file or image data flag = 1 read from file
                                                = 0 read from image
                   (R)
int fd;            File descriptor if read_file is set. (R)
void * img;        Pointer to start of image pixels data if read_file==0 (R)
int out_typ;       Format required for returned data
                   1 = unsigned byte data
                   2 = unsigned two-byte data (i2)
                   3 = signed integer data (i4)
                   4 = 'squashed i2' data (if stored I<0
                       ipix = -i*8
                   5 = 'squashed i2' data if stored I<0
                       ipix = (I+32768)*8
                   6 = 'squashed i2' data if stored I<0
                       ipix = (I+32768)*32
                   7 = 'squashed i2' data if stored I<0
                       ipix = -(I+1)*256 + 32768

                   Notes: out_typ==1 may only be used if typ==1 (see below)
                   (R)

int typ;           Format of stored (input) pixels data
                   1 = unsigned byte data
                   2 = unsigned two-byte data (i2)
                   3 = signed integer data (i4)
                   4 = 'squashed i2' data (if stored I<0
                       ipix = -i*8
                   5 = 'squashed i2' data if stored I<0
                       ipix = (I+32768)*8
                   6 = 'squashed i2' data if stored I<0
                       ipix = (I+32768)*32
                   7 = 'squashed i2' data if stored I<0
                       ipix = -(I+1)*256 + 32768
                   8 = Photon factory byte data on a log scale.
                   (R)

int order;         Order of the data in the input image 
                   with respect to two local axes ax1, ax2 e.g.
                   (xf, yf) as a number from 1 to 8.
                   1   +ax1 slow   +ax2 fast    (+xf, +yf)
                   2   +ax1 slow   -ax2 fast    (+xf, -yf)
                   3   -ax1 slow   +ax2 fast    (-xf, +yf)
                   4   -ax1 slow   -ax2 fast    (-xf, -yf)
                   5   +ax2 slow   +ax1 fast    (+yf, +xf)
                   6   +ax2 slow   -ax1 fast    (+yf, -xf)
                   7   -ax2 slow   +ax1 fast    (-yf, +xf)
                   8   -ax2 slow   -ax1 fast    (-yf, -xf)

                   If a value of 0 is used, the data will be
                   returned in its input order; otherwise
                   the data will be returned with the first
                   local axis (ax1) as the slow axis and the second
                   local axis (ax2) as the fast axis in the stored image.
                   (R)

int swap;          Swap bytes flag =1 yes, =0 no (R)
int nax1_rasts;    Number of rasters along the first local axis (R)
int nax2_rasts;    Number of rasters along the second local axis (R)
int recl;          Record length in bytes; if given as 0 then
                   record length will be calculated internally
                   from the numbers of rasters parameters and
                   data type assuming no padding (R)
(void *)()progress; Progress monitoring routine (may be null if not required)
                    (R)
void * img_full;   Array in which to return the image data (W)
int * minval;      Returns the minimum pixel value found (W)
int * maxval;      Returns the maximum pixel value found (W)
int * ovfl;        Flag =0 no data outside range
                        =1 pixel data overflows range (overflow values
                           will be set to largest value which can be
                           stored in the requested format.
                   (W)

Return = 0  OK
       > 0  Record number where a read error occurred
       =-10 Cannot allocate memory required for temp buffer
       =-20 Invalid parameters in call

-end*/

{
   int i, j, k, l;           /* Temp/loop variables */
   int irec;                 /* Record counter */
   int lrec;                 /* Record length */
   int step;                 /* Step through output array */
   int nrecs, nrasts;        /* No. of records, no. of rasters/record */
   unsigned char *buf;       /* Buffer from file or pointer to record in
                                input image */
   unsigned char tempc;      /* Temp unsigned character */
   int ir;		     /* Counter */
   int nbpix;                /* No. of bytes per stored pixel */
   int out_siz;              /* No. of bytes per pixel in output image */
   int ip;                   /* Progress routine flag */
   int intens;               /* Pixel intensity */
   unsigned char *imgbyt;    /* Pointer to input image */
   unsigned char *byte_data; /* Pointer to unsigned byte data */
   I2pix *i2_data;           /* Pointer to 'i2' data */
   Intpix *int_data;         /* Pointer to signed integer data */
   I2pix *i2_buf;            /* Pointer to buffer as 'i2' data */
   Intpix *int_buf;          /* Pointer to buffer as integer data */
   int (*unsq)();            /* Unsquash packed I2 routine */
   I2pix (*sqsh)();         /* Squash into packed I2 routine */

   /* Initialisations */

   *minval = 0;
   *maxval = 0;
   *ovfl = 0;
   *max_ok = 0;

   /* Check input parameters */

   if (order<1||order>8) return -20;
   if (out_typ<1||out_typ>7) return -20;
   if (typ<1||typ>8) return -20;
   if (out_typ==1&&typ!=1) return -20;

   *minval =  10000000;
   *maxval = -10000000;

   unsq = 0;
   sqsh = 0;

   if (!read_file) imgbyt = (unsigned char *) img;

   lrec = recl;

   switch (typ)
   {
       case 2:
       unsq = img_unsq1;
       break;

       case 4:
       unsq = img_unsq2;
       break;

       case 5:
       unsq = img_unsq3;
       break;

       case 6:
       unsq = img_unsq4;
       break;

       case 7:
       unsq = img_unsq5;
       break;
   }

   switch (out_typ)
   {
       case 1:
       *max_ok = 255;
       break;

       case 2:
       sqsh = img_sqsh1;
       *max_ok = 65535;
       break;

       case 3:
       *max_ok = 2147483647;
       break;

       case 4:
       sqsh = img_sqsh2;
       *max_ok = 262144;
       break;

       case 5:
       sqsh = img_sqsh3;
       *max_ok = 262136;
       break;

       case 6:
       sqsh = img_sqsh4;
       *max_ok = 1048544;
       break;

       case 7:
       sqsh = img_sqsh5;
       *max_ok = 8421120;
       break;
   }

   byte_data = (unsigned char *) img_full;
   i2_data = (I2pix *) img_full;
   int_data = (Intpix *) img_full;

   nbpix = 1;
   if (typ>1) nbpix = 2;
   if (typ==3) nbpix = 4;

   out_siz = 1;
   if (out_typ>1) out_siz = 2;
   if (out_typ==3) out_siz = 4;

   nrecs = nax1_rasts;
   nrasts = nax2_rasts;
   if (order>4)
   {
      nrecs = nax2_rasts;
      nrasts = nax1_rasts;
   }

   if (lrec==0) lrec = nrasts*nbpix;
 
   if (progress!=0)
   {
      ip = 0;
      ir = 0;
      progress (&ip, &nrecs, &ir);
   }
   
   /* Read data */

   if (read_file) buf = (unsigned char *) malloc(lrec);
   if (buf==0) return -10;
   irec = 1;
   for (ir=0;ir<nrecs;++ir)
   {
      if (read_file)
      {
         i = read (fd, buf, lrec);
         if (i!=nbpix*nrasts) goto tidy;
      }
      else
      {
         buf = (unsigned char *) &imgbyt[ir*lrec];
      }
      switch (order)
      {
         case 1:
            j = ir*nrasts;
            step = 1;
            break;
         case 2:
            j = (ir+1)*nrasts - 1;
            step = -1;
            break;
         case 3:
            j = (nrecs-ir-1)*nrasts;
            step = 1;
            break;
         case 4:
            j = (nrecs-ir)*nrasts - 1;
            step = -1;
            break;
         case 5:
            j = ir;
            step = nrecs;
            break;
         case 6:
            j = nrecs - 1 - ir;
            step = nrecs;
            break;
         case 7:
            j = (nrasts-1)*nrecs + ir;
            step = -nrecs;
            break;
         case 8:
            j = nrasts*nrecs - 1 - ir;
            step = -nrecs;
            break;
      }
      if (swap)
      {
         switch (nbpix)
         {
            case 1:
               break;
           
            case 2:
               l=0;
               for (k=0;k<nrasts;++k)
               {
                   tempc = buf[l];
                   buf[l] = buf[l+1];
                   buf[l+1] = tempc;
                   l += 2;
               }
               break;
           
            case 4:
               l=0;
               for (k=0;k<nrasts;++k)
               {
                  tempc = buf[l];
                  buf[l] = buf[l+3];
                  buf[l+3] = tempc;
                  tempc = buf[l+1];
                  buf[l+1] = buf[l+2];
                  buf[l+2] = tempc;
                  l += 4;
               }
               break;
           
            case 8:
               l=0;
               for (k=0;k<nrasts;++k)
               {
                  tempc = buf[l];
                  buf[l] = buf[l+7];
                  buf[l+7] = tempc;
                  tempc = buf[l+1];
                  buf[l+1] = buf[l+6];
                  buf[l+6] = tempc;
                  tempc = buf[l+2];
                  buf[l+2] = buf[l+5];
                  buf[l+5] = tempc;
                  tempc = buf[l+3];
                  buf[l+3] = buf[l+4];
                  buf[l+4] = tempc;
                  l += 8;
               }
               break;
         }
      }

      if (out_typ==typ)
      {
         switch (nbpix)
         {
            case 1:
            for (k=0;k<nrasts;++k)
            {
                byte_data[j] = buf[k];
                if (buf[k]<*minval) *minval = buf[k];
                if (buf[k]>*maxval) *maxval = buf[k];
                j += step;
            }
            break;

            case 2:
            i2_buf = (I2pix *) buf;
            for (k=0;k<nrasts;++k)
            {
                i2_data[j] = i2_buf[k];
                intens = unsq(i2_buf[k]);
                if (intens<*minval) *minval = intens;
                if (intens>*maxval) *maxval = intens;
                j += step;
            }
            break;

            case 4:
            int_buf = (Intpix *) buf;
            for (k=0;k<nrasts;++k)
            {
                int_data[j] = int_buf[k];
                intens = int_buf[k];
                if (intens<*minval) *minval = intens;
                if (intens>*maxval) *maxval = intens;
                j += step;
            }
            break;
         }
      }
      else
      {
         switch (out_siz)
         {
            case 2:
            switch (nbpix)
            {
               case 1:
               for (k=0;k<nrasts;++k)
               {
                   if(typ==8)
                   {
                      intens = img_byte_log(buf[k]);
                   }
                   else
                   {
                      intens = (int) buf[k];
                   }

                   i2_data[j] = sqsh(intens, ovfl);
                   if (intens<*minval) *minval = intens;
                   if (intens>*maxval) *maxval = intens;
                   j += step;
               }
               break;

               case 2:
               i2_buf = (I2pix *) buf;
               for (k=0;k<nrasts;++k)
               {
                   intens = unsq(i2_buf[k]);
                   i2_data[j] = sqsh(intens, ovfl);
                   if (intens<*minval) *minval = intens;
                   if (intens>*maxval) *maxval = intens;
                   j += step;
               }
               break;

               case 4:
               int_buf = (Intpix *) buf;
               for (k=0;k<nrasts;++k)
               {
                   intens = int_buf[k];
                   i2_data[j] = sqsh(intens, &ovfl);
                   if (intens<*minval) *minval = intens;
                   if (intens>*maxval) *maxval = intens;
                   j += step;
               }
               break;
            }
            break;

            case 4:
            switch (nbpix)
            {
               case 1:
               for (k=0;k<nrasts;++k)
               {
                   if(typ==8)
                   {
                      intens =  img_byte_log(buf[k]);
                   }
                   else
                   {
                      intens = (int) buf[k];
                   }
                   int_data[j] = intens;
                   if (intens<*minval) *minval = intens;
                   if (intens>*maxval) *maxval = intens;
                   j += step;
               }
               break;

               case 2:
               i2_buf = (I2pix *) buf;
               for (k=0;k<nrasts;++k)
               {
                   intens = unsq(i2_buf[k]);
                   int_data[j] = intens;
                   if (intens<*minval) *minval = intens;
                   if (intens>*maxval) *maxval = intens;
                   j += step;
               }
               break;
            }
            break;
         }
      }

      if (progress!=0)
      {
         ip = 1;
         progress(&ip, &nrecs, &irec);
      }
      irec++;
   }
   irec = 0;   
   
   /* Tidy up */

   tidy:
   if (read_file) free (buf);
   if (progress!=0)
   {
      ip = 2;
      ir = irec;
      if (irec==0) ir = nrecs;
      progress(&ip, &nrecs, &ir);
   }
   return irec;
}

/*-Section: Extracting Parts of an Image
Routines are available to extract a rectangle, line or point from an 
in memory image which may have its data in any of 8 possible orders with 
various options for the storage of the pixels data. All data are returned
as signed integer values.
-end*/

/*-Routine: Get rectangular section from image - img_getrec
The routine img_getrect (imgf_getrect) extracts a rectangular section
from the in-memory data returning it in an integer array.
-end*/

/*                                            
***********************
**   img_getrect     **
***********************

Purpose: Get a rectangular section from an in-memory image

Author:  John W. Campbell, July 1997

CD-Fortran:
      SUBROUTINE IMG_GETRECT (IMG, ITYPE, IORDER, NF_OFF, 
     +                        IAX1_ORIG, IAX2_ORIG, NAX1, NAX2,
     +                        NAX1_RASTS, NAX2_RASTS, IOUTLIM,
     +                        IDATA, IERR)
CD-end
*/
/*-Parameters:
IMG()      i (R)  Image data array (may hold signed integer data or
                  unsigned byte data packed into an integer array or
                  unsigned two byte data packed into an integer array
                  or 'squashed' two byte data packed into an integer
                  array.
ITYPE      i (R)  Image data type 
                  1 = unsigned byte data
                  2 = unsigned two-byte data (i2)
                  3 = signed integer data
                  4 = 'squashed i2' data (if stored I<0
                      ipix = -i*8
                  5 = 'squashed i2' data if stored I<0
                      ipix = (I+32768)*8
                  6 = 'squashed i2' data if stored I<0
                      ipix = (I+32768)*32
                  7 = 'squashed i2' data if stored I<0
                      ipix = -(I+1)*256 + 32768
IORDER     i (R)  Order of data in image wrt. the two local axes
                  (slow then fast)
                  =1 +ax1 +ax2
                  =2 +ax1 -ax2
                  =3 -ax1 +ax2
                  =4 -ax1 -ax2
                  =5 +ax2 +ax1
                  =6 +ax2 -ax1
                  =7 -ax2 +ax1
                  =8 -ax2 -ax1
NF_OFF     i (R)  Fast offset between (slow) rasters of image (e.g. 
                  first dimension of the image data if the data are 
                  held in a two dimensional array). This value is the 
                  number of elements for the image data type (bytes, 
                  2-bytes, integers or 'squashed i2' as set by ITYPE)
IAX1_ORIG  i (R)  The origin pixel of the required section along the 
                  first local axis
IAX2_ORIG  i (R)  The origin pixel of the required section along the 
                  second local axis
NAX1       i (R)  Size of the section in pixels along the first local
                  axis
NAX2       i (R)  Size of the section in pixels along the second local
                  axis
NAX1_RASTS i (R)  No. of rasters in the image along the first 
                  local axis    
NAX2_RASTS i (R)  No. of rasters in the image along the second 
                  local axis
IOUTLIM    i (R)  Value to which returned data pixels are to be set if
                  they lie outside the image
IDATA()    i (W)  Array returning the image section data first local
                  axis is the slow axis (2'nd dimension) and second
                  local axis is the fast axis (1'st) dimension.
IERR       i (W)  Error flag = 0 OK,
                             =-1 Invalid type or order flag
                             = 1 Box partly outside image
                             = 2 Box totally outside image

                 Note: pixels within the image are numbered from 1 upwards
-end*/
/*-C:*/
int img_getrect (int * img, int itype, int iorder, int nf_off, 
                 int iax1_orig, int iax2_orig, int nax1, int nax2,
                 int nax1_rasts, int nax2_rasts, int ioutlim,
                 int *idata)
/*end*/

/*-Parameters:
int * img;     Image data array (may hold signed integer data or
               unsigned byte data or unsigned two byte data or 'squashed' 
               two byte data - data will be cast to an appropriate type (R)
int itype;     Image data type 
               1 = unsigned byte data
               2 = unsigned two-byte data (i2)
               3 = signed integer data
               4 = 'squashed i2' data (if stored I<0
                   ipix = -i*8
               5 = 'squashed i2' data if stored I<0
                   ipix = (I+32768)*8
               6 = 'squashed i2' data if stored I<0
                   ipix = (I+32768)*32
               7 = 'squashed i2' data if stored I<0
                   ipix = -(I+1)*256 + 32768 
               (R)
int iorder;    Order of data in image wrt. the two local axes
               (slow then fast)
               =1 +ax1 +ax2
               =2 +ax1 -ax2
               =3 -ax1 +ax2
               =4 -ax1 -ax2
               =5 +ax2 +ax1
               =6 +ax2 -ax1
               =7 -ax2 +ax1
               =8 -ax2 -ax1
               (R)
int nf_off;    Fast offset between (slow) rasters of image (e.g. 
               first dimension of the image data if the data are 
               held in a two dimensional array). This value is the 
               number of elements for the image data type (bytes, 
               2-bytes, integers or 'squashed i2' as set by itype) (R)
int iax1_orig; The origin pixel of the required section along the 
               first local axis (R)
int iax2_orig; The origin pixel of the required section along the 
               second local axis (R)
int nax1;      Size of the section in pixels along the first local
               axis (R)
int nax2;      Size of the section in pixels along the second local
               axis (R)
int nax1_rasts No. of rasters in the image along the first 
               local axis (R)    
int nax2_rasts No. of rasters in the image along the second 
               local axis (R)
int ioutlim;   Value to which returned data pixels are to be set if
               they lie outside the image (R)
int * idata;   Array returning the image section data first local
               axis is the slow axis (2'nd dimension) and second
               local axis is the fast axis (1'st) dimension. (W)

               Note: input pixel numbers are numbered from 1 upwards

Return:  Error flag = 0 OK,
                    =-1 Invalid type or order flag
                    = 1 Box partly outside image
                    = 2 Box totally outside image

-end*/
{
   int islow_min;            /* Minimum pixel along image slow axis */
   int islow_max;            /* Maximum pixel along image slow axis */
   int ifast_min;            /* Minimum pixel along image fast axis */
   int ifast_max;            /* Maximum pixel along image fast axis */
   int ns_rasts;             /* No. of rasters in image along slow 
                                direction */
   int nf_rasts;             /* No. of rasters in image along fast 
                                direction */
   int iax1;                 /* Loop through local axis 1 */
   int iax1_start;           /* Start of loop through local axis 1 */
   int iax1_end;             /* End of loop through local axis 1 */
   int iax1_step;            /* Loop step for local axis 1 */
   int iax2;                 /* Loop through local axis 2 */
   int iax2_start;           /* Start of loop through local axis 2 */
   int iax2_end;             /* End of loop through local axis 2 */
   int iax2_step;            /* Loop step through local axis 2 */
   int ks;                   /* Counter for slow axis */
   int kf;                   /* Counter for fast axis */
   int koff;                 /* Offset to start of current fast axis 
                                raster */
   int ktype;                /* Data type flag */
   int err;                  /* Error flag */
   long ipix;                /* Pixel value */ 
   Intpix *int_data;         /* Pointer to signed integer data */
   I2uns *i2_data;           /* Pointer to image plate 'i2' data */
   I2squash *i2_squash;      /* Pointer to 'squashed i2' data */
   unsigned char *byte_data; /* Pointer to byte image data*/
   long (*unsq)();           /* Function to unpack a squashed
                                image pixel value */

   err = 0; 

/* Check for valid flag values */

   if (itype<1||itype>7) return -1;
   if (iorder<1||iorder>8) return -1;

/* Get pointer to image data */

   unsq = 0;
   ktype = itype;
   if (ktype>4) ktype = 4;
   switch (ktype)
   {
      case 1:
      byte_data = (unsigned char *) img;
      break;

      case 2:
      i2_data = (I2uns *) img;
      break;

      case 3:
      int_data = (Intpix *) img;
      break;

      case 4:
      i2_squash = (I2squash *) img;
      switch (itype)
      {
         case 4:
         unsq = img_desq;
         break;

         case 5:
         unsq = img_desq2;
         break;

         case 6:
         unsq = img_desq3;
         break;

         case 7:
         unsq = img_desq4;
         break;
      }
      break;
   }

/* Get Box limits in image array and loop parameters for returning
   data array */

   switch(iorder)
   {
      case 1:
         islow_min = iax1_orig;
         ifast_min = iax2_orig;
         iax1_start = 1;
         iax1_end = nax1;
         iax1_step = 1;
         iax2_start = 1;
         iax2_end = nax2;
         iax2_step = 1;
      break;

      case 2:
         islow_min = iax1_orig;
         ifast_min = nax2_rasts - iax2_orig - nax2 + 2;
         iax1_start = 1;
         iax1_end = nax1;
         iax1_step = 1;
         iax2_start = nax2;
         iax2_end = 1;
         iax2_step = -1;
      break;

      case 3:
         islow_min = nax1_rasts - iax1_orig - nax1 + 2;
         ifast_min = iax2_orig;
         iax1_start = nax1;
         iax1_end = 1;
         iax1_step = -1;
         iax2_start = 1;
         iax2_end = nax2;
         iax2_step = 1;
      break;

      case 4:
         islow_min = nax1_rasts - iax1_orig - nax1 + 2;
         ifast_min = nax2_rasts - iax2_orig - nax2 + 2;
         iax1_start = nax1;
         iax1_end = 1;
         iax1_step = -1;
         iax2_start = nax2;
         iax2_end = 1;
         iax2_step = -1;
      break;

      case 5:
         islow_min = iax2_orig;
         ifast_min = iax1_orig;
         iax2_start = 1;
         iax2_end = nax2;
         iax2_step = 1;
         iax1_start = 1;
         iax1_end = nax1;
         iax1_step = 1;
      break;

      case 6:
         islow_min = iax2_orig;
         ifast_min = nax1_rasts - iax1_orig - nax1 + 2;
         iax2_start = 1;
         iax2_end = nax2;
         iax2_step = 1;
         iax1_start = nax1;
         iax1_end = 1;
         iax1_step = -1;
      break;

      case 7:
         islow_min = nax2_rasts - iax2_orig - nax2 + 2;
         ifast_min = iax1_orig;
         iax2_start = nax2;
         iax2_end = 1;
         iax2_step = -1;
         iax1_start = 1;
         iax1_end = nax1;
         iax1_step = 1;
      break;

      case 8:
         islow_min = nax2_rasts - iax2_orig - nax2 + 2;
         ifast_min = nax1_rasts - iax1_orig - nax1 + 2;
         iax2_start = nax2;
         iax2_end = 1;
         iax2_step = -1;
         iax1_start = nax1;
         iax1_end = 1;
         iax1_step = -1;
      break;
   }
   if (iorder<5)
   {
      islow_max = islow_min + nax1 - 1;
      ifast_max = ifast_min + nax2 - 1;
      ns_rasts = nax1_rasts;
      nf_rasts = nax2_rasts;
   }
   else
   {
      islow_max = islow_min + nax2 - 1;
      ifast_max = ifast_min + nax1 - 1;
      ns_rasts = nax2_rasts;
      nf_rasts = nax1_rasts;
   }

/* If some or all of requested section lies outside image then
   initialise returned data array to the given out of limits value */

   if (islow_min<1||islow_max>ns_rasts||ifast_min<1||ifast_max>nf_rasts)
   {
      err = 1;
      for (iax1=0;iax1<nax1;++iax1)
      {
          for (iax2=0;iax2<nax2;++iax2)
          {
             idata[iax1*nax2+iax2] = ioutlim;
          }
      }
      if (islow_min>ns_rasts||islow_max<1) return 2;
      if (ifast_min>nf_rasts||ifast_max<1) return 2;
   }

/* Get image data */

   if (iorder<5)
   {
      ks = islow_min - 1; 
      koff = (ks-1)*nf_off;
      for (iax1=iax1_start;iax1<=iax1_end;iax1+=iax1_step)
      {
         ks++;
         koff = koff + nf_off;
         if (ks<1||ks>ns_rasts) continue;
         kf = ifast_min - 1;
         for (iax2=iax2_start;iax2<=iax2_end;iax2+=iax2_step)
         {
            kf++;
            if (kf<1||kf>nf_rasts) continue;
            switch (ktype)
            {
               case 1:
               ipix = (long) byte_data[koff+kf-1];
               break;

               case 2:
               ipix = (long) i2_data[koff+kf-1];
               break;

               case 3:
               ipix = (long) int_data[koff+kf-1];
               break;

               case 4:
               ipix = (long) unsq(i2_squash[koff+kf-1]);
               break;
            }
            idata[(iax1-1)*nax2+iax2-1] = ipix;
         }
      }
   }
   else
   {
      ks = islow_min - 1; 
      koff = (ks-1)*nf_off;
      for (iax2=iax2_start;iax2<=iax2_end;iax2+=iax2_step)
      {
         ks++;
         koff = koff + nf_off;
         if (ks<1||ks>ns_rasts) continue;
         kf = ifast_min - 1;
         for (iax1=iax1_start;iax1<=iax1_end;iax1+=iax1_step)
         {
            kf++;
            if (kf<1||kf>nf_rasts) continue;
            switch (ktype)
            {
               case 1:
               ipix = (long) byte_data[koff+kf-1];
               break;

               case 2:
               ipix = (long) i2_data[koff+kf-1];
               break;

               case 3:
               ipix = (long) int_data[koff+kf-1];
               break;

               case 4:
               ipix = (long) unsq(i2_squash[koff+kf-1]);
               break;
            }
            idata[(iax1-1)*nax2+iax2-1] = ipix;
         }
      }
   }
   return err;
}
/* Fortran binding: imgf_getrect */

void imgf_getrect (int * img, int * itype, int * iorder, int * nf_off, 
                   int * iax1_orig, int * iax2_orig, int * nax1, int * nax2,
                   int * nax1_rasts, int * nax2_rasts, int * ioutlim,
                   int *idata, int *ierr)
{
   *ierr = img_getrect (img, *itype, *iorder, *nf_off, 
                        *iax1_orig, *iax2_orig, *nax1, *nax2,
                        *nax1_rasts, *nax2_rasts, *ioutlim, idata);
   return;
} 

/*-Routine: Get first local axis line from image - img_ax1line
The routine img_axiline (imgf_ax1line) extracts an image line along
the direction of the first local axis from the in-memory data returning 
it in an integer array.
-end*/

/*                                            
***********************
**   img_ax1line     **
***********************

Purpose: Get a line (along local axis 1) from an in-memory image

Author:  John W. Campbell, July 1997

CD-Fortran:
      SUBROUTINE IMG_AX1LINE (IMG, ITYPE, IORDER, NF_OFF, 
     +                        IAX1_ORIG, IAX2_ORIG, NAX1,
     +                        NAX1_RASTS, NAX2_RASTS, IOUTLIM,
     +                        IDATA, IERR)
CD-end
*/
/*-Parameters:
IMG()      i (R)  Image data array (may hold signed integer data or
                  unsigned byte data packed into an integer array or
                  unsigned two byte data packed into an integer array
                  or 'squashed' two byte data packed into an integer
                  array.
ITYPE      i (R)  Image data type 
                  1 = unsigned byte data
                  2 = unsigned two-byte data (i2)
                  3 = signed integer data
                  4 = 'squashed i2' data (if stored I<0
                      ipix = -i*8
                  5 = 'squashed i2' data if stored I<0
                      ipix = (I+32768)*8
                  6 = 'squashed i2' data if stored I<0
                      ipix = (I+32768)*32
                  7 = 'squashed i2' data if stored I<0
                      ipix = -(I+1)*256 + 32768
IORDER     i (R)  Order of data in image wrt. the two local axes
                  (slow then fast)
                  =1 +ax1 +ax2
                  =2 +ax1 -ax2
                  =3 -ax1 +ax2
                  =4 -ax1 -ax2
                  =5 +ax2 +ax1
                  =6 +ax2 -ax1
                  =7 -ax2 +ax1
                  =8 -ax2 -ax1
NF_OFF     i (R)  Fast offset between (slow) rasters of image (e.g. 
                  first dimension of the image data if the data are 
                  held in a two dimensional array). This value is the 
                  number of elements for the image data type (bytes, 
                  2-bytes, integers or 'squashed i2' as set by ITYPE)
IAX1_ORIG  i (R)  The origin pixel of the required line along the 
                  first local axis
IAX2_ORIG  i (R)  The origin pixel of the required line along the 
                  second local axis
NAX1       i (R)  Size of the line in pixels along the first local axis
NAX1_RASTS i (R)  No. of rasters in the image along the first 
                  local axis    
NAX2_RASTS i (R)  No. of rasters in the image along the second 
                  local axis
IOUTLIM    i (R)  Value to which returned data pixels are to be set if
                  they lie outside the image
IDATA()    i (W)  Array returning the image line data
                  (NAX1 values)
IERR       i (W)  Error flag = 0 OK,
                             =-1 Invalid type or order flag
                             = 1 Line partly outside image
                             = 2 Line totally outside image

                 Note: pixels within the image are numbered from 1 upwards
-end*/
/*-C:*/
int img_ax1line (int * img, int itype, int iorder, int nf_off, 
                 int iax1_orig, int iax2_orig, int nax1,
                 int nax1_rasts, int nax2_rasts, int ioutlim,
                 int *idata)
/*end*/

/*-Parameters:
int * img;     Image data array (may hold signed integer data or
               unsigned byte data or unsigned two byte data or 'squashed' 
               two byte data - data will be cast to an appropriate type (R)
int itype;     Image data type 
               1 = unsigned byte data
               2 = unsigned two-byte data (i2)
               3 = signed integer data
               4 = 'squashed i2' data (if stored I<0
                   ipix = -i*8
               5 = 'squashed i2' data if stored I<0
                   ipix = (I+32768)*8
               6 = 'squashed i2' data if stored I<0
                   ipix = (I+32768)*32
               7 = 'squashed i2' data if stored I<0
                   ipix = -(I+1)*256 + 32768 
               (R)
int iorder;    Order of data in image wrt. the two local axes
               (slow then fast)
               =1 +ax1 +ax2
               =2 +ax1 -ax2
               =3 -ax1 +ax2
               =4 -ax1 -ax2
               =5 +ax2 +ax1
               =6 +ax2 -ax1
               =7 -ax2 +ax1
               =8 -ax2 -ax1
               (R)
int nf_off;    Fast offset between (slow) rasters of image (e.g. 
               first dimension of the image data if the data are 
               held in a two dimensional array). This value is the 
               number of elements for the image data type (bytes, 
               2-bytes, integers or 'squashed i2' as set by itype) (R)
int iax1_orig; The origin pixel of the required line along the 
               first local axis (R)
int iax2_orig; The origin pixel of the required line along the 
               second local axis (R)
int nax1;      Size of the line in pixels along the first local
               axis (R)
int nax1_rasts No. of rasters in the image along the first 
               local axis (R)    
int nax2_rasts No. of rasters in the image along the second 
               local axis (R)
int ioutlim;   Value to which returned data pixels are to be set if
               they lie outside the image (R)
int * idata;   Array returning the image line (nax1 values) (W)

               Note: input pixel numbers are numbered from 1 upwards

Return:  Error flag = 0 OK,
                    =-1 Invalid type or order flag
                    = 1 Line partly outside image
                    = 2 Line totally outside image

-end*/
{
   int islow_min;            /* Minimum pixel along image slow axis */
   int islow_max;            /* Maximum pixel along image slow axis */
   int ifast_min;            /* Minimum pixel along image fast axis */
   int ifast_max;            /* Maximum pixel along image fast axis */
   int ns_rasts;             /* No. of rasters in image along slow 
                                direction */
   int nf_rasts;             /* No. of rasters in image along fast 
                                direction */
   int iax1;                 /* Loop through local axis 1 */
   int iax1_start;           /* Start of loop through local axis 1 */
   int iax1_end;             /* End of loop through local axis 1 */
   int iax1_step;            /* Loop step for local axis 1 */
   int ks;                   /* Counter for slow axis */
   int kf;                   /* Counter for fast axis */
   int koff;                 /* Offset to start of current fast axis 
                                raster */
   int ktype;                /* Data type flag */
   int err;                  /* Error flag */
   long ipix;                /* Pixel value */ 
   Intpix *int_data;         /* Pointer to signed integer data */
   I2uns *i2_data;           /* Pointer to image plate 'i2' data */
   I2squash *i2_squash;      /* Pointer to 'squashed i2' data */
   unsigned char *byte_data; /* Pointer to byte image data*/
   long (*unsq)();           /* Function to unpack a squashed
                                image pixel value */

   err = 0; 

/* Check for valid flag values */

   if (itype<1||itype>7) return -1;
   if (iorder<1||iorder>8) return -1;

/* Get pointer to image data */

   unsq = 0;
   ktype = itype;
   if (ktype>4) ktype = 4;
   switch (ktype)
   {
      case 1:
      byte_data = (unsigned char *) img;
      break;

      case 2:
      i2_data = (I2uns *) img;
      break;

      case 3:
      int_data = (Intpix *) img;
      break;

      case 4:
      i2_squash = (I2squash *) img;
      switch (itype)
      {
         case 4:
         unsq = img_desq;
         break;

         case 5:
         unsq = img_desq2;
         break;

         case 6:
         unsq = img_desq3;
         break;

         case 7:
         unsq = img_desq4;
         break;
      }
      break;
   }

/* Get Box limits in image array and loop parameters for returning
   data array */

   switch(iorder)
   {
      case 1:
         islow_min = iax1_orig;
         ifast_min = iax2_orig;
         iax1_start = 1;
         iax1_end = nax1;
         iax1_step = 1;
      break;

      case 2:
         islow_min = iax1_orig;
         ifast_min = nax2_rasts - iax2_orig + 1;
         iax1_start = 1;
         iax1_end = nax1;
         iax1_step = 1;
      break;

      case 3:
         islow_min = nax1_rasts - iax1_orig - nax1 + 2;
         ifast_min = iax2_orig;
         iax1_start = nax1;
         iax1_end = 1;
         iax1_step = -1;
      break;

      case 4:
         islow_min = nax1_rasts - iax1_orig - nax1 + 2;
         ifast_min = nax2_rasts - iax2_orig + 1;
         iax1_start = nax1;
         iax1_end = 1;
         iax1_step = -1;
      break;

      case 5:
         islow_min = iax2_orig;
         ifast_min = iax1_orig;
         iax1_start = 1;
         iax1_end = nax1;
         iax1_step = 1;
      break;

      case 6:
         islow_min = iax2_orig;
         ifast_min = nax1_rasts - iax1_orig - nax1 + 2;
         iax1_start = nax1;
         iax1_end = 1;
         iax1_step = -1;
      break;

      case 7:
         islow_min = nax2_rasts - iax2_orig + 1;
         ifast_min = iax1_orig;
         iax1_start = 1;
         iax1_end = nax1;
         iax1_step = 1;
      break;

      case 8:
         islow_min = nax2_rasts - iax2_orig + 1;
         ifast_min = nax1_rasts - iax1_orig - nax1 + 2;
         iax1_start = nax1;
         iax1_end = 1;
         iax1_step = -1;
      break;
   }
   if (iorder<5)
   {
      islow_max = islow_min + nax1 - 1;
      ifast_max = ifast_min;
      ns_rasts = nax1_rasts;
      nf_rasts = nax2_rasts;
   }
   else
   {
      islow_max = islow_min;
      ifast_max = ifast_min + nax1 - 1;
      ns_rasts = nax2_rasts;
      nf_rasts = nax1_rasts;
   }

/* If some or all of requested section lies outside image then
   initialise returned data array to the given out of limits value */

   if (islow_min<1||islow_max>ns_rasts||ifast_min<1||ifast_max>nf_rasts)
   {
      err = 1;
      for (iax1=0;iax1<nax1;++iax1)
      {
         idata[iax1] = ioutlim;
      }
      if (islow_min>ns_rasts||islow_max<1) return 2;
      if (ifast_min>nf_rasts||ifast_max<1) return 2;
   }

/* Get image data */

   if (iorder<5)
   {
      ks = islow_min - 1; 
      koff = (ks-1)*nf_off;
      for (iax1=iax1_start;iax1<=iax1_end;iax1+=iax1_step)
      {
         ks++;
         koff = koff + nf_off;
         if (ks<1||ks>ns_rasts) continue;
         kf = ifast_min;
         if (kf<1||kf>nf_rasts) continue;
         switch (ktype)
         {
            case 1:
            ipix = (long) byte_data[koff+kf-1];
            break;

            case 2:
            ipix = (long) i2_data[koff+kf-1];
            break;

            case 3:
            ipix = (long) int_data[koff+kf-1];
            break;

            case 4:
            ipix = (long) unsq(i2_squash[koff+kf-1]);
            break;
         }
         idata[iax1-1] = ipix;
      }
   }
   else
   {
      ks = islow_min; 
      koff = (ks-1)*nf_off;
      if (ks<1||ks>ns_rasts) return err;
      kf = ifast_min - 1;
      for (iax1=iax1_start;iax1<=iax1_end;iax1+=iax1_step)
      {
         kf++;
         if (kf<1||kf>nf_rasts) continue;
         switch (ktype)
         {
            case 1:
            ipix = (long) byte_data[koff+kf-1];
            break;

            case 2:
            ipix = (long) i2_data[koff+kf-1];
            break;

            case 3:
            ipix = (long) int_data[koff+kf-1];
            break;

            case 4:
            ipix = (long) unsq(i2_squash[koff+kf-1]);
            break;
         }
         idata[iax1-1] = ipix;
      }
   }
   return err;
}
/* Fortran binding: imgf_ax1line */

void imgf_ax1line (int * img, int * itype, int * iorder, int * nf_off, 
                   int * iax1_orig, int * iax2_orig, int * nax1,
                   int * nax1_rasts, int * nax2_rasts, int * ioutlim,
                   int *idata, int *ierr)
{
   *ierr = img_ax1line (img, *itype, *iorder, *nf_off, 
                        *iax1_orig, *iax2_orig, *nax1,
                        *nax1_rasts, *nax2_rasts, *ioutlim, idata);
   return;
} 

/*-Routine: Get second local axis line from image - img_ax2line
The routine img_ax2line (imgf_ax2line) extracts an image line along
the direction of the second local axis from the in-memory data returning 
it in an integer array.
-end*/

/*                                            
***********************
**   img_ax2line     **
***********************

Purpose: Get a line (along local axis 2) from an in-memory image

Author:  John W. Campbell, July 1997

CD-Fortran:
      SUBROUTINE IMG_AX2LINE (IMG, ITYPE, IORDER, NF_OFF, 
     +                        IAX1_ORIG, IAX2_ORIG, NAX2,
     +                        NAX1_RASTS, NAX2_RASTS, IOUTLIM,
     +                        IDATA, IERR)
CD-end
*/
/*-Parameters:
IMG()      i (R)  Image data array (may hold signed integer data or
                  unsigned byte data packed into an integer array or
                  unsigned two byte data packed into an integer array
                  or 'squashed' two byte data packed into an integer
                  array.
ITYPE      i (R)  Image data type 
                  1 = unsigned byte data
                  2 = unsigned two-byte data (i2)
                  3 = signed integer data
                  4 = 'squashed i2' data (if stored I<0
                      ipix = -i*8
                  5 = 'squashed i2' data if stored I<0
                      ipix = (I+32768)*8
                  6 = 'squashed i2' data if stored I<0
                      ipix = (I+32768)*32
                  7 = 'squashed i2' data if stored I<0
                      ipix = -(I+1)*256 + 32768
IORDER     i (R)  Order of data in image wrt. the two local axes
                  (slow then fast)
                  =1 +ax1 +ax2
                  =2 +ax1 -ax2
                  =3 -ax1 +ax2
                  =4 -ax1 -ax2
                  =5 +ax2 +ax1
                  =6 +ax2 -ax1
                  =7 -ax2 +ax1
                  =8 -ax2 -ax1
NF_OFF     i (R)  Fast offset between (slow) rasters of image (e.g. 
                  first dimension of the image data if the data are 
                  held in a two dimensional array). This value is the 
                  number of elements for the image data type (bytes, 
                  2-bytes, integers or 'squashed i2' as set by ITYPE)
IAX1_ORIG  i (R)  The origin pixel of the required line along the 
                  first local axis
IAX2_ORIG  i (R)  The origin pixel of the required line along the 
                  second local axis
NAX2       i (R)  Size of the line in pixels along the second local
                  axis
NAX1_RASTS i (R)  No. of rasters in the image along the first 
                  local axis    
NAX2_RASTS i (R)  No. of rasters in the image along the second 
                  local axis
IOUTLIM    i (R)  Value to which returned data pixels are to be set if
                  they lie outside the image
IDATA()    i (W)  Array returning the image line data 
                  (NAX2 values)
IERR       i (W)  Error flag = 0 OK,
                             =-1 Invalid type or order flag
                             = 1 Box partly outside image
                             = 2 Box totally outside image

                 Note: pixels within the image are numbered from 1 upwards
-end*/
/*-C:*/
int img_ax2line (int * img, int itype, int iorder, int nf_off, 
                 int iax1_orig, int iax2_orig, int nax2,
                 int nax1_rasts, int nax2_rasts, int ioutlim,
                 int *idata)
/*end*/

/*-Parameters:
int * img;     Image data array (may hold signed integer data or
               unsigned byte data or unsigned two byte data or 'squashed' 
               two byte data - data will be cast to an appropriate type (R)
int itype;     Image data type 
               1 = unsigned byte data
               2 = unsigned two-byte data (i2)
               3 = signed integer data
               4 = 'squashed i2' data (if stored I<0
                   ipix = -i*8
               5 = 'squashed i2' data if stored I<0
                   ipix = (I+32768)*8
               6 = 'squashed i2' data if stored I<0
                   ipix = (I+32768)*32
               7 = 'squashed i2' data if stored I<0
                   ipix = -(I+1)*256 + 32768 
               (R)
int iorder;    Order of data in image wrt. the two local axes
               (slow then fast)
               =1 +ax1 +ax2
               =2 +ax1 -ax2
               =3 -ax1 +ax2
               =4 -ax1 -ax2
               =5 +ax2 +ax1
               =6 +ax2 -ax1
               =7 -ax2 +ax1
               =8 -ax2 -ax1
               (R)
int nf_off;    Fast offset between (slow) rasters of image (e.g. 
               first dimension of the image data if the data are 
               held in a two dimensional array). This value is the 
               number of elements for the image data type (bytes, 
               2-bytes, integers or 'squashed i2' as set by itype) (R)
int iax1_orig; The origin pixel of the required line along the 
               first local axis (R)
int iax2_orig; The origin pixel of the required line along the 
               second local axis (R)
int nax2;      Size of the line in pixels along the second local
               axis (R)
int nax1_rasts No. of rasters in the image along the first 
               local axis (R)    
int nax2_rasts No. of rasters in the image along the second 
               local axis (R)
int ioutlim;   Value to which returned data pixels are to be set if
               they lie outside the image (R)
int * idata;   Array returning the image line (nax2 values) (W)

               Note: input pixel numbers are numbered from 1 upwards

Return:  Error flag = 0 OK,
                    =-1 Invalid type or order flag
                    = 1 Line partly outside image
                    = 2 Line totally outside image

-end*/
{
   int islow_min;            /* Minimum pixel along image slow axis */
   int islow_max;            /* Maximum pixel along image slow axis */
   int ifast_min;            /* Minimum pixel along image fast axis */
   int ifast_max;            /* Maximum pixel along image fast axis */
   int ns_rasts;             /* No. of rasters in image along slow 
                                direction */
   int nf_rasts;             /* No. of rasters in image along fast 
                                direction */
   int iax2;                 /* Loop through local axis 2 */
   int iax2_start;           /* Start of loop through local axis 2 */
   int iax2_end;             /* End of loop through local axis 2 */
   int iax2_step;            /* Loop step through local axis 2 */
   int ks;                   /* Counter for slow axis */
   int kf;                   /* Counter for fast axis */
   int koff;                 /* Offset to start of current fast axis 
                                raster */
   int ktype;                /* Data type flag */
   int err;                  /* Error flag */
   long ipix;                /* Pixel value */ 
   Intpix *int_data;         /* Pointer to signed integer data */
   I2uns *i2_data;           /* Pointer to image plate 'i2' data */
   I2squash *i2_squash;      /* Pointer to 'squashed i2' data */
   unsigned char *byte_data; /* Pointer to byte image data*/
   long (*unsq)();           /* Function to unpack a squashed
                                image pixel value */

   err = 0; 

/* Check for valid flag values */

   if (itype<1||itype>7) return -1;
   if (iorder<1||iorder>8) return -1;

/* Get pointer to image data */

   unsq = 0;
   ktype = itype;
   if (ktype>4) ktype = 4;
   switch (ktype)
   {
      case 1:
      byte_data = (unsigned char *) img;
      break;

      case 2:
      i2_data = (I2uns *) img;
      break;

      case 3:
      int_data = (Intpix *) img;
      break;

      case 4:
      i2_squash = (I2squash *) img;
      switch (itype)
      {
         case 4:
         unsq = img_desq;
         break;

         case 5:
         unsq = img_desq2;
         break;

         case 6:
         unsq = img_desq3;
         break;

         case 7:
         unsq = img_desq4;
         break;
      }
      break;
   }

/* Get Box limits in image array and loop parameters for returning
   data array */

   switch(iorder)
   {
      case 1:
         islow_min = iax1_orig;
         ifast_min = iax2_orig;
         iax2_start = 1;
         iax2_end = nax2;
         iax2_step = 1;
      break;

      case 2:
         islow_min = iax1_orig;
         ifast_min = nax2_rasts - iax2_orig - nax2 + 2;
         iax2_start = nax2;
         iax2_end = 1;
         iax2_step = -1;
      break;

      case 3:
         islow_min = nax1_rasts - iax1_orig + 1;
         ifast_min = iax2_orig;
         iax2_start = 1;
         iax2_end = nax2;
         iax2_step = 1;
      break;

      case 4:
         islow_min = nax1_rasts - iax1_orig + 1;
         ifast_min = nax2_rasts - iax2_orig - nax2 + 2;
         iax2_start = nax2;
         iax2_end = 1;
         iax2_step = -1;
      break;

      case 5:
         islow_min = iax2_orig;
         ifast_min = iax1_orig;
         iax2_start = 1;
         iax2_end = nax2;
         iax2_step = 1;
      break;

      case 6:
         islow_min = iax2_orig;
         ifast_min = nax1_rasts - iax1_orig + 1;
         iax2_start = 1;
         iax2_end = nax2;
         iax2_step = 1;
      break;

      case 7:
         islow_min = nax2_rasts - iax2_orig - nax2 + 2;
         ifast_min = iax1_orig;
         iax2_start = nax2;
         iax2_end = 1;
         iax2_step = -1;
      break;

      case 8:
         islow_min = nax2_rasts - iax2_orig - nax2 + 2;
         ifast_min = nax1_rasts - iax1_orig + 1;
         iax2_start = nax2;
         iax2_end = 1;
         iax2_step = -1;
      break;
   }
   if (iorder<5)
   {
      islow_max = islow_min;
      ifast_max = ifast_min + nax2 - 1;
      ns_rasts = nax1_rasts;
      nf_rasts = nax2_rasts;
   }
   else
   {
      islow_max = islow_min + nax2 - 1;
      ifast_max = ifast_min;
      ns_rasts = nax2_rasts;
      nf_rasts = nax1_rasts;
   }

/* If some or all of requested section lies outside image then
   initialise returned data array to the given out of limits value */

   if (islow_min<1||islow_max>ns_rasts||ifast_min<1||ifast_max>nf_rasts)
   {
      err = 1;
      for (iax2=0;iax2<nax2;++iax2)
      {
         idata[iax2] = ioutlim;
      }
      if (islow_min>ns_rasts||islow_max<1) return 2;
      if (ifast_min>nf_rasts||ifast_max<1) return 2;
   }

/* Get image data */

   if (iorder<5)
   {
      ks = islow_min; 
      koff = (ks-1)*nf_off;
      if (ks<1||ks>ns_rasts) return err;
      kf = ifast_min - 1;
      for (iax2=iax2_start;iax2<=iax2_end;iax2+=iax2_step)
      {
         kf++;
         if (kf<1||kf>nf_rasts) continue;
         switch (ktype)
         {
            case 1:
            ipix = (long) byte_data[koff+kf-1];
            break;

            case 2:
            ipix = (long) i2_data[koff+kf-1];
            break;

            case 3:
            ipix = (long) int_data[koff+kf-1];
            break;

            case 4:
            ipix = (long) unsq(i2_squash[koff+kf-1]);
            break;
         }
         idata[iax2-1] = ipix;
      }
   }
   else
   {
      ks = islow_min - 1; 
      koff = (ks-1)*nf_off;
      for (iax2=iax2_start;iax2<=iax2_end;iax2+=iax2_step)
      {
         ks++;
         koff = koff + nf_off;
         if (ks<1||ks>ns_rasts) continue;
         kf = ifast_min;
         if (kf<1||kf>nf_rasts) continue;
         switch (ktype)
         {
            case 1:
            ipix = (long) byte_data[koff+kf-1];
            break;

            case 2:
            ipix = (long) i2_data[koff+kf-1];
            break;

            case 3:
            ipix = (long) int_data[koff+kf-1];
            break;

            case 4:
            ipix = (long) unsq(i2_squash[koff+kf-1]);
            break;
         }
         idata[iax2-1] = ipix;
      }
   }
   return err;
}
/* Fortran binding: imgf_ax2line */

void imgf_ax2line (int * img, int * itype, int * iorder, int * nf_off, 
                   int * iax1_orig, int * iax2_orig, int * nax2,
                   int * nax1_rasts, int * nax2_rasts, int * ioutlim,
                   int *idata, int *ierr)
{
   *ierr = img_ax2line (img, *itype, *iorder, *nf_off, 
                        *iax1_orig, *iax2_orig, *nax2,
                        *nax1_rasts, *nax2_rasts, *ioutlim, idata);
   return;
} 

/*-Routine: Get pixel from image - img_getpix
The routine img_getpix (imgf_getpix) extracts a pixel
from the in-memory data returning its value in an integer variable.
-end*/

/*                                            
***********************
**   img_getpix      **
***********************

Purpose: Get a pixel from an in-memory image

Author:  John W. Campbell, July 1997

CD-Fortran:
      SUBROUTINE IMG_GETPIX (IMG, ITYPE, IORDER, NF_OFF, 
     +                       IAX1_PIX, IAX2_PIX,
     +                       NAX1_RASTS, NAX2_RASTS, IOUTLIM,
     +                       IDATA, IERR)
CD-end
*/
/*-Parameters:
IMG()      i (R)  Image data array (may hold signed integer data or
                  unsigned byte data packed into an integer array or
                  unsigned two byte data packed into an integer array
                  or 'squashed' two byte data packed into an integer
                  array.
ITYPE      i (R)  Image data type 
                  1 = unsigned byte data
                  2 = unsigned two-byte data (i2)
                  3 = signed integer data
                  4 = 'squashed i2' data (if stored I<0
                      ipix = -i*8
                  5 = 'squashed i2' data if stored I<0
                      ipix = (I+32768)*8
                  6 = 'squashed i2' data if stored I<0
                      ipix = (I+32768)*32
                  7 = 'squashed i2' data if stored I<0
                      ipix = -(I+1)*256 + 32768
IORDER     i (R)  Order of data in image wrt. the two local axes
                  (slow then fast)
                  =1 +ax1 +ax2
                  =2 +ax1 -ax2
                  =3 -ax1 +ax2
                  =4 -ax1 -ax2
                  =5 +ax2 +ax1
                  =6 +ax2 -ax1
                  =7 -ax2 +ax1
                  =8 -ax2 -ax1
NF_OFF     i (R)  Fast offset between (slow) rasters of image (e.g. 
                  first dimension of the image data if the data are 
                  held in a two dimensional array). This value is the 
                  number of elements for the image data type (bytes, 
                  2-bytes, integers or 'squashed i2' as set by ITYPE)
IAX1_PIX   i (R)  The pixel position of the required pixel along the 
                  first local axis
IAX2_PIX   i (R)  The pixel position of the required pixel along the 
                  second local axis
NAX1_RASTS i (R)  No. of rasters in the image along the first 
                  local axis    
NAX2_RASTS i (R)  No. of rasters in the image along the second 
                  local axis
IOUTLIM    i (R)  Value to which returned data pixels are to be set if
                  they lie outside the image
IDATA      i (W)  The returned pixel value
IERR       i (W)  Error flag = 0 OK,
                             =-1 Invalid type or order flag
                             = 2 Pixel outside image

                 Note: pixels within the image are numbered from 1 upwards
-end*/
/*-C:*/
int img_getpix (int * img, int itype, int iorder, int nf_off, 
                int iax1_pix, int iax2_pix,
                int nax1_rasts, int nax2_rasts, int ioutlim,
                int *idata)
/*end*/

/*-Parameters:
int * img;     Image data array (may hold signed integer data or
               unsigned byte data or unsigned two byte data or 'squashed' 
               two byte data - data will be cast to an appropriate type (R)
int itype;     Image data type 
               1 = unsigned byte data
               2 = unsigned two-byte data (i2)
               3 = signed integer data
               4 = 'squashed i2' data (if stored I<0
                   ipix = -i*8
               5 = 'squashed i2' data if stored I<0
                   ipix = (I+32768)*8
               6 = 'squashed i2' data if stored I<0
                   ipix = (I+32768)*32
               7 = 'squashed i2' data if stored I<0
                   ipix = -(I+1)*256 + 32768 
               (R)
int iorder;    Order of data in image wrt. the two local axes
               (slow then fast)
               =1 +ax1 +ax2
               =2 +ax1 -ax2
               =3 -ax1 +ax2
               =4 -ax1 -ax2
               =5 +ax2 +ax1
               =6 +ax2 -ax1
               =7 -ax2 +ax1
               =8 -ax2 -ax1
               (R)
int nf_off;    Fast offset between (slow) rasters of image (e.g. 
               first dimension of the image data if the data are 
               held in a two dimensional array). This value is the 
               number of elements for the image data type (bytes, 
               2-bytes, integers or 'squashed i2' as set by itype) (R)
int iax1_pix;  The pixel position of the required pixel along the 
               first local axis (R)
int iax2_pix ; The pixel position of the required pixel along the 
               second local axis (R)
int nax1_rasts No. of rasters in the image along the first 
               local axis (R)    
int nax2_rasts No. of rasters in the image along the second 
               local axis (R)
int ioutlim;   Value to which returned data pixels are to be set if
               they lie outside the image (R)
int * idata;   The returned pixel value (W)

               Note: input pixel numbers are numbered from 1 upwards

Return:  Error flag = 0 OK,
                    =-1 Invalid type or order flag
                    = 2 Pixel outside image

-end*/
{
   int ns_ras;               /* No. of slow rasters */
   int nf_ras;               /* No. of fast rasters */
   int axs_pix;              /* Pixel position along slow axis */
   int axf_pix;              /* Pixel position along fast axis */
   int ktype;                /* Data type flag */
   int ioff;                 /* Offset into image */
   Intpix *int_data;         /* Pointer to signed integer data */
   I2uns *i2_data;           /* Pointer to image plate 'i2' data */
   I2squash *i2_squash;      /* Pointer to 'squashed i2' data */
   unsigned char *byte_data; /* Pointer to byte image data*/
   long (*unsq)();           /* Function to unpack a squashed
                                image pixel value */

   *idata = ioutlim;

/* Check for valid flag values */

   if (itype<1||itype>7) return -1;
   if (iorder<1||iorder>8) return -1;

/* Get pixel position */

   if (iorder<5)
   {
      ns_ras = nax1_rasts;
      nf_ras = nax2_rasts;
   }
   else
   {
      ns_ras = nax2_rasts;
      nf_ras = nax1_rasts;
   }

   switch (iorder) 
   {
       case 1:  /* +ax1, +ax2 */
          axs_pix = iax1_pix;
          axf_pix = iax2_pix;
       break;

       case 2:  /* +ax1, -ax2 */
          axs_pix = iax1_pix;
          axf_pix = nf_ras - iax2_pix  + 1;
       break;         

       case 3:  /* -ax1, +ax2 */
          axs_pix = ns_ras - iax1_pix + 1;
          axf_pix = iax2_pix;
       break;

       case 4:  /* -ax1, -ax2 */
          axs_pix = ns_ras - iax1_pix + 1;
          axf_pix = nf_ras - iax2_pix + 1;
       break;

       case 5:  /* +ax2, +ax1 */
          axs_pix = iax2_pix;
          axf_pix = iax1_pix;
       break;

       case 6:  /* +ax2, -ax1 */
          axs_pix = iax2_pix;
          axf_pix = nf_ras - iax1_pix + 1;
       break;         

       case 7:  /* -ax2, +ax1 */
          axs_pix = ns_ras - iax2_pix + 1;
          axf_pix = iax1_pix;
       break;

       case 8:  /* -ax2, -ax1 */
          axs_pix = ns_ras - iax2_pix + 1;
          axf_pix = nf_ras - iax1_pix + 1;
       break;
   }
   if (axs_pix<1||axs_pix>ns_ras) return 2;
   if (axf_pix<1||axf_pix>nf_ras) return 2;

   ioff = (axs_pix-1)*nf_off + axf_pix - 1;

   unsq = 0;
   ktype = itype;
   if (ktype>4) ktype = 4;
   switch (ktype)
   {
      case 1:
      byte_data = (unsigned char *) img;
      *idata = (int) byte_data[ioff];
      break;

      case 2:
      i2_data = (I2uns *) img;
      *idata = (int) i2_data[ioff];
      break;

      case 3:
      int_data = (Intpix *) img;
      *idata = (int) int_data[ioff];
      break;

      case 4:
      i2_squash = (I2squash *) img;
      switch (itype)
      {
         case 4:
         unsq = img_desq;
         break;

         case 5:
         unsq = img_desq2;
         break;

         case 6:
         unsq = img_desq3;
         break;

         case 7:
         unsq = img_desq4;
         break;
      }
      *idata = (int) unsq(i2_squash[ioff]);
      break;
   }
   return 0;
} 
/* Fortran binding: imgf_getpix */

void imgf_getpix (int * img, int * itype, int * iorder, int * nf_off, 
                  int * iax1_pix, int * iax2_pix,
                  int * nax1_rasts, int * nax2_rasts, int * ioutlim,
                  int *idata, int *ierr)
{
   *ierr = img_getpix (img, *itype, *iorder, *nf_off, 
                        *iax1_pix, *iax2_pix,
                        *nax1_rasts, *nax2_rasts, *ioutlim, idata);
   return;
} 
/*-Section: File Name from Template
This is a routine for creating an image file name from a template name.
-end*/

/*-Routine: Get file name from template - img_filename
The routine img_filename (imgf_filename) returns the file name for an image
given a directory string, a template string and an image number.
-end*/ 
/*
***************************
**     img_filename      **
***************************

Purpose: Form an image file name from directory, template and image no.

Author:  John W. Campbell, November 1997

*/
/*-Fortran:
         CALL IMGF_FILENAME (IMGSTR(DIRECTORY), LEND, IMGSTR(TEMPLATE),
        +                    LENT, IMGNUM, IMGSTR(FILNAM), MAXF, IERR)
-end*/
 
/*-Parameters:
DIRECTORY   c (R) The directory string
                  ** Pass address using the IMGSTR routine **
LEND        i (R) Length of the directory string
TEMPLATE    c (R) Image file name template string
                  ** Pass address using the IMGSTR routine **
LENT        i (R) Length of the template string
IMGNUM      i (R) The image number
FILNAM      c (W) Returns the file name string
                  ** Pass address using the IMGSTR routine **
MAXF        i (R) Maximum length for the file name string
IERR        i (W) Error return flag from img_filename
-end*/

/*-C:*/
int img_filename (char * directory, int lend, char * template, int lent,
                  int imgnum, char * filnam, int maxf)
/*end*/
 
/*-Parameters:
char * directory;   Directory string (may include trailing '/' (R)
int lend;           Length of the directory string - may be 0 if
                    a null terminated string passed (R)
char * template;    File name template - hash field will be replaced
                    by image number (filled on the left with zeroes if
                    needed) (R)
int lent;           Length of the template string - may be 0 if
                    a null terminated string passed (R)
int imgnum;         The image number (R) 
char * filnam;      String in which to return the file name (W) 
int maxf;           if >0 Maximum length of file name string allowed
                          (excluding terminating null - filnam must be at
                          least maxf + 1 characters in length)
                    if <0 abs(maxf) characters will be returned padded
                          with blanks if needed (for 'fortran' use) (R)

Return:  =0   OK
         -1   blank template
         -2   cannot allocate temporary memory for the routine
-end*/
{
   char *indx1;
   char *indx2;
   int n;
   char * templ_str;
   int use_dir;
   int l;
   char format[50];
   char str[20];
   char *tmpstr;
   char * tmpfil;

   img_copy_chars (" ", 0, filnam, maxf);

   if (img_blank_string (template, lent)) return -1;
   if (lent==0) lent = strlen(template);
   tmpstr = (char *) malloc (lent+1);
   if (tmpstr==0) return -2;
   tmpstr[0] = 0;
   strncat (tmpstr,template,lent);
   templ_str = (char *) malloc (lent+1);
   if (templ_str==0) return -2;
   use_dir = 1;
   sscanf (tmpstr,"%s", templ_str);
   if (templ_str[0]=='/'||templ_str[0]=='.') use_dir = 0;
   free(tmpstr);

   tmpfil = (char *) malloc (lend+lent+21);
   if (tmpfil==0) return -2;
   if (lend==0) lend = strlen(directory);
   tmpstr = (char *) malloc (lend+1);
   if (tmpstr==0) return -2;
   tmpstr[0] = 0;
   strncat (tmpstr,directory,lend);   
   if (use_dir&&!img_blank_string(directory, lend))
   {
      sscanf (tmpstr, "%s", tmpfil);
      l = strlen (tmpfil);
      if (tmpfil[l-1]!='/'&&tmpfil[l-1]!='\\'
          &&tmpfil[l-1]!=']'&&tmpfil[l-1]!=':') 
      {
         tmpfil[l]='/'; 
         tmpfil[l+1]=0;
      }
   }
   free (tmpstr);

   indx1 = strchr(templ_str, (int) '#');
   if (indx1==0)
   {
      strcat (tmpfil, templ_str);
      goto done;
   }
   indx2 = strrchr(templ_str, (int) '#');
   n = indx1 - templ_str;
   if (n>0) strncat (tmpfil, templ_str, n);
   n = indx2 - indx1 + 1;
   sprintf (format,"%%0%dd",n);
   sprintf (str, format, imgnum);
   strcat (tmpfil, str);
   strcat (tmpfil, indx2+1);
   done:
   img_copy_chars (tmpfil, 0, filnam, maxf);
   free (templ_str);
   free (tmpfil);
   return 0;
}
/* Fortran binding: imgf_filename */

void imgf_filename (int * directory, int * lend, int * template, int * lent,
                    int * imgnum, int * filnam, int * maxf, int * ierr)
{
   int maxfl;

   maxfl = -(*maxf);
   if (maxfl>0) maxfl = -maxfl;
   *ierr = img_filename (IMG_pointers[*directory], *lend,
                         IMG_pointers[*template], *lent,
                         *imgnum, IMG_pointers[*filnam], maxfl);
}   


/*-Section: Axis Order Decoding Routine
This is a routine for decoding the axis order codes used to indicate the
order of the data in an image file.
-end*/

/*-Routine: Interpret axis order string - img_axord
The routine img_axord (imgf_axord) is used to interpret a user supplied
string defining the axis order in an input image file. The string defines 
the axis order in terms of x and y (the film xf, yf axes). The slower 
moving axis is given first followed by the faster moving axis. Minus signs 
are used when the order of an axis is reversed. The code 's' if present 
indicates that byte swapping is to be done and the code 'n' indicates that no
byte swapping is to be done (byte swapping only relevant for image plate
data). Characters, other than 'x', 'y', '-', '+', 's' and 'n', or their
upper case equivalents, are ignored. A special code of 'unset' is
allowed. Some examples follow and the returned flags are indicated:
.
      unset                  order = 0, swaptyp = 0
      xy                     order = 1, swaptyp = 0
      +x+y                   order = 1, swaptyp = 0
      -xy                    order = 3, swaptyp = 0
      x-y s                  order = 2, swaptyp = 1
      -Y, +X, n              order = 7, swaptyp = 0
      S                      order = 0, swaptyp = 1
      +yf, +xf               order = 4, swaptyp = 0

Invalid strings: +x-x      (Same axis defined twice)
                 +x+y-x    (More than two axes defined)
                 xysn      (Byte swapping and no byte swapping both 
                            defined)
.
-end*/
/*
***************************
**     img_axord         **
***************************

Purpose: Interpret image axes order and byte swapping option string

Author:  John W. Campbell, November 1991

*/
/*-Fortran:
         CALL IMGF_AXORD (STR, LEN, IORD, ISWAP, IERR)
-end*/
 
/*-Parameters:
STR       c  (R)  Character string containing the axis order
                  (and byte swapping option) definition
                  (see str)
                  ** Pass address using the IMGSTR routine **
LEN       i  (R)  The length of the character string
IORD      i  (W)  Returns the axis order as a number from 1-8
                  (0 if 'unset') 
                  (see order)
ISWAP     i  (W)  Returns the byte swap flag =0 no byte swap (or 'unset')
                                             =1 byte swap data
IERR      i  (W)  Returns the status from the rdm_axord call
-end*/

/*-C:*/
int img_axord (char * str, int len, int * order, int * swaptyp)
/*end*/
 
/*-Parameters:
char *str;      String containing the axes order and byte swap flags (R)
                    
int len;        Length of the string; if 0 then assume a null
                terminated string was passed (R) 
int *order;     Returns the axis order given as a number from 1 to 8
                    (If 'unset' return 0)
                     1   +xf slow   +yf fast
                     2   +xf slow   -yf fast
                     3   -xf slow   +yf fast
                     4   -xf slow   -yf fast
                     5   +yf slow   +xf fast
                     6   +yf slow   -xf fast
                     7   -yf slow   +xf fast
                     8   -yf slow   -xf fast
                     (W)
int *swaptyp;   Returns byte swap type   =0 do not swap bytes (or 'unset')
                                         =1 swap bytes
                    (W)
Return:  Error flag =0 no error; 
                    >0 syntax error   =1 Only one axis defined.
                                      =2 Both axes the same.
                                      =3 More than two axes defined.
                                      =4 Byte swapping and no byte
                                         swapping defined.
-end*/
{
   int saxis, ssign;
   int faxis, fsign;
   int n;
   int l;
   char * p;
   char tmpstr[81];
   int err;

   /* Initialisations */

   if (len<=0) len=strlen(str);
   saxis = 0;
   ssign = 1;
   faxis = 0;
   fsign = 1;
   err = 0;
   *swaptyp = 0;
   *order = 0;
   if (len==0) return 0;
   tmpstr[0] = 0;
   l = len;
   if (l>80) l = 80;
   strncat (tmpstr,str,l);
   p = strtok (tmpstr," ");
   if (p!=0)
   {
      if (strcasecmp(p,"unset")==0) return 0;
   }
   n = 1;

   /* Decode the string */

   for (l=0;l<len;++l)
   {
      switch (str[l])
      {
         case '-':
            if (n==1) { ssign = -1; }else{ fsign = -1; }
            break;
         case '+':
            if (n==1) { ssign = 1; }else{ fsign = 1; }
            break;
         case 'x':
         case 'X':
            if (n==1) 
            {
               saxis = 1;
               n = 2;
            }
            else
            {
               if (n==3) { err=3; goto synterr; }
               faxis = 1;
               n = 3;
            }
            break;
         case 'y':
         case 'Y':
            if (n==1) 
            {
               saxis = 2;
               n = 2;
            }
            else
            {
               if (n==3) { err=3; goto synterr; }
               faxis = 2;
               n = 3;
            }
            break;
         case 's':
         case 'S':
            if (*swaptyp>0) { err=4; goto synterr; }
            *swaptyp = 2;
            break;
         case 'n':
         case 'N':
            if (*swaptyp>0) goto synterr;
            *swaptyp = 1;
            break;
      }
   }
   if (*swaptyp==0) *swaptyp = 1;
   *swaptyp -= 1;
   if (saxis==0&&faxis==0) return 0;
   if (saxis==0||faxis==0) { err=1; goto synterr; }
   if (saxis==faxis) { err=2; goto synterr; }
   *order = 4*saxis - 3;
   if (ssign<0) *order += 2;
   if (fsign<0) *order += 1;
   return 0;
   synterr:
   *order = 0;
   *swaptyp = 0;
   return err;
}
/* Fortran binding: imgf_axord */

void imgf_axord (str, len, iord, iswap, ierr)
int *str, *len, *iord, *iswap, *ierr;
{
   *ierr = img_axord (IMG_pointers[*str], *len, iord, iswap);
   return;
}

/*-Section: Background Image Calculation
This is a  routine for calculating a background image from an
original single crystal X-ray diffraction image.
-end*/

/*-Routine: Background image calculation - img_bgcalc
The routine img_bgcalc (imgf_bgcalc) is used to calculate
a background image from an image. This is for single 
crystal X-ray diffraction images (i.e. one with spots over a fairly slowly 
changing background). The method calculates the background at a pixel by taking
the average of the lowest pixel values in a box surrounding the pixel 
such that the number of such values is a requested percentage of the
total number of pixels in the box. To optimise the calculation, the
box is walked through the image moving a pixel at a time starting
at the top left. It then repeats the process of left from to right,
then one pixel down, then from right to left, and then one
pixel down etc. until the film image has been covered. (In this
description the slow moving image index is visualised as being from top to 
bottom and the fast one from left to right). The routine will cope with
a maximum background level up to around 65000. If desired, routines may be 
passed to it to monitor the progress of the background calculation and to 
cancel the calculation if required.
-end*/

/*
*************************
**   img_bgcalc        **
*************************

Purpose: Calculate a background image for an image using a 2-D search

Author:  John Campbell

*/
/*-Fortran:
         CALL IMGF_BGCALC (IMG, ITYPE, NSRASTS, NFRASTS, NFOFF, 
        +                   NCMP, NPBOX, IPCNT, NULL_PIXEL,
        +                   IMG_BG, NFBGOFF, 
        +                   PROGRESS, CANCEL, IERR)
-end*/
 
/*-Parameters:
IMG        i (R)   Array holding the image data 
                i  (normally packed into an integer array) 
                 i (see 'img')
ITYPE      i (R)   Image data type =1 unsigned byte, 
                                   =2 unsigned two-byte, 
                                   =3 signed integer 
                                   =4 'squashed i2' data (if stored I<0
                                      ipix = -i*8
                                   =5 'squashed i2' data if stored I<0
                                       ipix = (I+32768)*8
                                   =6 'squashed i2' data if stored I<0
                                      ipix = (I+32768)*32
                                   =7 'squashed i2' data if stored I<0
                                      ipix = -(I+1)*256 + 32768
                   (see 'type')
NSRASTS    i (R)   No. rasters along slow image axis 
                   (see 'ns_rasts')
NFRASTS    i (R)   No. rasters along fast image axis
                   (see 'nf_rasts')
NFOFF      i (R)   Offset (in pixels) between successive slow
                   rasters (see 'nf_off')
NCMP       i (R)   Compress (sample) NCMP into 1 pixel in each 
                   direction for background image (see ncmp)
NPBOX      i (R)   Dimension of background calculation box in
                   pixels of compressed imagee (preferably odd) 
                   (see 'npbox')
IPCNT      i (R)   Percentage of pixels to treat as background.
NULL_PIXEL i (R)   Value of a pixel to be ignored and treated as
                   null.
IMG_BG     i (W)   Array to hold the returned background image.
                   (pass an integer array large enough to hold 
                   the returned image data) (see 'img_bg')
NFBGOFF    i (R/W) Offset between start of lines in returned
                   background image (allows for padding if 
                   required). If zero on input then value will
                   be returned for no padding cases (see nf_bgoff)
PROGRESS   s (R)   Subroutine to monitor progress. Will be
                   called after each processing 'strip' 
                   as follows:

                   CALL PROGRESS (IFLAG, NSTRIP, ISTRIP)

                   IFLAG    (R)   =0 start, =1 in progress, 
                                  =2 done
                   NSTRIP   (R)   Total no. of strips of image to 
                                  process
                   ISTRIP   (R)   Current strip (1 to NSTRIP)

CANCEL     s (R)   This routine will be called at intervals to
                   see if a cancel state has been set.
                 
                   The call is:
               
                   CALL CANCEL (ICANCL)
                 
                   ICANCL   (W)  Return 0 to continue, 1 to cancel
                
IERR       i (W)   Returns the status from the img_bgcalc
                   call
-end*/

/*-C:*/
int img_bgcalc (int *img, int type, int ns_rasts, int nf_rasts, 
                int nf_off, int ncmp, int npbox, int percent, 
                int null_pixel, int *img_bg, int *nf_bgoff, 
                void (*progress)(), void (*cancel)())
/*end*/
 
/*-Parameters:
int *img;                Image from which to calculate the background 
                         image (R)
int type;                Image (and backgound image) data type
                              =1 unsigned byte
                              =2 unsigned two-byte data
                              =3 signed integer data
                              =4 'squashed i2' data (if stored I<0
                                 ipix = -i*8
                              =5 'squashed i2' data if stored I<0
                                  ipix = (I+32768)*8
                              =6 'squashed i2' data if stored I<0
                                 ipix = (I+32768)*32
                              =7 'squashed i2' data if stored I<0
                                 ipix = -(I+1)*256 + 32768
                         (R)
int ns_rasts;            No. of slow moving rasters in the image (R)
int nf_rasts;            No. of fast moving rasters in the image (R)
int nf_off;              Offset in the image (in pixels) between start
                         of each line of the image as stored (allows
                         for padding at the end of each line if 
                         used) (R)
int ncmp;                No. of pixels to compress into 1 in each
                         direction for the background image. The
                         size of the background image will be
                         ns_rasts/ncmp * nf_rasts/ncmp (integer
                         divisions) (R)
int npbox;               Size of box (after compression of the image)
                         for each background calculation
                         in pixels (box is npbox*npbox) - preferably
                         an odd number (R)
int percent;             Percentage of pixels to treat as background (R)
int null_pixel;          Value of a pixel to be ignored and treated as
                         null (R)
int *img_bg;             Returns background image in same data format
                         as the input image. The size of the 
                         background image will be
                         ns_rasts/ncmp * nf_rasts/ncmp (integer
                         divisions) (see also mf_bgoff) (W)
int *nf_bgoff;           Offset between start of image lines in
                         returned background image. If value
                         given is less than nf_rasts/ncmp then
                         it will be reset to nf_rasts/ncmp (R/W)
void (*progress)();      Function passed to the routine to monitor
                         the progress of the background calculation.
                         called as follows: 

                         progress (iflag, nstrip, istrip);

                         int *iflag;    =0 start, =1 in progress, 
                                        =2 done
                         int *nstrip;   Total no. of strips of image to 
                                        process
                         int *istrip;   Current strip (1 to nstrip)

                         (May be a NULL)
                         (R)
void (*cancel)();        Function passed to the routine to enable
                         cancellation. Called as follows:

                         cancel (icancl);

                         int *icancl;  Function must return 0 to 
                                       continue, or 1 to cancel.
                         (May be a NULL)
                         (R)                              
-end*/
/*-Doc:
Return:  =0 OK; =1 background too high (> 65000); =2 box too big for image
         -1 interrupted
-end*/
{

   int count[2048];   /* Array to hold intensity counts */
   int i,j,k,l;       /* Temp/loop variables */
   int offset;        /* Offset to center of background box (compressed) 
                         from its origin */
   int nbgmin;        /* The minimum number of background pixels required */
   int max_f;         /* The number of fast axis steps required */
   int max_s;         /* The number of slow axis steps required */
   int ip;            /* Index of pixel being removed in current step */
   int jp;            /* Index of pixel being added in current step */
   int ipx;           /* Index of current pixel for which background is being
                         found */
   int offset1;       /* Offset addition for sampled data */
   int offset2;       /* Pixel offset from top to bottom of box */
   int intens_tot;    /* Total intensity up to maximum level used in
                         background calculation */
   int ni;            /* No. of intensity values included in intens_tot */
   int nl;            /* Max intensity currently used in background
                         calculation (index into 'count') */
   enum step {down,up,right} nxt, last; /*Next and last step directions */
   int kf;            /* Step counter in fast direction */
   int ks;            /* Step counters in slow direction */
   int ii;           /* Temp intensity value */
   int jj;            /* Offset into count array for intensities>1023 */
   int iii;           /* Temp offset value */
   int nc;            /* Count of intensities in top level bin used for
                         background */
   int ne;            /* No. of extra intensities in intens_tot over those
                         required */
   int min_foff;      /* Minimum fast axis offset for which background 
                         pixel found */
   int max_foff;      /* Maximum fast axis offset for which background 
                         pixel found */
   int min_soff;      /* Minimum slow axis offset for which background 
                         pixel found */
   int max_soff;      /* Maximum slow axis offset for which background 
                         pixel found */
   int too_high;      /* Background level too high flag */
   int iflag;         /* Flag for progress routiine */
   int istrip;        /* Strip no. for progress routine */
   int icancl;        /* Cancel routine flag */
   int ktype;         /* Type 1-4, 4 includes all squashed i2 types */
   unsigned char *byte_data; /* Pointer to unsigned byte image data */
   I2uns *i2_data;    /* Pointer to unsigned two-byte image data */
   I2pix *i2_squash;  /* Pointer to 'squashed i2' image data */
   Intpix *int_data;  /* Pointer to signed integer image data */
   unsigned char *byte_bgdata; /* Pointer to unsigned byte background 
                                  image data */
   I2uns *i2_bgdata;   /* Pointer to unsigned two-byte background image data */
   I2pix *i2_bgsquash; /* Pointer to 'squashed i2' background image data */
   Intpix *int_bgdata; /* Pointer to signed integer background image data */
   int (*unsq)();     /* Pointer to 'un-squash' i2 routine */
   I2pix (*sqsh)();    /* Pointer to 'squash' i2 routine */


   too_high = 0;
   for(i=0;i<2048;++i) count[i]=0;
   nbgmin = (npbox*npbox*percent)/100;
   nf_rasts = ncmp*(nf_rasts/ncmp);
   ns_rasts = ncmp*(ns_rasts/ncmp);
   if (npbox>nf_rasts/ncmp||npbox>ns_rasts/ncmp) return 2;
   offset = npbox/2;
   max_f = nf_rasts/ncmp - npbox;
   max_s = ns_rasts/ncmp - npbox;
   min_foff = offset;
   max_foff = nf_rasts/ncmp + min_foff - npbox;
   min_soff = offset;
   max_soff = ns_rasts/ncmp + min_soff - npbox;
   offset1 = (ncmp/2)*nf_off + (ncmp/2);
   offset2 = npbox*ncmp*nf_off;
   int_data = (Intpix *) img;
   i2_data = (I2uns *) img;
   i2_squash = (I2pix *) img;
   byte_data = (unsigned char *) img;
   int_bgdata = (Intpix *) img_bg;
   i2_bgdata = (I2uns *) img_bg;
   i2_bgsquash = (I2pix *) img_bg;
   byte_bgdata = (unsigned char *) img_bg;
   if (*nf_bgoff<nf_rasts/ncmp) *nf_bgoff = nf_rasts/ncmp;

   unsq = 0;
   sqsh = 0;
   ktype = type;
   if (ktype>4) ktype = 4;
   if (ktype==4)
   {
      switch (type)
      {
         case 4:
           unsq = img_unsq2;
           sqsh = img_sqsh2;
         break;

         case 5:
           unsq = img_unsq3;
           sqsh = img_sqsh3;
         break;

         case 6:
           unsq = img_unsq4;
           sqsh = img_sqsh4;
         break;

         case 7:
           unsq = img_unsq5;
           sqsh = img_sqsh5;
         break;
      }
   }

   /* Initial call to progress monitor */

   iflag = 0;
   istrip = 1;
   if (progress!=0) (*progress)(&iflag, &max_f, &istrip);
   iflag = 1;

   /* Full scan of image */

   /* Process initial background box */
   k=offset1;
   for (i=0;i<npbox;i++)
   {
      j = k;
      for (l=0;l<npbox;l++) 
      {
         switch (ktype)
         {
            case 1:
               ii = byte_data[j]; j+=ncmp;
            break;

            case 2:
               ii = i2_data[j]; j+=ncmp;
            break;

            case 3:
               ii = int_data[j]; j+=ncmp;
            break;

            case 4:
               ii = unsq(i2_squash[j]); j+=ncmp;
            break;
         }
         if (ii==null_pixel) ii = 0;
         if (ii>1023)
         {
            jj = 1024 + ((ii-992)>>6);
            if (jj>2047) jj = 2047;
            count[jj]++;
         }
         else
         {
            count[ii]++;
         }
      }
      k += ncmp*nf_off;
   }
   nbgmin = ( npbox*npbox - count[0] ) * percent / 100;
   ni=0; intens_tot=0;
   for (i=1,nl=1;i<2048;++i,++nl)
   {
      ni += count[i];
      if (i>1023)
      {
         intens_tot += count[i]*(1024+((i-1024)<<6));
         if (i==2047) too_high = 1;
      }
      else
      {
         intens_tot += count[i]*i;
      }
      if (ni>=nbgmin) break;
   }
   ipx = offset*(*nf_bgoff) + offset;
   switch (ktype)
   {
      case 1:
      	byte_bgdata[ipx] = ni ? intens_tot/ni : 0;
      break;

      case 2:
        i2_bgdata[ipx] = ni ? intens_tot/ni : 0;
      break;

      case 3:
        int_bgdata[ipx] = ni ? intens_tot/ni : 0;
      break;

      case 4:
        i2_bgsquash[ipx] = ni ? sqsh(intens_tot/ni) : 0;
      break;
   }
   ipx += *nf_bgoff;

   /* Start walk through the image */
   
   kf = 0;
   ks = 1;
   ip = offset1;
   jp = offset1 + offset2;
   nxt=down;

   next_box:

   /* Remove intensities for left (or top or right) row of current box
      depending on the direction the box is being moved */

   /* Add intensities for right (or bottom or left) row of shifted box */

   if (nxt!=right)   
   {
      for (i=0;i<npbox;++i)
      {
         switch (ktype)
         {
            case 1:
               ii = byte_data[ip]; ip+=ncmp;
            break;

            case 2:
               ii = i2_data[ip]; ip+=ncmp;
            break;

            case 3:
               ii = int_data[ip]; ip+=ncmp;
            break;

            case 4:
               ii = unsq(i2_squash[ip]); ip+=ncmp;
            break;
         }
         if (ii==null_pixel) ii = 0;
         if (ii>1023)
         {
            iii = (ii-992)>>6;
            jj = 1024+iii;
            if (jj>2047) jj = 2047;
            count[jj]--;
            if (jj<=nl){intens_tot-=(1024+(iii<<6)); ni--;}
         }
         else
         {
            count[ii]--;
            if (ii<=nl) {intens_tot-=ii; ni--;}
            if (ii<=nl && !ii) {intens_tot+=ii; ni++;}
         }
         switch (ktype)
         {
            case 1:
               ii = byte_data[jp]; jp+=ncmp;
            break;

            case 2:
               ii = i2_data[jp]; jp+=ncmp;
            break;

            case 3:
               ii = int_data[jp]; jp+=ncmp;
            break;

            case 4:
               ii = unsq(i2_squash[jp]); jp+=ncmp;
            break;
         }
         if (ii==null_pixel) ii = 0;
         if (ii>1023)
         {
            iii = (ii-992)>>6;
            jj = 1024+iii;
            if (jj>2047) jj = 2047;
            count[jj]++;
            if (jj<=nl){intens_tot+=(1024+(iii<<6)); ni++;}
         }
         else
         {
            count[ii]++;
            if (ii<=nl) {intens_tot+=ii; ni++;}
            if (ii<=nl && !ii) {intens_tot-=ii; ni--;}
         }
      }
   }
   else
   {
      for (i=0;i<npbox;++i)
      {
         switch (ktype)
         {
            case 1:
               ii = byte_data[ip]; ip+=ncmp*nf_off;
            break;

            case 2:
               ii = i2_data[ip]; ip+=ncmp*nf_off;
            break;

            case 3:
               ii = int_data[ip]; ip+=ncmp*nf_off;
            break;

            case 4:
               ii = unsq(i2_squash[ip]); ip+=ncmp*nf_off;
            break;
         }
         if (ii==null_pixel) ii = 0;
         if (ii>1023)
         {
            iii = (ii-992)>>6;
            jj = 1024+iii;
            if (jj>2047) jj = 2047;
            count[jj]--;
            if (jj<=nl){intens_tot-=(1024+(iii<<6)); ni--;}
         }
         else
         {
            count[ii]--;
            if (ii<=nl) {intens_tot-=ii; ni--;}
            if (ii<=nl && !ii) {intens_tot+=ii; ni++;}
         }
         switch (ktype)
         {
            case 1:
               ii = byte_data[jp]; jp+=ncmp*nf_off;
            break;

            case 2:
               ii = i2_data[jp]; jp+=ncmp*nf_off;
            break;

            case 3:
               ii = int_data[jp]; jp+=ncmp*nf_off;
            break;

            case 4:
               ii = unsq(i2_squash[jp]); jp+=ncmp*nf_off;
            break;
         }
         if (ii==null_pixel) ii = 0;
         if (ii>1023)
         {
            iii = (ii-992)>>6;
            jj = 1024+iii;
            if (jj>2047) jj = 2047;
            count[jj]++;
            if (jj<=nl){intens_tot+=(1024+(iii<<6)); ni++;}
         }
         else
         {
            count[ii]++;
            if (ii<=nl) {intens_tot+=ii; ni++;}
            if (ii<=nl && !ii) {intens_tot-=ii; ni--;}
         }
      }
   }

   /* Move up or down 'count' array to find new background level */

   nbgmin = ( npbox*npbox - count[0] ) * percent / 100;
   if (ni>=nbgmin)
   {
      nc = count[nl];
      while (ni-nc>=nbgmin)
      {
         if (nl>1023)
         {
            intens_tot -= nc*(1024+((nl-1024)<<6));
         }
         else
         {
            intens_tot -= nc*nl;
         }
         ni -= nc;
         nc = count[--nl];
      }
   }
   else
   {
      while (ni<nbgmin)
      {
         nc = count[++nl];
         ni += nc;
         if (nl>1023)
         {
            intens_tot += nc*(1024+((nl-1024)<<6));
            if (nl==2047) too_high = 1;
         }
         else
         {
            intens_tot += nc*nl;
         }
      }
   }

   /* Set background pixel value */

   ne = ni - nbgmin;
   if (nl>1023)
   {
      switch (ktype)
      {
         case 1:
            byte_bgdata[ipx] = nbgmin ? (intens_tot-ne*(1024+((nl-1024)<<6)))
                               /nbgmin : 0;
         break;

         case 2:
            i2_bgdata[ipx] = nbgmin ? (intens_tot-ne*(1024+((nl-1024)<<6)))
                             /nbgmin : 0;
         break;

         case 3:
            int_bgdata[ipx] = nbgmin ? (intens_tot-ne*(1024+((nl-1024)<<6)))
            /nbgmin : 0;
         break;

         case 4:
            i2_bgsquash[ipx]= nbgmin ? sqsh((intens_tot-ne
                              *(1024+((nl-1024)<<6)))/nbgmin) : 0;
         break;
     }
   }
   else
   {
      switch (ktype)
      {
         case 1:
            byte_bgdata[ipx] = nbgmin ? (intens_tot-ne*nl)/nbgmin : 0;
         break;

         case 2:
            i2_bgdata[ipx] = nbgmin ? (intens_tot-ne*nl)/nbgmin : 0;
         break;

         case 3:
            int_bgdata[ipx] = nbgmin ? (intens_tot-ne*nl)/nbgmin : 0;
         break;

         case 4:
            i2_bgsquash[ipx] = nbgmin ? sqsh((intens_tot-ne*nl)/nbgmin) : 0;
         break;
      }
   }

   /* Get direction for next step and key pixel positions */

   last = nxt;
   switch (last)
   {
      case down:
         if (ks<max_s) 
         {
            ip = offset1 + ks*ncmp*nf_off + kf*ncmp;
            jp = ip + offset2;
            ipx += *nf_bgoff;
            ks++;
         }
         else
         {
            if (kf>=max_f) goto done;
            nxt = right;
            ip = offset1 + ks*ncmp*nf_off + kf*ncmp;
            jp = ip + npbox*ncmp;
            ipx++;
            kf++;
         }
         goto next_box;
      
      case up:
         if (ks>0) 
         {
            ks--;
            jp = offset1 + ks*ncmp*nf_off + kf*ncmp;
            ip = jp + offset2;
            ipx -= *nf_bgoff;
         }
         else
         {
            if (kf>=max_f) goto done;
            nxt = right;
            ip = offset1 + kf*ncmp;
            jp = ip + npbox*ncmp;
            ipx++;
            kf++;
            if (progress!=0) 
            {
               istrip = kf;
               (*progress)(&iflag, &max_f, &istrip);
	    }
            if (cancel!=0)
	    {
               (*cancel)(&icancl);
                if (icancl==1) return -1;
            }
         }
         goto next_box;

      case right:
         if (ks==0)
         {
            nxt = down;
            ip = offset1 + kf*ncmp;
            jp = ip + offset2;
            ipx += *nf_bgoff;
            ks = 1;
         }
         else
         {
            nxt = up;
            ks--;
            jp = offset1 + ks*ncmp*nf_off + kf*ncmp;
            ip = jp + offset2;
            ipx -= *nf_bgoff;
         }
         goto next_box;
   }
   done:

   /* Fill in strip around edge */

   for (ks=0;ks<ns_rasts/ncmp;++ks)
   {
      ip = ks*(*nf_bgoff);
      for (kf=0;kf<min_foff;++kf)
      {
         if (ks<min_soff)
         {
            jp = min_soff*(*nf_bgoff)+min_foff;
         }
         else if (ks>max_soff)
         {
            jp = max_soff*(*nf_bgoff)+min_foff;
         }
         else
         {
            jp = ks*(*nf_bgoff)+min_foff;
         }
         switch (ktype)
         {
            case 1:
               byte_bgdata[ip] = byte_bgdata[jp];
            break;

            case 2:
               i2_bgdata[ip] = i2_bgdata[jp];
            break;

            case 3:
               int_bgdata[ip] = int_bgdata[jp];
            break;

            case 4:
               i2_bgsquash[ip] = i2_bgsquash[jp];
            break;
         }
         ip++;
      }
      ip = ks*(*nf_bgoff) + max_foff + 1;
      for (kf=max_foff+1;kf<nf_rasts/ncmp;++kf)
      {
         if (ks<min_soff)
         {
            jp = min_soff*(*nf_bgoff)+max_foff;
         }
         else if (ks>max_soff)
         {
            jp = max_soff*(*nf_bgoff)+max_foff;
         }
         else
         {
            jp = ks*(*nf_bgoff)+max_foff;
         }
         switch (ktype)
         {
            case 1:
               byte_bgdata[ip] = byte_bgdata[jp];
            break;

            case 2:
               i2_bgdata[ip] = i2_bgdata[jp];
            break;

            case 3:
               int_bgdata[ip] = int_bgdata[jp];
            break;

            case 4:
               i2_bgsquash[ip] = i2_bgsquash[jp];
            break;
         }
         ip++;
      }
   }
   for (ks=0;ks<min_soff;++ks)
   {
      ip = ks*(*nf_bgoff) + min_foff;
      jp = min_soff*(*nf_bgoff) + min_foff;
      for (kf=min_foff;kf<=max_foff;++kf)
      {
         switch (ktype)
         {
            case 1:
               byte_bgdata[ip++] = byte_bgdata[jp++];
            break;

            case 2:
               i2_bgdata[ip++] = i2_bgdata[jp++];
            break;

            case 3:
               int_bgdata[ip++] = int_bgdata[jp++];
            break;

            case 4:
               i2_bgsquash[ip++] = i2_bgsquash[jp++];
            break;
         }
      }
   }
   for (ks=max_soff+1;ks<ns_rasts/ncmp;++ks)
   {
      ip = ks*(*nf_bgoff) + min_foff;
      jp = max_soff*(*nf_bgoff) + min_foff;
      for (kf=min_foff;kf<=max_foff;++kf)
      {
         switch (ktype)
         {
            case 1:
               byte_bgdata[ip++] = byte_bgdata[jp++];
            break;

            case 2:
               i2_bgdata[ip++] = i2_bgdata[jp++];
            break;

            case 3:
               int_bgdata[ip++] = int_bgdata[jp++];
            break;
            case 4:
               i2_bgsquash[ip++] = i2_bgsquash[jp++];
            break;
         }
      }
   }
   if (progress!=0)
   {
      iflag = 2;
      istrip = max_f;
      (*progress)(&iflag, &max_f, &istrip);
   }

   return too_high;
}
/* Fortran binding: imgf_bgcalc */

void imgf_bgcalc (img, itype, nsrasts, nfrasts, nfoff, ncmp, npbox, 
                   ipcnt, nullpix, img_bg, nfbgoff, progress, cancel, ierr)
int *img, *itype, *nsrasts, *nfrasts, *nfoff, *ncmp, *npbox;
int *ipcnt, *nullpix, *img_bg, *nfbgoff, *ierr;
void (*progress)();
void (*cancel)();
{
   *ierr = img_bgcalc (img, *itype, *nsrasts, *nfrasts, *nfoff, *ncmp,
                        *npbox, *ipcnt, *nullpix, img_bg, nfbgoff, 
                       progress, cancel);
   return;
}


/*-Section: Find Spots on an Image
A routine is available to search for spots in an in-memory image and a
routine enables details of the found spots to be retrieved.
-end*/

/*-Routine: Find spots on an image - img_findspots
The routine img_findspots (imgf_findspots) is used to find the 
diffraction spot positions in an image stored in memory. A pre-calculated 
background image is passed to the routine and the routine searches for pixels 
at a certain threshold above background to determine those pixels which are 
presumed to belong to spots. The routine continuously updates the spot position
by updating the calculation of its centre of gravity as the spot is
built up from contiguous above threshold pixels.  A list of spot details
is stored in an in-memory spot list managed by the 'DML' set of routines. 
-end*/

/*                                            
*************************
**   img_findspots     **
*************************

Purpose: Find spots on an mage

Author:  John W. Campbell, July 1997

CD-Fortran:
      SUBROUTINE IMGF_FINDSPOTS (IMG, ITYPE, IORDER, NF_OFF, 
     +                           NAX1_RASTS, NAX2_RASTS,
     +                           IMG_BG, NF_BGOFF, NBGCMP, IDATA, 
     +                           AX1_CEN, AX2_CEN, RMIN, RMAX,
     +                           RAX1_CEN, RAX2_CEN, 
     +                           AX1_LOW, AX1_HIGH, AX2_LOW, AX2_YHIGH, 
     +                           REXCLUDE, AX1_SPLIT, AX2_SPLIT,
     +                           NULL_PIXEL,
     +                           THRESHFAC, MINPIX, PROGRESS, CANCEL,
     +                           MINDX_SPOT, NSPOT, IERR) 
CD-end
*/
/*-Parameters:
IMG()      i (R)  Image data array (may hold signed integer data or
                  unsigned byte data packed into an integer array or
                  unsigned two byte data packed into an integer array
                  or 'squashed' two byte data packed into an integer
                  array.
ITYPE      i (R)  Image data type 
                  1 = unsigned byte data
                  2 = unsigned two-byte data (i2)
                  3 = signed integer data
                  4 = 'squashed i2' data (if stored I<0
                      ipix = -i*8
                  5 = 'squashed i2' data if stored I<0
                      ipix = (I+32768)*8
                  6 = 'squashed i2' data if stored I<0
                      ipix = (I+32768)*32
                  7 = 'squashed i2' data if stored I<0
                      ipix = -(I+1)*256 + 32768
IORDER     i (R)  Order of data in image wrt. the two local axes
                  (slow then fast)
                  =1 +ax1 +ax2
                  =2 +ax1 -ax2
                  =3 -ax1 +ax2
                  =4 -ax1 -ax2
                  =5 +ax2 +ax1
                  =6 +ax2 -ax1
                  =7 -ax2 +ax1
                  =8 -ax2 -ax1
NF_OFF     i (R)  Fast offset between (slow) rasters of image (e.g. 
                  first dimension of the image data if the data are 
                  held in a two dimensional array). This value is the 
                  number of elements for the image data type (bytes, 
                  2-bytes, integers or 'squashed i2' as set by ITYPE)
IMG_BG()   i (R)  Background image (same data format as IMG)
NF_BGOFF   i (R)  Fast offset between (slow) rasters of background 
                  image (e.g. first dimension of the background image 
                  data if the data are held in a two dimensional array).
                  This value is the number of elements for the image 
                  data type (bytes, 2-bytes or integers as set by 
                  ITYPE)
NBGCMP     i (R)  No. of pixels in each direction which were compressed
                  into 1 for the background calculation.
NAX1_RASTS i (R)  No. of rasters in the image along the first
                  local axis.
NAX2_RASTS i (R)  No. of rasters in the image along the second
                  local axis .
IDATA      i (*)  Work array to hold the spot data returned from the
                  image. Must be dimensioned to at least the maximum
                  no. of rasters in either of the two dimensions of
                  the full image.
AX1_CEN    r (R)  Centre of pattern image along the first local axis
                  in rasters (corrected)
AX2_CEN    r (R)  Centre of pattern image along the second local axis
                  in rasters (corrected)
RMIN       r (R)  Minimum radius of active area (rasters)
RMAX       r (R)  Maximum radius of active area (rasters)
RAX1_CEN   r (R)  Centre of pattern image along the first local axis
                  in rasters for RMIN calculation (Use AX1_CEN if
                  RAX1_CEN=0.0 & RAX2_CEN=0.0)
RAX2_CEN   r (R)  Centre of pattern image along the second local axis
                  in rasters for RMIN calculation (Use AX2_CEN if
                  RAX1_CEN=0.0 & RAX2_CEN=0.0)
AX1_LOW    r (R)  Low limit on 1'st local axis of area to include 
                  (0.0=image min.) (rasters)
AX1_HIGH   r (R)  High limit on 1'st local axis of area to include 
                  (0.0=image max.) (rasters)
AX2_LOW    r (R)  Low limit on 2'nd local axis of area to include 
                  (0.0=image min.) rasters)
AX2_HIGH   r (R)  High limit on on 2'nd local axis of area to include 
                  (0.0=image max.) (rasters)
REXCLUDE   r (R)  The routine uses RMIN+REXCLUDE and RMAX-REXCLUDE as 
                  actual limits for min and max radii to avoid edge of
                  detector problems with background calculation 
                  (rasters). Also uses AX1_LOW+REXCLUDE, AX1_HIGH-REXCLUDE, 
                  AX2_LOW+REXCLUDE and AX2_HIGH-REXCLUDE as limits if 
                  AX1_LOW, AX2_HIGH etc. are set. A suggestion is to use 
                  a value value of (dimension in pixels of background
                  calculation box)/sqrt(2) if XDLF_BG_CALC was used
                  to calculate the background image.
AX1_SPLIT  r (R)  Spots less than AX1_SPLIT rasters apart along the 1'st 
                  local axis are considered as one spot.
AX2_SPLIT  r (R)  Spots less than AX2_SPLIT rasters apart along the 2'nd 
                  local axis are considered as one spot.
NULL_PIXEL i (R)  The value of a 'null' pixel
THRESHFAC  i (R)  Uses pixels more than THRESHFAC*sqrt(background) above 
                  local background as being pixels which belong to spots.
MINPIX     i (R)  Minimum no. of pixels to consider as a spot
PROGRESS   s ( )  Subroutine which will be called at intervals during
                  the spots find which may be used to display its 
                  progress if required (See below for its arguments).
                  Note that the routine must be declared as EXTERNAL 
                  in the calling program. Will be called after each
                  processing 'strip' as follows:

                  CALL PROGRESS (IFLAG, NSTRIP, ISTRIP)
                  IFLAG    (R)   =0 start, =1 in progress, 
                                 =2 done
                  NSTRIP   (R)   Total no. of strips of image to 
                                 process
                  ISTRIP   (R)   Current strip (1 to NSTRIP)

CANCEL     s (R)  This routine will be called at intervals to
                  see if a cancel state has been set.
                  The call is:

                  CALL CANCEL (ICANCL)
                  ICANCL   (W)  Return 0 to continue, 1 to cancel

                  Note that the routine must be declared as EXTERNAL 
                  in the calling program.

MINDX_SPOT i (W)  'DML' record storage routines index.
                  Get data using IMGF_SPOT_DETAILS
NSPOT      i (W)  No. of spots found.
IERR       i (W)  Error flag  =0 OK
                             !=0 as returned by 'img_findspots'
-end*/

/*-C:*/
int img_findspots (int *img, int itype, int iorder, int nf_off, 
                   int nax1_rasts, int nax2_rasts,
                   int * img_bg, int nf_bgoff, int nbgcmp, int * idata, 
                   float ax1_cen, float ax2_cen, float rmin, float rmax, 
                   float rax1_cen, float rax2_cen,
                   float ax1_low, float ax1_high, 
                   float ax2_low, float ax2_high, 
                   float rexclude, float ax1_split, float ax2_split,
                   int null_pixel,
                   float threshfac, int minpix, 
                   void (*progress)(), void (*cancel)(),
                   int * mindx_spot, int * nspot) 
/*end*/
 
/*-Parameters:
int *img;         Image data array (may hold signed integer data or
                  unsigned byte data packed into an integer array or
                  unsigned two byte data packed into an integer array
                  or 'squashed' two byte data packed into an integer
                  array. (R)
int itype;        Image data type 
                  1 = unsigned byte data
                  2 = unsigned two-byte data (i2)
                  3 = signed integer data
                  4 = 'squashed i2' data (if stored I<0
                      ipix = -i*8
                  5 = 'squashed i2' data if stored I<0
                      ipix = (I+32768)*8
                  6 = 'squashed i2' data if stored I<0
                      ipix = (I+32768)*32
                  7 = 'squashed i2' data if stored I<0
                      ipix = -(I+1)*256 + 32768
                  (R)
int iorder        Order of data in image wrt. the two local axes
                  (slow then fast)
                  =1 +ax1 +ax2
                  =2 +ax1 -ax2
                  =3 -ax1 +ax2
                  =4 -ax1 -ax2
                  =5 +ax2 +ax1
                  =6 +ax2 -ax1
                  =7 -ax2 +ax1
                  =8 -ax2 -ax1
                  (R)
int nf_off;       Fast offset between (slow) rasters of image (e.g. 
                  first dimension of the image data if the data are 
                  held in a two dimensional array). This value is the 
                  number of elements for the image data type (bytes, 
                  2-bytes, integers or 'squashed i2' as set by 'itype' 
                  (R)
int *img_bg;      Background image (same data format as 'img') (R)
int nf_bgoff;     Fast offset between (slow) rasters of background 
                  image (e.g. first dimension of the background image 
                  data if the data are held in a two dimensional array).
                  This value is the number of elements for the image 
                  data type (bytes, 2-bytes or integers as set by 
                  'itype') (R)
int nbgcmp;       No. of pixels in each direction which were compressed
                  into 1 for the background calculation. (R)
int nax1_rasts;   No. of rasters in the image along the first
                  local axis. (R)
int nax2_rasts;   No. of rasters in the image along the second
                  local axis. (R)
int *idata;       Work array to hold the spot data returned from the
                  image. Must be dimensioned to at least the maximum
                  no. of rasters in either of the two dimensions of
                  the full image. (*)
float ax1_cen;    Centre of pattern image along the first local axis
                  in rasters (corrected) (R)
float ax2_cen;    Centre of pattern image along the second local axis
                  in rasters (corrected) (R)
float rmin;       Minimum radius of active area (rasters) (R)
float rmax;       Maximum radius of active area (rasters) (R)
float rax1_cen;   Centre of pattern image along the first local axis
                  in rasters for rmin calculation (Use ax1_cen if
                  rax1_cen=0.0 & rax2_cen=0.0) (R)
float rax2_cen;   Centre of pattern image along the second local axis
                  in rasters for rmin calculation (Use ax2_cen if
                  rax1_cen=0.0 & rax2_cen=0.0) (R)
float ax1_low;    Low limit on 1'st local axis of area to include 
                  (0.0=image min.) (rasters) (R)
float ax1_high;   High limit on 1'st local axis of area to include 
                  (0.0=image max.) (rasters) (R)
float ax2_low;    Low limit on 2'nd local axis of area to include 
                  (0.0=image min.) rasters) (R)
float ax2_high;   High limit on on 2'nd local axis of area to include 
                  (0.0=image max.) (rasters) (R)
float rexclude;   The routine uses RMIN+REXCLUDE and RMAX-REXCLUDE as 
                  actual limits for min and max radii to avoid edge of
                  detector problems with background calculation 
                  (rasters). Also uses AX1_LOW+REXCLUDE, AX1_HIGH-REXCLUDE, 
                  AX2_LOW+REXCLUDE and AX2_HIGH-REXCLUDE as limits if 
                  AX1_LOW, AX2_HIGH etc. are set. A suggestion is to use 
                  a value value of (dimension in pixels of background
                  calculation box)/sqrt(2) if XDLF_BG_CALC was used
                  to calculate the background image. (R)
float ax1_split;  Spots less than AX1_SPLIT rasters apart along the 1'st 
                  local axis are considered as one spot. (R)
float ax2_split;  Spots less than AX2_SPLIT rasters apart along the 2'nd 
                  local axis are considered as one spot. (R)
int null_pixel;   The value of a 'null' pixel (R)
float threshfac;  Uses pixels more than THRESHFAC*sqrt(background) above 
                  local background as being pixels which belong to spots. (R)
int minpix;       Minimum number of pixels to consider as a spot. (R)
void (*progress)() Function which will be called at intervals during
                  the spots find which may be used to display its 
                  progress if required (See below for its arguments).
                  Will be called after each processing 'strip' as follows:

                  progress (iflag, nstrip, istrip);

                  int *iflag;    =0 start, =1 in progress, 
                                           =2 done
                  int *nstrip;   Total no. of strips of image to 
                                             process
                  int *istrip;   Current strip (1 to nstrip)

                  Note: Use of pointers as may be a Fortran based routine.

                  (May be a NULL)
                  (R)

void (*cancel)(); Function passed to the routine to enable
                  cancellation. Called as follows:

                  cancel (icancl);

                  int *icancl;  Function must return 0 to 
                                continue, or 1 to cancel.

                  Note: Use of pointer as may be a Fortran based routine.

                  (May be a NULL)
                  (R)                             
int * mindx_spot; 'DML' record storage routines index. (W)
                  Get data using img_spot_details
int *nspot;       No. of spots found. (W)

-end*/

/*-Doc:

      Return     =0,   OK
                =-1,   Cannot allocate memory for temp storage
                =-2,   Cannot create found spots list
                =-3,   Cannot allocate memory for found spots (list
                       will be truncated)
               =100,   Cancelled (Note: No spots list returned)
-end*/
{

/* Note: In this routine the first local axis is called 'x' and the
         second logal axi is called 'y' */

   int mxcent;                  /* Current max no. of elements in active 
                                   spots array */
   int mxcent_old;              /* Saved value of mxcent */
   int i;                       /* Temp/loop variable */
   int j;                       /* Index */
   int l;                       /* Index */
   int ierr;                    /* Error flag */
   int kxmin;                   /* Minimum rectangle rasters limit along 'x' */
   int kxmax;                   /* Maximum rectangle rasters limit along 'x' */
   int kymin;                   /* Minimum rectangle rasters limit along 'y' */
   int kymax;                   /* Maximum rectangle rasters limit along 'y' */
   int istrip;                  /* Strip counter */
   int nstrips;                 /* No. of strips */
   int kflag;                   /* Flag for progress bar */
   int irmin;                   /* Minimum radius to use in rasters */
   int irmax;                   /* Maximum radius to use in rasters */
   int irminsq;                 /* irmin**2 */
   int ixcen;                   /* Nearest pixel 'x' centre in rasters */
   int iycen;                   /* Nearest pixel 'y' centre in rasters */
   int jxcen;                   /* Nearest pixel 'x' centre in rasters 
                                   for 'rmin' limits calculation */
   int jycen;                   /* Nearest pixel 'y' centre in rasters  
                                   for 'rmin' limits calculation */
   int ixsplit;                 /* Allowed spot split in 'x' in rasters */
   int iysplit;                 /* Allowed spot split in 'y' in rasters */
   int ixrast;                  /* Loop counter through 'x' rasters */
   int icent;                   /* Loop counter through active spots */
   int ncent;                   /* Current no. of active spots */
   int idx;                     /* 'x' distance in rasters for current 
                                   stripe */
   int idelx;                   /* 'y' distance in rasters from 'y'
                                   centre along current stripe */
   int ithresh;                 /* Threshold/intensity/value/flag */
   int ixl;                     /* 'x' low limit for search */
   int ixh;                     /* 'x' high limit for search */
   int iyl;                     /* 'y' low limit for search */
   int iyh;                     /* 'y' high limit for search */
   int ixbg;                    /* Background pixel position in 'x' */
   int iybg;                    /* Background pixel position in 'y' */
   int ibl;                     /* Beginning of line of pixels above 
                                   threshold   */
   int iel;                     /* End of line of pixels above threshold */
   int ilen;                    /* Length of line of pixels above threshold */
   int line;                    /* Continuous line of over threshold
                                   pixels in progress flag =1 yes, =0 no */
   int iypt;                    /* 'y' pixels loop counter */
   int iymid;                   /* 'y' mid point */
   int kerr;                    /* Error flag */
   int ibgval;                  /* Background value */
   int isbg;                    /* Approx sqrt of background value */
   int itmp;                    /* Temporary value */
   int icancl;                  /* 'Cancel' flag */
   float sumod;                 /* Sum of optical densities */
   float sumyod;                /* Sum of y*(optical density) */
   float fxval;                 /* 'x' position as a float */
   float r1;                    /* Radius */
   double rr;                   /* Temp value */

   Active_Spot *actspots;       /* Active spots array */
   Active_Spot *p;              /* Active spots array - temp pointer */
   int * ipt;                   /* Index array used along with actspots
                                   array */
   int *pp;                     /* Temp pointer */
   Found_Spot spot;             /* Found spot structure */

/* Initialisations */

   *nspot = 0;
   ierr = 0;
   mxcent = MAX_ACT_START;
   actspots = (Active_Spot *) malloc (mxcent*sizeof(Active_Spot));
   if (actspots==0) return -1;
   ipt = (int *)  malloc (mxcent*sizeof(int));
   if (ipt==0)
   {
      free(actspots);
      return -1;
   }
   for (i=0;i<mxcent;++i) actspots[i].npixcent = 0;
   kxmin = 1;
   if (ax1_low>0.0) kxmin = nint(ax1_low);
   kxmax = nax1_rasts;
   if (ax1_high>0.0) kxmax = nint(ax1_high);
   kymin = 1;
   if (ax2_low>0.0) kymin = nint(ax2_low);
   kymax = nax2_rasts;
   if (ax2_high>0.0) kymax = nint(ax2_high);
   kxmin = nint(kxmin+rexclude);
   kxmax = nint(kxmax-rexclude);
   kymin = nint(kymin+rexclude);
   kymax = nint(kymax-rexclude);

/* Create output spots list */

   *mindx_spot = dml_init(INIT_SPOT_RECS, INC_SPOT_RECS ,sizeof(Found_Spot));
   if (*mindx_spot<0) return -2;

   irmax = nint(rmax-rexclude);
   irmin = 0;
   if (rmin>0.0) irmin = nint(rmin+rexclude);
   irminsq = irmin*irmin;

/* Calculate limits along first local axis */

   ixcen = nint(ax1_cen);
   iycen = nint(ax2_cen);
   jxcen = ixcen;
   jycen = iycen;
   if (rax1_cen!=0.0&&rax2_cen!=0.0)
   {
      jxcen = nint(rax1_cen);
      jycen = nint(rax2_cen);
   }
   ixl = 1;
   if (ixcen-irmax>ixl) ixl = ixcen-irmax;
   if (kxmin>ixl) ixl = kxmin;
   ixh = nax1_rasts;
   if (ixcen+irmax<ixh) ixh = ixcen+irmax;
   if (kxmax<ixh ) ixh = kxmax;
   nstrips = ixh - ixl + 1;
   ixsplit = nint(ax1_split);
   iysplit = nint(ax2_split);
   istrip = 0;
   kflag = 0;
   if (progress!=0) (*progress)(&kflag, &nstrips, &istrip);
   kflag = 1;
   ncent = 0;

/* Loop over stripes */

   for (ixrast=ixl;ixrast<=ixh;++ixrast)
   {
      idx = ixcen - ixrast;

/* Determine limits on second local axis */

      rr = (double)(irmax*irmax-idx*idx);
      if (rr<0.0) continue;
      idelx = nint(sqrt(rr));
      iyl = max(1,max(iycen-idelx,kymin));
      iyl = 1;
      if (iycen-idelx>iyl) iyl = iycen-idelx;
      if (kymin>iyl) iyl = kymin;
      iyh = nax2_rasts;
      if (iycen+idelx<iyl) iyl = iycen+idelx;
      if (kymax<iyl) iyl = kxmax;
/* Get line of pixels from image */

      kerr = img_ax2line(img,itype,iorder,nf_off,ixrast,1,
                         nax2_rasts,nax1_rasts,nax2_rasts,0,idata);

/* Loop over pixels to find lines of continuous threshold density */

      fxval = (float) ixrast;
      line = 0;
      ixbg = (ixrast-1)/nbgcmp + 1;

      for (iypt=iyl;iypt<=iyh;++iypt)
      {
         iybg = (iypt-1)/nbgcmp + 1;
         kerr = img_getpix (img_bg, itype, iorder, nf_bgoff, ixbg, iybg,
                            nax1_rasts/nbgcmp, nax2_rasts/nbgcmp, 0, &ibgval);
 
         if (ibgval>1)
	 {
            isbg = ibgval>>1;
            isbg += ibgval/isbg;
            isbg >>=1;
            isbg += ibgval/isbg;
            isbg >>=1;
            isbg += ibgval/isbg;
            isbg >>=1;
         }
         else
	 {
            isbg = ibgval;
	 }

         ithresh = idata[iypt-1] - ibgval - threshfac*isbg;
         if (abs(jxcen-ixrast)<=irmin)
         {
            if ((iypt-jycen)*(iypt-jycen) 
               +(ixrast-jxcen)*(ixrast-jxcen)<=irminsq) ithresh = -1;
         }
         if (idata[iypt-1]==null_pixel) ithresh = -1;

         if (!line)
         {
            if (ithresh<=0) continue;
            line = 1;
            ibl = iypt;
            sumyod = (float)(ithresh*iypt);
            sumod = (float) ithresh;
         }
         else if (ithresh<=0||iypt==iyh)
         {
            line = 0;
            iel = iypt - 1;
            ilen = iel - ibl + 1;

/* Inner loop to decide which spots are ovelapping */
            
            for (icent=1;icent<=ncent;++icent)
            {
               j = ipt[icent-1];

/* skipping condition */

               itmp = abs(iymid-actspots[j].iycent);
               if (abs((actspots[j].iyend+actspots[j].iybeg) -(iel+ibl))>
                   abs((actspots[j].iyend-actspots[j].iybeg) +(iel-ibl))&&
                   itmp>iysplit) continue;

               actspots[j].sumxi += sumod*fxval;
               actspots[j].sumyi += sumyod;
               actspots[j].sumi += sumod;
               actspots[j].ixcent = nint(actspots[j].sumxi/actspots[j].sumi);
               actspots[j].iycent = nint(actspots[j].sumyi/actspots[j].sumi);
               actspots[j].npixcent = actspots[j].npixcent + ilen;
               actspots[j].iybeg = ibl;
               actspots[j].iyend = iel;
               actspots[j].ixmax = ixrast;
               actspots[j].iymin = min(ibl,actspots[j].iymin);
               actspots[j].iymax = max(iel,actspots[j].iymax);

/* Flag spot if line overlaps search area limits */

               if (ibl==iyl||iypt==iyh) actspots[j].iflag = 1;
               goto lab70;
            }

/* Count this active spot */

            ncent = ncent + 1;
            if (ncent>mxcent)
            {
               p = (Active_Spot *) realloc (actspots,
                                   (mxcent+MAX_ACT_INC)*sizeof(Active_Spot));
               if (p==0)
               {
                  free (actspots);
                  free (ipt);
                  return -1;
               }
               actspots = p;
               pp = (int *) realloc (ipt,
                                   (mxcent+MAX_ACT_INC)*sizeof(int));
               if (pp==0)
               {
                  free (actspots);
                  free (ipt);
                  return -1;
               }
               ipt = pp;
               mxcent_old = mxcent;
               mxcent += MAX_ACT_INC;
               for (i=mxcent_old;i<mxcent;++i) actspots[i].npixcent = 0;
	    } 

/* Find empty slot in which to store info about this spot */

            for(i=0;i<mxcent;++i)
            {
               if (actspots[i].npixcent==0)
               {
                  ipt[ncent-1] = i;
                  break;
               }
            }

            l = ipt[ncent-1];
            actspots[l].npixcent = ilen;
            actspots[l].sumxi = sumod*fxval;
            actspots[l].sumyi = sumyod;
            actspots[l].sumi = sumod;
            actspots[l].iybeg = ibl;
            actspots[l].iyend = iel;
            actspots[l].ixcent = nint(actspots[l].sumxi/actspots[l].sumi);
            actspots[l].iycent = nint(actspots[l].sumyi/actspots[l].sumi);
            actspots[l].ixmin = ixrast;
            actspots[l].ixmax = ixrast;
            actspots[l].iymin = ibl;
            actspots[l].iymax = iel;

/* Flag spots if srip overlaps search area */

            if (ibl==iyl||ixrast==ixl)
            {
               actspots[l].iflag = 1;
            }
            else
            {
               actspots[l].iflag = 0;
            } 

/* Ongoing line */
        
         }
         else
         {
            sumyod += (float)(iypt*ithresh);
            sumod += (float) ithresh;
         }
         lab70: continue;
      }

/* End of loop over pixels */

/* Loop over active spots and store those that have finished */

/* Skip if no active spots */

      if (ncent==0) goto lab120;
      icent = 0;
      lab80:
      icent = icent + 1;
      j = ipt[icent-1];

/* Spots considered finished if no pixels for spot found on last stripe
   and current stripe at least ixsplit from centroid */
 
      if (ixrast-actspots[j].ixcent<=ixsplit||
          ixrast==actspots[j].ixmax) goto lab110;


/* Reject spot if it contains less than minpix pixels */

      if (actspots[j].npixcent<minpix) goto lab90;

/* Reject spot if it overlaps limits of search area */

      if (actspots[j].iflag==1) goto lab90;

/* Spots coordinates as centre of gravity */

      spot.ax1_pos = actspots[j].sumxi/actspots[j].sumi;
      spot.ax2_pos = actspots[j].sumyi/actspots[j].sumi;

/* Store spots if centroid lies within rmax limit */

      spot.siz_ax1 = abs(actspots[j].ixmax-actspots[j].ixmin) + 1;
      spot.siz_ax2 = abs(actspots[j].iymax-actspots[j].iymin) + 1;
      r1 = sqrt((ixcen-spot.ax1_pos)*(ixcen-spot.ax1_pos)+
                (iycen-spot.ax2_pos)*(iycen-spot.ax2_pos));
      if (r1>irmax) goto lab90;
      *nspot = *nspot + 1;
      spot.sumint = actspots[j].sumi;
      spot.npix = actspots[j].npixcent;
      ierr = dml_addrec(*mindx_spot, &spot);
      if (ierr!=0)
      {
         *nspot = *nspot - 1;
         free (actspots);
         free (ipt);
         return -3;
      }

/* Spot stored or rejected */
/* Set flag and shuffle IPT down to fill gap in array */
      lab90:
      actspots[j].npixcent = 0;
      ncent = ncent - 1;

/* Skip if last spot */

      if (icent>ncent) goto lab120;

      for (i=icent;i<=ncent;++i)
      {
         ipt[i-1] = ipt[i];
      }

      lab110:
      if (icent<ncent) goto lab80;
      lab120:
      if (cancel!=0)
      {
         (*cancel)(&icancl);
         if (icancl==1)
         { 
            free(actspots);
            free (ipt);
            dml_free (*mindx_spot);
            *mindx_spot = -1;
            if (progress!=0)
            {
               kflag = 2;
               (*progress)(&kflag, &nstrips, &istrip);
            }
            return -100;
         }
      }
      istrip = ixrast - ixl + 1;
      if (progress!=0) 
      {
         (*progress)(&kflag, &nstrips, &istrip);
      }
   }
   free(actspots);
   free (ipt);
   if (progress!=0)
   {
      kflag = 2;
      (*progress)(&kflag, &nstrips, &istrip);
   }
   return 0;
}

/* Fortran binding: imgf_findspots */

void imgf_findspots (int *img, int *itype, int *iorder, int *nf_off, 
                     int *nax1_rasts, int *nax2_rasts,
                     int *img_bg, int *nf_bgoff, int *nbgcmp, int *idata, 
                     float *ax1_cen, float *ax2_cen, float *rmin, float *rmax,
                     float *rax1_cen, float *rax2_cen, 
                     float *ax1_low, float *ax1_high, 
                     float *ax2_low, float *ax2_high, 
                     float *rexclude, float *ax1_split, float *ax2_split,
                     int *null_pixel,
                     float *threshfac, int *minpix, 
                     void (*progress)(), void (*cancel) (),
                     int * mindx_spot, int * nspot, int *ierr)
{
 
*ierr = img_findspots (img, *itype, *iorder, *nf_off, 
                       *nax1_rasts, *nax2_rasts,
                       img_bg, *nf_bgoff, *nbgcmp, idata, 
                       *ax1_cen, *ax2_cen, *rmin, *rmax,
                       *rax1_cen, *rax2_cen, 
                       *ax1_low, *ax1_high, 
                       *ax2_low, *ax2_high, 
                       *rexclude, *ax1_split, *ax2_split,
                       *null_pixel,
                       *threshfac, *minpix, 
                       progress, cancel,
                       mindx_spot, nspot);
   return;
} 
/*-Routine: Get spot details - img_spot_details
Return the details for a found spot.
-end*/

/*                                            
****************************
**   img_spot_details     **
****************************

Purpose: Return spot details

Author:  John W. Campbell, February 1998

CD-Fortran:
      SUBROUTINE IMGF_SPOT_DETAILS (MINDX, ISPOT, AX1_POS, AX2_POS,
     +                              SPOTINT, NPIX, ISIZ_AX1, ISIZ_AX2,
     +                              IERR
CD-end
*/
/*-Parameters:
MINDX       i  (R)  Index for the DML spots list from IMGF_FINDSPOTS
ISPOT       i  (R)  Spot number from 1 upwards
AX1_POS     r  (W)  Spot position along first local axis in rasters
AX2_POS     r  (W)  Spot position along second local axis in rasters
SPOTINT     r  (W)  Spot intensity
NPIX        i  (W)  No pixels in spot
ISIZ_AX1    i  (W)  Spot size in pixels along 1'st local axis
ISIZ_AX2    i  (W)  Spot size in pixels along 2'nd local axis
IERR        i  (W)  Error return flags fromm img_spot_details
-end*/


/*-C:*/
int img_spot_details (int mindx, int ispot, float *ax1_pos, float *ax2_pos,
                      float * spotint, int * npix, int * isiz_ax1,
                      int * isiz_ax2)
/*end*/
 
/*-Parameters:
int mindx;       Index for the DML spots list from img_findspots (R)
int ispot;       Spot number from 1 upwards. 
float * ax1_pos; Spot position along 1'st local axis in rasters (W)  
float * ax2_pos; Spot position along 2'nd local axis in rasters (W)
float * spotint; Spot intensity (W)
int * npix;      No. of pixels in spot (W)  
int *isiz_ax1;   Spot size in pixels along 1'st local axis (W)
int *isiz_ax2;   Spot size in pixels along 2'nd local axis (W)

Return     =0,   OK
           =1,   Invalid index
           =2,   Spot out of range
-end*/

{
   int nspots;
   Found_Spot * spot;

   nspots = dml_numrecs (mindx);
   if (nspots<1) return 1;
   if (ispot<1||ispot>nspots) return 2;

   spot = (Found_Spot *) dml_recpointer (mindx, ispot);
   *ax1_pos = spot->ax1_pos;
   *ax2_pos = spot->ax2_pos;
   *spotint = spot->sumint;
   *npix = spot->npix;
   *isiz_ax1 = spot->siz_ax1;
   *isiz_ax2 = spot->siz_ax2;
   return 0;
}
/* Fortran binding: imgf_spot_details */

void imgf_spot_details (int * mindx, int * ispot, float * ax1_pos, 
                        float * ax2_pos, float * spotint, int * npix, 
                        int * isiz_ax1, int * isiz_ax2, int * ierr) 
{
*ierr = img_spot_details (*mindx, *ispot, ax1_pos, ax2_pos, spotint,
                          npix, isiz_ax1, isiz_ax2);
}




/*-Section: Passing Fortran Character Strings
A special function is supplied for use when passing character string
parameters from a Fortran to 'C' image handling routine.
-end*/


/*-Routine: Pointers list index of a Fortran character string - imgstr
The routine imgstr is used to return an index to a list of character string
pointers held within 'imglib' and used where a Fortran callable routine 
passes character data to or from the routine. The actual paramter(s) expected 
by the C routine are machine dependent.  The Fortran call is of the form:

               IMGSTR (CHS)   where CHS is a Fortran character variable 
                              or string.

The function IMGSTR must be declared in the calling Fortran program as
an INTEGER FUNCTION.
-end*/
/*
*********************
**     imgstr      **
*********************

Purpose: Return the index in IMG_pointers array of a Fortran character string
         being passed as a parameter from a Fortran program

Author:  John W. Campbell, Nov 1997

*/

#if   CHAR_STRING_TYPE == 1

int imgstr (ch, len)

char *ch;    /* Address of the character string (R)*/
int len;     /* Length of the character string (R)*/

{
   IMG_pointers_index++;
   if (IMG_pointers_index>=MAXPOINTERS) IMG_pointers_index = 0;
   IMG_pointers[IMG_pointers_index] = ch;
   return IMG_pointers_index;
}

#endif

#if CHAR_STRING_TYPE == 2

typedef struct
{
   short int len;
   unsigned char type;
   unsigned char class;
   char * addr;
}
   Vax_ch_desc;

int imgstr (ch_desc)
 
Vax_ch_desc *ch_desc;     /* Pointer to character descriptor (R)*/

/*-Doc:
Return:  Index in the internal IMG_pointers list from 0 to MAXPOINTERS-1
-end*/
{
   IMG_pointers_index++;
   if (IMG_pointers_index>=MAXPOINTERS) IMG_pointers_index = 0;
   IMG_pointers[IMG_pointers_index] = ch_desc->addr;
   return IMG_pointers_index;
}

#endif

/*==================================================*/
/*         'Private' SERVICE ROUTINES               */
/*==================================================*/


/*
****************************
**     img_desq           **
****************************

Purpose: Get intensity value from 'squashed i2' pixel (type 4)

Author:  John W. Campbell

*/

long img_desq (ipx)

I2squash ipx;               /* Stored pixel value */
{
   if (ipx>=0) return (long) ipx;
   return (long) -ipx*8;
}

/*
****************************
**     img_desq2          **
****************************

Purpose: Get intensity value from 'squashed i2' pixel (type 5 - RAXIS II)

Author:  John W. Campbell

*/

long img_desq2 (ipx)

I2squash ipx;               /* Stored pixel value */
{
   if (ipx>=0) return (long) ipx;
   return (long) (ipx+32768)*8;
}

/*
****************************
**     img_desq3          **
****************************

Purpose: Get intensity value from 'squashed i2' pixel (type 6 - RAXIS IV)

Author:  John W. Campbell, September 1996

*/

long img_desq3 (ipx)

I2squash ipx;               /* Stored pixel value */
{
   if (ipx>=0) return (long) ipx;
   return (long) (ipx+32768)*32;
}

/*
****************************
**     img_desq4          **
****************************

Purpose: Get intensity value from 'squashed i2' pixel (type 7 - MAC SCIENCE)

Author:  John W. Campbell

*/

long img_desq4 (ipx)

I2squash ipx;               /* Stored pixel value */
{
   if (ipx>=0) return (long) ipx;
   return (long) -(ipx+1)*256 + 32768;
}

/*
****************************
**     img_unsq1          **
****************************

Purpose: Get intensity value from unsigned 'i2' pixel (type 2)

Author:  John W. Campbell, August 1997

*/

int img_unsq1 (ipx)

I2pix ipx;               /* Stored pixel value */
{
   if (ipx>=0) return (int) ipx;
   return (int) 65536 + ipx;
}

/*
****************************
**     img_unsq2          **
****************************

Purpose: Get intensity value from 'squashed i2' pixel (type 4)

Author:  John W. Campbell,  August 1997

*/

int img_unsq2 (ipx)

I2pix ipx;               /* Stored pixel value */
{
   if (ipx>=0) return (int) ipx;
   return (int) -ipx*8;
}

/*
****************************
**     img_unsq3          **
****************************

Purpose: Get intensity value from 'squashed i2' pixel (type 5 - RAXIS II)

Author:  John W. Campbell,  August 1997

*/

int img_unsq3 (ipx)

I2pix ipx;               /* Stored pixel value */
{
   if (ipx>=0) return (int) ipx;
   return (int) (ipx+32768)*8;
}

/*
****************************
**     img_unsq4          **
****************************

Purpose: Get intensity value from 'squashed i2' pixel (type 6 - RAXIS IV)

Author:  John W. Campbell,  August 1997

*/

int img_unsq4 (ipx)

I2pix ipx;               /* Stored pixel value */
{
   if (ipx>=0) return (int) ipx;
   return (int) (ipx+32768)*32;
}

/*
****************************
**     img_unsq5          **
****************************

Purpose: Get intensity value from 'squashed i2' pixel (type 7 - MAC SCIENCE)

Author:  John W. Campbell,  August 1997

*/

int img_unsq5 (ipx)

I2pix ipx;               /* Stored pixel value */
{
   if (ipx>=0) return (int) ipx;
   return (int) -(ipx+1)*256 + 32768;
}
/*
****************************
**     img_byte_log       **
****************************

Purpose: Get intensity value from Photon factory log byte data

Author:  John W. Campbell,  November 1997

*/

int img_byte_log (ipx)

unsigned char ipx;               /* Stored pixel value */
{
   return nint(pow(10.0,4.0*(float)ipx/255.0)-1.0);
}

/*
****************************
**     img_sqsh1          **
****************************

Purpose: Set intensity value for unsigned 'i2' pixel (type 2)

Author:  John W. Campbell, August 1997

*/

I2pix img_sqsh1 (ipx, ovfl)

int ipx;               /* Pixel value to store */
int *ovfl;             /* Overflow flag (set to 1 if overflow found) */
{
   if (ipx<0) ipx = 0;
   if (ipx<32768) return (I2pix) ipx;
   if (ipx>65535)
   {
      *ovfl = 1;
      ipx = 65535;
   }
   return (I2pix) (ipx - 65536);
}

/*
****************************
**     img_sqsh2          **
****************************

Purpose: Set intensity value for 'squashed i2' pixel (type 4)

Author:  John W. Campbell, August 1997

*/

I2pix img_sqsh2 (ipx, ovfl)

int ipx;               /* Pixel value to store */
int *ovfl;             /* Overflow flag (set to 1 if overflow found) */
{
   if (ipx<0) ipx = 0;
   if (ipx<32768) return (I2pix) ipx;
   if (ipx>262144)
   {
      *ovfl = 1;
      ipx = 262144;
   }
   return (I2pix) (- (ipx+3)/8);
}

/*
****************************
**     img_sqsh3          **
****************************

Purpose:  Set intensity value for 'squashed i2' pixel (type 5 - RAXIS II)

Author:  John W. Campbell, August 1997

*/

I2pix img_sqsh3 (ipx, ovfl)

int ipx;               /* Pixel value to store */
int *ovfl;             /* Overflow flag (set to 1 if overflow found) */
{
   if (ipx<0) ipx = 0;
   if (ipx<32768) return (I2pix) ipx;
   if (ipx>262136)
   {
      *ovfl = 1;
      ipx = 262136;
   }
   return (I2pix) ((ipx+3)/8 - 32768);
}

/*
****************************
**     img_sqsh4          **
****************************

Purpose:  Set intensity value for 'squashed i2' pixel (type 6 - RAXIS IV)

Author:  John W. Campbell, August 1997

*/

I2pix img_sqsh4 (ipx, ovfl)

int ipx;               /* Pixel value to store */
int *ovfl;             /* Overflow flag (set to 1 if overflow found) */
{
   if (ipx<0) ipx = 0;
   if (ipx<32768) return (I2pix) ipx;
   if (ipx>1048544)
   {
      *ovfl = 1;
      ipx = 1048544;
   }
   return (I2pix) ((ipx+15)/32 - 32768);
}

/*
****************************
**     img_sqsh5          **
****************************

Purpose:  Set intensity value for 'squashed i2' pixel (type 7 - MAC SCIENCE)

Author:  John W. Campbell, August 1997

*/

I2pix img_sqsh5 (ipx, ovfl)

int ipx;               /* Pixel value to store */
int *ovfl;             /* Overflow flag (set to 1 if overflow found) */
{
   if (ipx<0) ipx = 0;
   if (ipx<32768) return (I2pix) ipx;
   if (ipx>8421120)
   {
      *ovfl = 1;
      ipx = 8421120;
   }
   return (I2pix) ((32385-ipx)/256);
}

/*
****************************
**     set_mar_ovpix      **
****************************

Purpose:  Set MAR overload pixel value into image

Author:  John W. Campbell,November 1997

**** SERVICE ROUTINE ONLY ****
*/
void set_mar_ovpix (void *img, int * ovpix, int iord, 
                   int nxrasts, int nyrasts, int out_typ, int * ovfl)
{
   int s;
   int f;
   int nf;
   int xoff;
   int yoff;
   int ioff;
   int ipix;
   Intpix * img_int;
   I2pix * img_i2;

   img_int = (Intpix *) img;
   img_i2 = (I2pix *) img;
   ipix = ovpix[1];
   nf = nyrasts;
   if (iord>4) nf = nxrasts;

   s = (ovpix[0]-1)/nf;
   f = (ovpix[0]-1)%nf;

   switch (iord) 
   {
       case 1:  /* +ax1, +ax2 */
          xoff = s;
          yoff = f;
       break;

       case 2:  /* +ax1, -ax2 */
          xoff = s;
          yoff = nxrasts - f  - 1;
       break;         

       case 3:  /* -ax1, +ax2 */
          xoff = nxrasts - s - 1;
          yoff = f;
       break;

       case 4:  /* -ax1, -ax2 */
          xoff = nxrasts - s - 1;
          yoff = nyrasts - f - 1;
       break;

       case 5:  /* +ax2, +ax1 */
          xoff = f;
          yoff = s;
       break;

       case 6:  /* +ax2, -ax1 */
          xoff = f;
          yoff = nyrasts - s - 1;
       break;         

       case 7:  /* -ax2, +ax1 */
          xoff = nxrasts - f - 1;
          yoff = s;
       break;

       case 8:  /* -ax2, -ax1 */
          xoff = nxrasts - f - 1;
          yoff = nyrasts - s - 1;
       break;
   }
   ioff = xoff*nxrasts + yoff;

   switch (out_typ)
   {
      case 2:
      img_i2[ioff] = img_sqsh1 (ipix, ovfl);
      break;

      case 3:
      img_int[ioff] = ipix;
      break;

      case 4:
      img_i2[ioff] = img_sqsh2 (ipix, ovfl);
      break;

      case 5:
      img_i2[ioff] = img_sqsh3 (ipix, ovfl);
      break;

      case 6:
      img_i2[ioff] = img_sqsh4 (ipix, ovfl);
      break;

      case 7:
      img_i2[ioff] = img_sqsh5 (ipix, ovfl);
      break;
   }
   return;
}
/*
******************************
**      img_copy_chars      **
******************************

Purpose: Copy a string; The output string may be either a 'c' type string
         which will be null terminated, or a 'fortran' type string which
         will be padded with blanks if required but will have no terminating
         null.

Author:  John Campbell, November 1997 (modified copy of xdl_copy_chars)

**** SERVICE ROUTINE ONLY ****
*/
void img_copy_chars (char * string_in, int lin, char * string_out, int max_len)
/*
char * string_in;    The string to be copied  (R)
int lin;             Length of string_in; may be 0 if string is null
                     terminated (R)
char * string_out;   The output string address for the copied string (W) 
int    max_len;      If >0 it is the maximum length allowed for the null
                           terminated copied string. string_out must be 
                           at least (max_len + 1) characters in length.
                     if <0 then up to abs(max_len) characters are copied 
                           to the output string. If the number of 
                           characters in the input string is less than 
                           this then the output string will be padded to 
                           abs(max_len) characters with blanks. (R) 
*/
{
   int len;     /* String length */
   int i;       /* Loop counter */

   if (max_len==0) return;

   len = lin;
   if (len<=0) len =strlen (string_in);
   if (max_len>0)
   {
      if (len>max_len) len = max_len;
      strncpy (string_out, string_in, len);
      string_out[len] = 0;
      return;
   }
   else
   {
      max_len = - max_len;
      if (len>max_len) len = max_len;
      strncpy (string_out, string_in, len);
      if (max_len>len)
      {
         for (i=len;i<max_len;++i) string_out[i] = ' ';
      }
   }
   return;
}
/*
***************************
**    img_blank_string   **
***************************

Purpose: Test for a blank(/null) character string

Author:  J. W. Campbell, November 1997

**** SERVICE ROUTINE ONLY ****
*/

int img_blank_string (char * str, int lens)
 
/*
char str[];   String to test (null terminated) (R)
int lens;     Length of string - may be 0 if null terminated (R)

Return:  =1 (true) blank or null string found,
         =0 not a blank/null string
*/

{
   int i;     /*Temp/loop variables*/
   int l;     /*Offset of last character in string*/

   if (str==0) return 1;
   l = lens;
   if (l==0) l = strlen (str);
   if (l==0) return 1;
   for (i=0;i<l;++i)
   {
      if (str[i]!=' ') return 0;
   }
   return 1;
}
