/*===========================================*/
/*       PRL Reflection list   routines      */
/*===========================================*/

/*-Title: PREDICTED REFLECTION LISTS (PRL)
-end*/

/*-Intro:
These are a series of routines concerned with the allocation
and management of memory for lists of reflection records to be 
used with the RPF set of reflection prediction routines. The routines make
use of the DML set of memory management routines for in-memory record lists. 
-end*/

/*Routines:
The following routines are available:

     prl_init      Allocate and initialise a reflections list
     prl_addref    Add a reflection to a reflections list
     prl_puthkl    Update reflection indices in a reflections list
     prl_putref    Update a reflection in a reflections list
     prl_numrefs   Get number of reflections in a list
     prl_gethkl    Get reflection indices  for a reflection
     prl_getref    Get reflection details for a reflection
     prl_getphrc   Get further information for a reflection
     prl_getcoords Get coordinates for a reflection    
     prl_delimg    Delete reflections for a given image
     prl_clear     Clear a reflections list
     prl_free      Clear and free a reflections list
     prl_sort      Sort a reflections list
     prl_ovlp1     Determine spatial overlaps

     The equivalent Fortran interface routines are named prlf...
*/

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include <stdio.h>

#include "dml.h"

#include "prl_systyp.h"
#include "prl.h"

/* Define Fortran binding names */

#if    LINKTYP == 1

#define prlf_init prlf_init_
#define prlf_addref prlf_addref_ 
#define prlf_puthkl prlf_puthkl_
#define prlf_putref prlf_putref_
#define prlf_numrefs prlf_numrefs_
#define prlf_gethkl prlf_gethkl_
#define prlf_getref prlf_getref_
#define prlf_getphrc prlf_getphrc_
#define prlf_getcoords prlf_getcoords_
#define prlf_delimg prlf_delimg_
#define prlf_clear prlf_clear_
#define prlf_free prlf_free_
#define prlf_sort prlf_sort_
#define prlf_ovlp1 prlf_ovlp1_

#endif

#if  LINKTYP == 2

#define prlf_init PRLF_INIT
#define prlf_addref PRLF_ADDREF 
#define prlf_puthkl PRLF_PUTHKL
#define prlf_putref PRLF_PUTREF
#define prlf_numrefs PRLF_NUMREFS
#define prlf_gethkl PRLF_GETHKL
#define prlf_getref PRLF_GETREF
#define prlf_getphrc PRLF_GETPHRC
#define prlf_getcoords PRLF_GETCOORDS
#define prlf_delimg PRLF_DELIMG
#define prlf_clear PRLF_CLEAR
#define prlf_free PRLF_FREE
#define prlf_sort PRLF_SORT
#define prlf_ovlp1 PRLF_OVLP1

#endif

void prlf_init();
void prlf_addref();
void prlf_puthkl();
void prlf_putref();
void prlf_numrefs();
void prlf_gethkl();
void prlf_getref();
void prlf_getphrc();
void prlf_getcoords();
void prlf_delimg();
void prlf_clear();
void prlf_free();
void prlf_sort();
void prlf_ovlp1();

typedef struct _PRLreflection
{
   int hkl;     /* Packed reflection indices 2**20(h+512)+2**10(k+512)+l+512 */
   float ax1_coord; /* Axis 1 (e.g. x) detector coordinate in mm */
   float ax2_coord; /* Axis 2 (e.g. y) detector coordinate in mm */
   float phi;   /* Phi value when reflection centered on Ewald sphere */
   float xrl;   /* Reciprocal lattice 'x' coordinate (wrt laboratory axes)*/
   float yrl;   /* Reciprocal lattice 'y' coordinate (wrt laboratory axes)*/
   float zrl;   /* Reciprocal lattice 'y' coordinate (wrt laboratory axes)*/
   float rlorf; /* Lorentz factor */
   unsigned short icrys;     /* Crystal number */
   unsigned short imgnum;    /* Image number */
   unsigned nimg :   12;     /* No. of images (see below) */
   signed  ifirst : 12;      /* Offset for first image (see below) */
   unsigned ir_flag : 3;     /* Recorded flag (see below) */
   unsigned inr_flag : 3;    /* Not-recorded flag (see below) */
   unsigned spat : 1;        /* Spatially overlapped flag (see below) */
   unsigned char ksym;       /* Symmetry no. 0=not defined (usually 0 if
                                measured/predicted indices stored or set
                                to symmetry no. if indices have been
                                converted to unique set */
}PRLreflection;

/*
    Flags:

    nimg =  No. of images that the reflection is spread over (1=fully
            recorded spot, 0 = wider than the maximum requested limit
            when reflection prediction done) 

  ifirst =  First image for the spot as an offset wrt. the current image.

 ir_flag =  Recorded flag       0 = not recorded
                                1 = full
                                2 = good partial (<= nwmax images)
                                3 = too-wide partial
                                4 = within cusp but part visible

inr_flag =  Not recorded flag   0 = recorded
                                1 = on rotation axis
                                2 = within cusp
                                3 = > extended resol limit
                                4 = not-recorded but still in list

    spat =  Spatially overlapped flag 1=yes, 0=no
*/
typedef struct _PRLimgcrys
{
  unsigned short icrys;
  unsigned short imgnum;
}PRLimgcrys;

#define two20 1048576
#define two10 1024
#define MAXSPOTS_IN_STRIP 1000


/*-Section: Creating and Writing 'PRL' lists
Routines are available to create new PRL reflection lists, to add records
to such lists and to modify data in such lists.
-end*/

/*-Routine: Allocate and initialise a reflections list  - prl_init
The routine prl_init (prlf_init) checks to see if a spare 
slot is available and, if so, allocates the memory for a new reflections list 
with an initial allocation of records. The routine returns an index (handle) 
to be used in all other routines for accessing that list.
-end*/

/*                                            
******************************
**    prl_init              **
******************************

Purpose: Allocate and initialise a reflections list (For PRL functions)

Author:  John W. Campbell, May 1995

CD-Fortran:
         CALL PRLF_INIT  (INIT_ALLOC, INT_ALLOC, MINDX)
CD-end
*/
/*-Parameters:
INIT_ALLOC   i  (R)  Initial allocation required (no. of reflection
                     records).
INC_ALLOC    i  (R)  No. of records by which to extend the 
                     reflections list each time an extension is 
                     required.
MINDX        i  (W)  Index for the allocated reflection list (>=0)
                     A negative return value indicates an error and
                     the value is the return code from the routine
                     prl_init.
-end*/

/*-C:*/
int prl_init (int init_alloc, int inc_alloc)
/*end*/

/*-Parameters:
int init_alloc;      No. of records to be allocated initially (R)
int inc_alloc;       No. of records to be added each time an extension
                     to the list is required (R)

Return:  >=0, the index of the allocated list
          -1, no spare lists available
          -2, cannot allocate initial memory
          -3, init_alloc<1 or inc_alloc<1
-end*/
{
   return dml_init (init_alloc, inc_alloc, sizeof(PRLreflection));
}

/* Fortran binding: prlf_init */

void prlf_init (int * init_alloc, int * inc_alloc, int *mindx)
{
   *mindx = prl_init (*init_alloc, *inc_alloc);
   return;
}

/*-Routine: Add a reflection to a reflections list  - prl_addref
The routine prl_addref (prlf_addref) adds a new reflection to 
the end of the requested reflections list providing that the list has been 
initialised and that sufficient memory is available or can be allocated for it.
-end*/

/*                                            
*******************************
**    prl_addref             **
*******************************

Purpose: Add a reflection record to a reflections list

Author:  John W. Campbell, May 1995


CD-Fortran:
         CALL PRLF_ADDREF  (MINDX, ICRYS, IMGNUM, IH, IK, IL, KSYM,
        +                          AX1_COORD, AX2_COORD, NIMG, IFIRST, 
        +                          IR_FLAG, INR_FLAG, ISPAT,
        +                          PHI, XRL, YRL, ZRL, , RLORF, IERR)
CD-end
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by PRLF_INIT
ICRYS        i  (R)  Crystal number
IMGNUM       i  (R)  Image number
IH           i  (R)  Reflection 'h' index
IK           i  (R)  Reflection 'k' index
IL           i  (R)  Reflection 'l' index
KSYM         i  (R)  Symmetry no. (0 if indices are as measured)
AX1_COORD    r  (R)  Axis 1 (e.g. 'x') detector coordinate in mm
AX2_COORD    r  (R)  Axis 2 (e.g. 'y') detector coordinate in mm
NIMG         i  (R)  No. of images over which the reflection is
                     spread. (1=fully recorded spot, 0 = wider than 
                     the maximum requested limit when reflection 
                     prediction done.)
IFIRST       i  (R)  First image for the spot as an offset wrt. the 
                     current image.
IR_FLAG      i  (R)  Recorded flag       
                                  0 = not recorded
                                  1 = full
                                  2 = good partial (<= nwmax images)
                                  3 = too-wide partial
                                  4 = within cusp but part visible
INR_FLAG     i  (R)  Not recorded flag   
                                  0 = recorded
                                  1 = on rotation axis
                                  2 = within cusp
                                  3 = > extended resol limit
                                  4 = not-recorded but still in list
ISPAT        i  (R)  Spatially overlapped flag 1=yes, 0=no
PHI          r  (R)  Phi value
XRL          r  (R)  Reciprocal lattice 'x' coordinate (wrt 
                     laboratory axes)
YRL          r  (R)  Reciprocal lattice 'y' coordinate (wrt 
                     laboratory axes)
ZRL          r  (R)  Reciprocal lattice 'z' coordinate (wrt 
                     laboratory axes)
RLORF        r  (R)  Lorentz factor
IERR         i  (W)  Error return code from prl_addref
-end*/

/*-C:*/
int prl_addref (int mindx, int icrys, int imgnum, 
                        int h, int k, int l, int ksym,
                        float ax1_coord, float ax2_coord,
                        int nimg, int ifirst, int ir_flag, int inr_flag,
                        int spat, float phi, 
                        float xrl, float yrl, float zrl, float rlorf)
/*end*/

/*-Parameters:
int mindx;       Index as returned by prl_init (R)
int icrys;       Crystal number (R)
int imgnum;      Image number (R)
int h;           Reflection 'h' index (R)
int k;           Reflection 'k' index (R)
int l;           Reflection 'l' index (R)
int ksym;        Symmetry no. (0 if indices are as measured) (R)
float ax1_coord; Axis 1 (e.g. 'x') coordinate in mm (R)
float ax2_coord; Axis 2 (e.g. 'y') coordinate in mm (R)
int nimg;        No. of images over which the reflection is spread.
                 (1=fully recorded spot, 0 = wider than the maximum 
                 requested limit when reflection prediction done.)
int ifirst;      First image for the spot as an offset wrt. the 
                 current image.
int ir_flag;     Recorded flag       
                                     0 = not recorded
                                     1 = full
                                     2 = good partial (<= nwmax images)
                                     3 = too-wide partial
                                     4 = within cusp but part visible
                 (R)
int inr_flag;    Not recorded flag   
                                     0 = recorded
                                     1 = on rotation axis
                                     2 = within cusp
                                     3 = > extended resol limit
                                     4 = not-recorded but still in list
                 (R)
int spat;        Spatially overlapped flag 1=yes, 0=no (R)
float phi;       Phi value (R)
float xrl;       Reciprocal lattice 'x' coordinate (wrt laboratory 
                 axes) (R)
float yrl;       Reciprocal lattice 'y' coordinate (wrt laboratory 
                 axes) (R)
float zrl;       Reciprocal lattice 'z' coordinate (wrt laboratory 
                 axes) (R)
float rlorf;     Lorentz factor (R)

Return:  =  0, OK
         = -1, Invalid index given or list not initialised
         = -2, Cannot allocate required memory 
-end*/
{
   PRLreflection refl;

   refl.hkl = two20*(h+512)+two10*(k+512)+l+512;
   refl.ksym = ksym;
   refl.ax1_coord = ax1_coord;
   refl.ax2_coord = ax2_coord;
   refl.phi = phi;
   refl.xrl = xrl;
   refl.yrl = yrl;
   refl.zrl = zrl;
   refl.rlorf = rlorf;
   refl.icrys = icrys;
   refl.imgnum = imgnum;
   refl.nimg = nimg;
   refl.ifirst = ifirst;
   refl.ir_flag = ir_flag;
   refl.inr_flag = inr_flag;
   refl.spat = spat;
   return dml_addrec (mindx, (void *) &refl);
}

/* Fortran binding: prlf_addref */

void prlf_addref (int *mindx, int *icrys, int *imgnum, 
                        int *h, int *k, int *l, int *ksym,
                        float *ax1_coord, float *ax2_coord,
                        int *nimg, int *ifirst, int *ir_flag, int *inr_flag,
                        int *ispat, float *phi, 
                        float *xrl, float *yrl, float *zrl, 
                        float *rlorf, int *ierr)
{
   *ierr = prl_addref (*mindx, *icrys, *imgnum, *h, *k, *l, *ksym,
                                *ax1_coord, *ax2_coord, *nimg, *ifirst,
                                *ir_flag, *inr_flag, *ispat, *phi,
                                *xrl, *yrl, *zrl, *rlorf);
   return;
}

/*-Routine: Update reflection indices - prl_puthkl
The routine prl_puthkl (prlf_puthkl) updates the indices
and symmetry no. for a reflection in the requested reflections list. This
would normally be used to change from measured indices & symmetry=0 to
unique indices with the no. of the symmetry operator used for the
conversion (or vice versa)
-end*/

/*                                            
*******************************
**    prl_puthkl             **
*******************************

Purpose: Update a reflection record indices & symmetry in a reflections list

Author:  John W. Campbell, July 1995


CD-Fortran:
         CALL PRLF_PUTHKL  (MINDX, IREF, IH, IK, IL, KSYM, IERR)
CD-end
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by PRLF_INIT
IREF         i  (R)  The reflection number (from 1 upwards)
IH           i  (R)  Reflection 'h' index
IK           i  (R)  Reflection 'k' index
IL           i  (R)  Reflection 'l' index
KSYM         i  (R)  Symmetry no. (0 if indices are as measured)
IERR         i  (W)  Error return code from prl_puthkl
-end*/

/*-C:*/
int prl_puthkl (int mindx, int iref, int h, int k, int l, int ksym)
/*end*/

/*-Parameters:
int mindx;       Index as returned by prl_init (R)
int iref;        The reflection number (from 1 upwards) (R)
int h;           Reflection 'h' index (R)
int k;           Reflection 'k' index (R)
int l;           Reflection 'l' index (R)
int ksym;        Symmetry no. (0 if indices are as measured) (R)

Return:  =  0, OK
         = -1, Invalid index given, list not initialised or reflection  
               record not present
-end*/
{
   PRLreflection *refl;

   refl = (PRLreflection *) dml_recpointer (mindx, iref);
   refl->hkl = two20*(h+512)+two10*(k+512)+l+512;
   refl->ksym = ksym;
   return dml_putrec (mindx, iref, (void *) refl);
}

/* Fortran binding: prlf_puthkl */

void prlf_puthkl (int *mindx, int *iref, int *h, int *k, int *l, 
                          int *ksym, int *ierr)
{
   *ierr = prl_puthkl (*mindx, *iref, *h, *k, *l, *ksym);
   return;
}

/*-Routine: Update a reflection in a reflections list  - prl_putref
The routine prl_putref (prlf_putref) updates the data for a 
reflection in the requested reflections list.
-end*/

/*                                            
*******************************
**    prl_putref             **
*******************************

Purpose: Update a reflection record in a reflections list

Author:  John W. Campbell, May 1995


CD-Fortran:
         CALL PRLF_PUTREF  (MINDX, IREF, ICRYS, IMGNUM, IH, IK, IL,
        +                          KSYM, AX1_COORD, AX2_COORD, NIMG, IFIRST, 
        +                          IR_FLAG, INR_FLAG, ISPAT,
        +                          PHI, XRL, YRL, ZRL, RLORF, IERR)
CD-end
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by PRLF_INIT
IREF         i  (R)  The reflection number (from 1 upwards)
ICRYS        i  (R)  Crystal number
IMGNUM       i  (R)  Image number
IH           i  (R)  Reflection 'h' index
IK           i  (R)  Reflection 'k' index
IL           i  (R)  Reflection 'l' index
KSYM         i  (R)  Symmetry no. (0 if indices are as measured)
AX1_COORD    r  (R)  Axis 1 (e.g. 'x') detector coordinate in mm
AX2_COORD    r  (R)  Axis 2 (e.g. 'y') detector coordinate in mm
NIMG         i  (R)  No. of images over which the reflection is 
                     spread. (1=fully recorded spot, 0 = wider than 
                     the maximum requested limit when reflection 
                     prediction done.)
IFIRST       i  (R)  First image for the spot as an offset wrt. the 
                     current image.
IR_FLAG      i  (R)  Recorded flag       
                                  0 = not recorded
                                  1 = full
                                  2 = good partial (<= nwmax images)
                                  3 = too-wide partial
                                  4 = within cusp but part visible
INR_FLAG     i  (R)  Not recorded flag   
                                  0 = recorded
                                  1 = on rotation axis
                                  2 = within cusp
                                  3 = > extended resol limit
                                  4 = not-recorded but still in list
ISPAT        i  (R)  Spatially overlapped flag 1=yes, 0=no
PHI          r  (R)  Phi value
XRL          r  (R)  Reciprocal lattice 'x' coordinate (wrt 
                     laboratory axes)
YRL          r  (R)  Reciprocal lattice 'y' coordinate (wrt 
                     laboratory axes)
ZRL          r  (R)  Reciprocal lattice 'z' coordinate (wrt 
                     laboratory axes)
RLORF        r  (R)  Lorentz factor
IERR         i  (W)  Error return code from prl_putref
-end*/

/*-C:*/
int prl_putref (int mindx, int iref, int icrys, int imgnum, 
                        int h, int k, int l, int ksym,
                        float ax1_coord, float ax2_coord,
                        int nimg, int ifirst, int ir_flag, int inr_flag,
                        int spat, float phi, 
                        float xrl, float yrl, float zrl, float rlorf)
/*end*/

/*-Parameters:
int mindx;       Index as returned by prl_init (R)
int iref;        The reflection number (from 1 upwards) (R)
int icrys;       Crystal number (R)
int imgnum;      Image number (R)
int h;           Reflection 'h' index (R)
int k;           Reflection 'k' index (R)
int l;           Reflection 'l' index (R)
int ksym;        Symmetry no. (0 if indices are as measured) (R)
float ax1_coord; Axis 1 (e.g. 'x') coordinate in mm (R)
float ax2_coord; Axis 2 (e.g. 'y') coordinate in mm (R)
int nimg;        No. of images over which the reflection is spread.
                 (1=fully recorded spot, 0 = wider than the maximum 
                 requested limit when reflection prediction done.)
int ifirst;      First image for the spot as an offset wrt. the 
                 current image.
int ir_flag;     Recorded flag       
                                     0 = not recorded
                                     1 = full
                                     2 = good partial (<= nwmax images)
                                     3 = too-wide partial
                                     4 = within cusp but part visible
                 (R)
int inr_flag;    Not recorded flag   
                                     0 = recorded
                                     1 = on rotation axis
                                     2 = within cusp
                                     3 = > extended resol limit
                                     4 = not-recorded but still in list
                 (R)
int spat;        Spatially overlapped flag 1=yes, 0=no (R)
float phi;       Phi value (R)
float xrl;       Reciprocal lattice 'x' coordinate (wrt laboratory 
                 axes) (R)
float yrl;       Reciprocal lattice 'y' coordinate (wrt laboratory 
                 axes) (R)
float zrl;       Reciprocal lattice 'z' coordinate (wrt laboratory 
                 axes) (R)
float rlorf;     Lorentz factor (R)

Return:  =  0, OK
         = -1, Invalid index given, list not initialised or reflection  
               record not present
-end*/
{
   PRLreflection refl;

   refl.hkl = two20*(h+512)+two10*(k+512)+l+512;
   refl.ksym = ksym;
   refl.ax1_coord = ax1_coord;
   refl.ax2_coord = ax2_coord;
   refl.phi = phi;
   refl.xrl = xrl;
   refl.yrl = yrl;
   refl.zrl = zrl;
   refl.rlorf = rlorf;
   refl.icrys = icrys;
   refl.imgnum = imgnum;
   refl.nimg = nimg;
   refl.ifirst = ifirst;
   refl.ir_flag = ir_flag;
   refl.inr_flag = inr_flag;
   refl.spat = spat;
   return dml_putrec (mindx, iref, (void *) &refl);
}

/* Fortran binding: prlf_putref */

void prlf_putref (int *mindx, int *iref, int *icrys, int *imgnum, 
                        int *h, int *k, int *l, int *ksym,
                        float *ax1_coord, float *ax2_coord,
                        int *nimg, int *ifirst, int *ir_flag, int *inr_flag,
                        int *ispat, float *phi, 
                        float *xrl, float *yrl, float *zrl, 
                        float *rlorf, int *ierr)
{
   *ierr = prl_putref (*mindx, *iref, *icrys, *imgnum, *h, *k, *l,
                                *ksym, *ax1_coord, *ax2_coord, *nimg, *ifirst,
                                *ir_flag, *inr_flag, *ispat, *phi,
                                *xrl, *yrl, *zrl, *rlorf);
   return;
}

/*-Section: Get Data from 'PRL' Lists
Routines are available to get the number of reflection records and details
of the individual reflections from a PRL predicted reflections list.
-end*/

/*-Routine: Get number of reflections in a list  - prl_numrefs
The routine prl_numrefs (prlf_numrefs) returns the number of 
reflections currently stored in a reflections list.
-end*/

/*                                            
********************************
**    prl_numrefs             **
********************************

Purpose: Get no. of reflections in reflections list

Author:  John W. Campbell, May 1995

CD-Fortran:
         CALL PRLF_NUMREFS  (MINDX, NUMREFS)
CD-end
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by PRLF_INIT
NUMREFS      i  (W)  No. of reflections in the reflections list. If
                     there is an error a value of -1 will be returned
                     from the prl_numrefs call
-end*/

/*-C:*/
int prl_numrefs (int mindx)
/*end*/

/*-Parameters:
int mindx;       Index as returned by prl_init (R)

Return:  >= 0, The number of reflections
         = -1, Invalid index given or list not initialised
-end*/
{
   return dml_numrecs (mindx);
}

/* Fortran binding: prlf_numrefs */

void prlf_numrefs (int *mindx, int*numrefs)
{
   *numrefs = prl_numrefs (*mindx);
   return;
}

/*-Routine: Get reflection indices for a reflection  - prl_gethkl
The routine prl_gethkl (prlf_gethkl) gets the reflection 
indices and symmetry no. (if set) for a given reflection in a reflection list.
-end*/

/*                                            
********************************
**    prl_gethkl              **
********************************

Purpose: Get reflection indices from a reflections list

Author:  John W. Campbell, July 1995


CD-Fortran:
         CALL PRLF_GETHKL  (MINDX, IREF, ICRYS, IMGNUM, IH, IK, IL,
        +                           KSYM, IERR)
CD-end
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by PRLF_INIT
IREF         i  (R)  The reflection number (from 1 upwards)
ICRYS        i  (W)  Crystal number
IMGNUM       i  (W)  Image number
IH           i  (W)  Reflection 'h' index
IK           i  (W)  Reflection 'k' index
IL           i  (W)  Reflection 'l' index
KSYM         i  (W)  Symmetry no. (0 if indices are as measured)
IERR         i  (W)  Error return code from prl_gethkl
-end*/

/*-C:*/
int prl_gethkl (int mindx, int iref, int *icrys, int *imgnum, 
                        int *h, int *k, int *l, int *ksym)
/*end*/

/*-Parameters:
int mindx;        Index as returned by prl_init (R)
int iref;         Reflection number (from 1 upwards) (R)
int *icrys;       Crystal number (W)
int *imgnum;      Image number (W)
int *h;           Reflection 'h' index (W)
int *k;           Reflection 'k' index (W)
int *l;           Reflection 'l' index (W)
int ksym;         Symmetry no. (0 if indices are as measured) (W)

Return:  =  0, OK
         = -1, Invalid index given, list not initialised or reflection
               record not present.
-end*/
{
   PRLreflection *refl;
   int jj;

   refl = (PRLreflection *) dml_recpointer (mindx, iref);
   if (refl==0) return -1;
   *icrys = refl->icrys;
   *imgnum = refl->imgnum;
   jj = refl->hkl; 
   *h = jj/two20;
   jj -= two20*(*h);
   *h -= 512;
   *k = jj/two10;
   *l = jj - two10*(*k) - 512;
   *k -= 512;
   *ksym = refl->ksym;
   return 0;
}
/* Fortran binding: prlf_gethkl */

void prlf_gethkl (int *mindx, int * iref, int *icrys, int *imgnum, 
                          int *h, int *k, int *l, int *ksym, int *ierr)
{
   *ierr = prl_gethkl (*mindx, *iref, icrys, imgnum, h, k, l, ksym);
   return;
}

/*-Routine: Get reflection details for a reflection  - prl_getref
The routine prl_getref (prlf_getref) gets the main reflection 
details for a given reflection in a reflection list. The remaining details 
may be retrieved via other routines such as prl_getphrc 
(prlf_getphrc).
-end*/

/*                                            
********************************
**    prl_getref              **
********************************

Purpose: Get reflection record details from a reflections list

Author:  John W. Campbell, May 1995


CD-Fortran:
         CALL PRLF_GETREF  (MINDX, IREF, ICRYS, IMGNUM, IH, IK, IL,
        +                           KSYM, AX1_COORD, AX2_COORD, NIMG, IFIRST, 
        +                           IR_FLAG, INR_FLAG, ISPAT, IERR)
CD-end
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by PRLF_INIT
IREF         i  (R)  The reflection number (from 1 upwards)
ICRYS        i  (W)  Crystal number
IMGNUM       i  (W)  Image number
IH           i  (W)  Reflection 'h' index
IK           i  (W)  Reflection 'k' index
IL           i  (W)  Reflection 'l' index
KSYM         i  (W)  Symmetry no. (0 if indices are as measured)
AX1_COORD    r  (W)  Axis 1 (e.g. 'x') detector coordinate in mm
AX2_COORD    r  (W)  Axis 2 (e.g. 'y') detector coordinate in mm
NIMG         i  (W)  No. of images over which the reflection is 
                     spread. (1=fully recorded spot, 0 = wider than 
                     the maximum requested limit when reflection 
                     prediction done.)
IFIRST       i  (W)  First image for the spot as an offset wrt. the 
                     current image.
IR_FLAG      i  (W)  Recorded flag       
                                  0 = not recorded
                                  1 = full
                                  2 = good partial (<= nwmax images)
                                  3 = too-wide partial
                                  4 = within cusp but part visible
INR_FLAG     i  (W)  Not recorded flag   
                                  0 = recorded
                                  1 = on rotation axis
                                  2 = within cusp
                                  3 = > extended resol limit
                                  4 = not-recorded but still in list
ISPAT        i  (W)  Spatially overlapped flag 1=yes, 0=no
IERR         i  (W)  Error return code from prl_getref
-end*/

/*-C:*/
int prl_getref (int mindx, int iref, int *icrys, int *imgnum, 
                        int *h, int *k, int *l, int *ksym,
                        float *ax1_coord, float *ax2_coord, 
                        int * nimg, int * ifirst,
                        int * ir_flag, int *inr_flag, int *spat)
/*end*/

/*-Parameters:
int mindx;        Index as returned by prl_init (R)
int iref;         Reflection number (from 1 upwards) (R)
int *icrys;       Crystal number (W)
int *imgnum;      Image number (W)
int *h;           Reflection 'h' index (W)
int *k;           Reflection 'k' index (W)
int *l;           Reflection 'l' index (W)
int ksym;         Symmetry no. (0 if indices are as measured) (W)
float *ax1_coord; Axis 1 (e.g. 'x') coordinate in mm (W)
float *ax2_coord; Axis 2 (e.g. 'y') coordinate in mm (W)
int *nimg;        No. of images over which the reflection is spread.
                  (1=fully recorded spot, 0 = wider than the maximum 
                  requested limit when reflection prediction done.)
int *ifirst;      First image for the spot as an offset wrt. the 
                  current image.
int *ir_flag;     Recorded flag       
                                     0 = not recorded
                                     1 = full
                                     2 = good partial (<= nwmax images)
                                     3 = too-wide partial
                                     4 = within cusp but part visible
                  (W)
int *inr_flag;    Not recorded flag   
                                     0 = recorded
                                     1 = on rotation axis
                                     2 = within cusp
                                     3 = > extended resol limit
                                     4 = not-recorded but still in list
                  (W)
int *spat;        Spatially overlapped flag 1=yes, 0=no (W)

Return:  =  0, OK
         = -1, Invalid index given, list not initialised or reflection
               record not present.
-end*/
{
   PRLreflection *refl;
   int jj;

   refl = (PRLreflection *) dml_recpointer (mindx, iref);
   if (refl==0) return -1;
   *icrys = refl->icrys;
   *imgnum = refl->imgnum;
   jj = refl->hkl; 
   *h = jj/two20;
   jj -= two20*(*h);
   *h -= 512;
   *k = jj/two10;
   *l = jj - two10*(*k) - 512;
   *k -= 512;
   *ksym = refl->ksym;
   *ax1_coord = refl->ax1_coord;
   *ax2_coord = refl->ax2_coord;
   *nimg = refl->nimg;
   *ifirst = refl->ifirst;
   *ir_flag = refl->ir_flag;
   *inr_flag = refl->inr_flag;
   *spat = refl->spat;
   return 0;
}
/* Fortran binding: prlf_getref */

void prlf_getref (int *mindx, int * iref, int *icrys, int *imgnum, 
                          int *h, int *k, int *l, int *ksym,
                          float *ax1_coord, float *ax2_coord,
                          int *nimg, int *ifirst, int *ir_flag, int *inr_flag,
                          int *ispat, int *ierr)
{
   *ierr = prl_getref (*mindx, *iref, icrys, imgnum, h, k, l, ksym,
                               ax1_coord, ax2_coord, nimg, ifirst,
                               ir_flag, inr_flag, ispat);
   return;
}

/*-Routine: Get further information for a reflection  - prl_getphrc
The routine prl_getphrc (prlf_getphrc) gets the phi value,
reciprocal lattice coordinates and Lorentz facor for a reflection for a 
given reflection in a reflection list. The main details may be retrieved 
via the routine prl_getref (prlf_getref).
-end*/

/*                                            
********************************
**    prl_getphrc             **
********************************

Purpose: Get reflection record details from a reflections list

Author:  John W. Campbell, May 1995


CD-Fortran:
         CALL PRLF_GETPHRC  (MINDX, IREF, PHI,
        +                    XRL, YRL, ZRL, RLORF, IERR)
CD-end
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by PRLF_INIT
IREF         i  (R)  The reflection number (from 1 upwards)
PHI          r  (W)  Phi value
XRL          r  (W)  Reciprocal lattice 'x' coordinate (wrt 
                     laboratory axes)
YRL          r  (W)  Reciprocal lattice 'y' coordinate (wrt 
                     laboratory axes)
ZRL          r  (W)  Reciprocal lattice 'z' coordinate (wrt 
                     laboratory axes)
RLORF        r  (W)  Lorentz factor
IERR         i  (W)  Error return code from prl_getphrc
-end*/

/*-C:*/
int prl_getphrc (int mindx, int iref, float * phi, float * xrl,
                         float *yrl, float * zrl, float *rlorf)
/*end*/

/*-Parameters:
int mindx;       Index as returned by prl_init (R)
int iref;        Reflection number (from 1 upwards) (R)
float *phi;      Phi value (W)
float *xrl;      Reciprocal lattice 'x' coordinate (wrt laboratory 
                 axes) (W)
float *yrl;      Reciprocal lattice 'y' coordinate (wrt laboratory 
                 axes) (W)
float *zrl;      Reciprocal lattice 'z' coordinate (wrt laboratory
                 axes (W)
float *rlorf;    Lorentz factor (W)

Return:  =  0, OK
         = -1, Invalid index given, list not initialised or reflection 
               record not present.
-end*/
{
   PRLreflection *refl;

   refl = (PRLreflection *) dml_recpointer (mindx, iref);
   if (refl==0) return -1;
   *phi = refl->phi;
   *xrl = refl->xrl;
   *yrl = refl->yrl;
   *zrl = refl->zrl;
   *rlorf = refl->rlorf;
   return 0;
}

/* Fortran binding: prlf_getphrc */

void prlf_getphrc (int *mindx, int * iref, float *phi,
                           float *xrl, float *yrl, float *zrl, 
                           float *rlorf, int *ierr)
{
   *ierr = prl_getphrc (*mindx, *iref, phi, xrl, yrl, zrl, rlorf);
   return;
}

/*-Routine: Get coordinates for a reflection  - prl_getcoords
The routine prl_getcoords (prlf_getcoords) gets the 
coordinates in mm from the pattern centre for a reflection for a given 
reflection in a reflection list. Fuller details, including the coordinates,
may be retrieved via the routine prl_getref (prlf_getref).
-end*/

/*                                            
********************************
**    prl_getcoords           **
********************************

Purpose: Get reflection coordinates from a reflections list

Author:  John W. Campbell, May 1995


CD-Fortran:
         CALL PRLF_GETXY  (MINDX, IREF, AX1_COORD, AX2_COORD,
        +                          IERR)
CD-end
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by PRLF_INIT
IREF         i  (R)  The reflection number (from 1 upwards)
AX1_COORD    r  (W)  Axis 1 coordinate
AX2_COORD    r  (W)  Axis 2 coordinate
IERR         i  (W)  Error return code from prl_getphrc
-end*/

/*-C:*/
int prl_getcoords (int mindx, int iref, float * ax1_coord, 
                           float * ax2_coord)
/*end*/

/*-Parameters:
int mindx;        Index as returned by prl_init (R)
int iref;         Reflection number (from 1 upwards) (R)
float *ax1_coord; Axis 1 coordinate in mm (W)
float *ax2_coord; Axis 2 coordinate in mm (W)

Return:  =  0, OK
         = -1, Invalid index given, list not initialised or reflection 
               record not present.
-end*/
{
   PRLreflection *refl;

   refl = (PRLreflection *) dml_recpointer (mindx, iref);
   if (refl==0) return -1;
   *ax1_coord = refl->ax1_coord;
   *ax2_coord = refl->ax2_coord;
   return 0;
}

/* Fortran binding: prlf_getcoords */

void prlf_getcoords (int * mindx, int * iref, float *ax1_coord,
                             float * ax2_coord, int *ierr)
{
   *ierr = prl_getcoords (*mindx, *iref, ax1_coord, ax2_coord);
   return;
}

/*-Section: Delete and Clear 'PRL' Data
Routines enable the data stored in a PRL reflection list to be
deleted for a given image and also for complete lists to be cleared
or removed.
-end*/

/*-Routine: Delete reflections for a given image - prl_delimg
The routine prl_delimg (prlf_delimg) deletes the reflections
for a given image from a reflections list. Unused space at the end of the list
will be automatically freed.
-end*/

/*                                            
******************************
**    prl_delimg            **
******************************

Purpose: Delete the reflections for a given image from a reflections list

Author:  John W. Campbell, May 1995

CD-Fortran:
         CALL PRLF_DELIMG  (MINDX, ICRYS, IMGNUM, IERR)
CD-end
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by PRLF_INIT
ICRYS        i  (R)  Crystal number
IMGNUM       i  (R)  Image number
IERR         i  (W)  Error return code from prl_delimg
-end*/

/*-C:*/
int prl_delimg (int mindx, int icrys, int imgnum)
/*end*/

/*-Parameters:
int mindx;       Index as returned by prl_init (R)
int icrys;       Crystal number (R)
int imgnum;      Image number (R)

Return:  =  0, OK
         = -1, Invalid index given or list not initialised
-end*/
{
   PRLimgcrys client_data;
   int i;

   client_data.icrys = icrys;
   client_data.imgnum = imgnum;
   i = dml_delrecs (mindx, prl_del, (void *) &client_data);
   if (i<0) return i;
   dml_freespace (mindx);
   return 0;
}

/* Fortran binding: prlf_delimg */

void prlf_delimg (int *mindx, int *icrys, int *imgnum, int *ierr)
{
   *ierr = prl_delimg (*mindx, *icrys, *imgnum);
   return;
}

/*-Routine: Clear a reflections list - prl_clear
The routine prl_clear (prlf_clear) clears a reflections list.
The memory at the end of the list is released but the index returned 
from the prl_init (prlf_init) call can still be used and new reflections 
may be added etc. as required. 
-end*/

/*                                            
******************************
**    prl_clear             **
******************************

Purpose: Clear a reflections list (keep memory and list parameters)

Author:  John W. Campbell, May 1995

CD-Fortran:
         CALL PRLF_CLEAR  (MINDX, IERR)
CD-end
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by PRLF_INIT
IERR         i  (W)  Error return code from prl_free
-end*/

/*-C:*/
int prl_clear (int mindx)
/*end*/

/*-Parameters:
int mindx;       Index as returned by prl_init (R)

Return:  =  0, OK
         = -1, Invalid index given or list not initialised
-end*/
{
   dml_clear (mindx);
   return dml_freespace (mindx);
}

/* Fortran binding: prlf_clear */

void prlf_clear (int *mindx, int *ierr)
{
   *ierr = prl_clear (*mindx);
   return;
}

/*-Routine: Clear and free a reflections list - prl_free
The routine prl_free (prlf_free) clears a reflections list
and makes it available for re-use via another initialisation call. All the
memory that was allocated for the list is freed.
-end*/

/*                                            
******************************
**    prl_free              **
******************************

Purpose: Clear a reflections list and free it for re-use

Author:  John W. Campbell, May 1995

CD-Fortran:
         CALL PRLF_FREE  (MINDX, IERR)
CD-end
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by PRLF_INIT
IERR         i  (W)  Error return code from prl_free
-end*/

/*-C:*/
int prl_free (int mindx)
/*end*/

/*-Parameters:
int mindx;       Index as returned by prl_init (R)

Return:  =  0, OK
         = -1, Invalid index given or list not initialised
-end*/
{
   return dml_free (mindx);
}

/* Fortran binding: prlf_free */

void prlf_free (int *mindx, int *ierr)
{
   *ierr = prl_free (*mindx);
   return;
}
/*-Section: Sort 'PRL' Lists
A routine enables PRL reflection lists to be sorted on a number
of the stored parameters.
-end*/

/*-Routine: Sort a reflections list - prl_sort
The routine prl_sort (prlf_sort) sorts a reflections list
in a requested order. The order of any records, with the same sort key value,
is preserved.
-end*/

/*                                            
******************************
**    prl_sort              **
******************************

Purpose: Sort a reflections list.

Author:  John W. Campbell, May 1995

CD-Fortran:
         CALL PRLF_SORT  (MINDX, IORD, IERR)
CD-end
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by PRLF_INIT
IORD         i  (R)  Sort order =1 Sort on hkl (ascending)
                                =2 Sort on ax1_coord (ascending)
                                =3 Sort on ax3_coord (ascending)
IERR         i  (W)  Error return code from prl_sort
-end*/

/*-C:*/
int prl_sort (int mindx, int iord)
/*end*/

/*-Parameters:
int mindx;       Index as returned by prl_init (R)
int iord;        Sort order  =1 Sort on hkl (ascending)
                             =2 Sort on ax1_coord (ascending)
                             =3 Sort on ax3_coord (ascending)
                 (R)

Return:  =  0, OK
         = -1, Invalid index given or list not initialised
         = -2, Invalid sort order requested
-end*/
{
   if (iord<1||iord>3) return -2;
   switch (iord)
   {
      case 1:
      return dml_sort (mindx, prl_sorxx1);

      case 2:
      return dml_sort (mindx, prl_sorxx2);

      case 3:
      return dml_sort (mindx, prl_sorxx3);
   }
   return 0;
}

/* Fortran binding: prlf_sort */

void prlf_sort (int *mindx, int *iord, int *ierr)
{
   *ierr = prl_sort (*mindx, *iord);
   return;
}
/*-Section: Determine Spatial Overlaps
A routine is available to determine and flag spots which spatially overlapped.
-end*/

/*-Routine: Determine spatial overlaps - prl_ovlp1
The routine prl_ovlp1 (prlf_ovlp1) determines whether 
reflections are spatially overlapped on an image and sets the required
spatial overlap flags internally. This overlap determination assumes
circular spots of constant diameter througout the image. All spatial overlap 
flags for the requested image will be cleared before the overlap calculation is
performed. Note also that the routine sorts the reflection data in order
of ascending values for the second detector coordinate.
-end*/

/*                                            
******************************
**    prl_ovlp1             **
******************************

Purpose: Find spatial overlaps for a requested image assuming circular spots

Author:  John W. Campbell, May 1995

CD-Fortran:
         CALL PRLF_OVLP1  (MINDX, ICRYS, IMGNUM, SPOTD, 
        +                  NREF, NSPAT, IERR)
CD-end
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by PRLF_INIT
ICRYS        i  (R)  Crystal number
IMGNUM       i  (R)  Image number
SPOTD        r  (R)  The spot diameter im mm
NREF         i  (W)  No. of reflections for the requested image
NSPAT        i  (W)  No. of spatially overlapped reflections for the 
                     requested image
IERR         i  (W)  Error return code from prl_ovlp1
-end*/

/*-C:*/
int prl_ovlp1 (int mindx, int icrys, int imgnum, float spotd,
                       int *nref, int *nspat)
/*end*/

/*-Parameters:
int mindx;       Index as returned by prl_init (R)
int icrys;       Crystal number for the required image (R)
int imgnum;      Image number for the required image (R)
float spotd;     Spot diameter im mm (R)
int *nref;       No. of reflections for the requested image (W)
int *nspat;      No. of spatially overlapped reflections for the requested
                 image (W)

Return:  =  0, OK
         = -1, Invalid index given or list not initialised
         = -2, Cannot allocate memory for temporary work array
-end*/
{
   int nr;
   int ns;
   int num_strips;
   int strip;
   PRLreflection *refl;
   int i;
   int j;
   int i1;
   int i2;
   int n;
   int *indx;
   float x_start;
   float x_end;
   float x_step;
   float xmin;
   float xmax;
   float xdiff;
   float ydiff;
   float d2;
   float spd2;

   *nref = 0;
   *nspat = 0;

   spd2 = spotd*spotd;

   /* Sort data */

   i = prl_sort (mindx, 3);
   if (i<0) return i;

   nr = prl_numrefs (mindx);
   if (nr==0) return 0;

   refl = (PRLreflection *) dml_data (mindx);

   /* Clear spatial overlap flags for requested image */

   ns = 0;
   x_end = -1.0e20;
   x_start = 1.0e20;
   for (i=0;i<nr;++i)
   {
      if (refl[i].icrys!=icrys||refl[i].imgnum!=imgnum) continue;
      refl[i].spat = 0;
      if (refl[i].ax1_coord<x_start) x_start = refl[i].ax1_coord;
      if (refl[i].ax1_coord>x_end) x_end = refl[i].ax1_coord;
      ns++;
   }
   *nref = ns;
  
   indx = (int *) malloc (ns*sizeof(int));
   if (indx==0) return -2;

   /* Determine spatial overlaps */
   num_strips = ns/(MAXSPOTS_IN_STRIP/3.0) + 2;
   x_step = (x_end-x_start)/num_strips;
   xmin = x_start - spotd - x_step;
   xmax = xmin + x_step + 2.0*spotd;
   for (strip=0;strip<num_strips;++strip)
   {
      xmin += x_step;
      xmax += x_step; 
      n = 0;
      for (i=0;i<nr;++i)
      {
         if (refl[i].icrys!=icrys||refl[i].imgnum!=imgnum) continue;
         if (refl[i].ax1_coord<xmin) continue;
         if (refl[i].ax1_coord>xmax) continue;
         indx[n] = i;
         n++;
      }
      if (n<=1) continue;
      for (i=0;i<n-1;++i)
      {
         for (j=i+1;j<n;++j)
	 {
            i1 = indx[i];
            i2 = indx[j];
            xdiff = refl[i1].ax1_coord-refl[i2].ax1_coord;
            if (xdiff<0.0) xdiff = -xdiff;
            ydiff = refl[i1].ax2_coord-refl[i2].ax2_coord;
            if (ydiff<0.0) ydiff = -ydiff;
            if (ydiff>spotd) break;
            if (xdiff<=spotd)
	    {
               d2 = xdiff*xdiff + ydiff*ydiff;
               if (d2<=spd2)
	       {
                  if (refl[i1].spat==0){refl[i1].spat = 1; *nspat += 1;}
                  if (refl[i2].spat==0){refl[i2].spat = 1; *nspat += 1;}
               }
            }
         }  
      }
   }
   free(indx);
   return 0;
}

/* Fortran binding: prlf_ovlp1 */

void prlf_ovlp1 (int *mindx, int *icrys, int *imgnum, float *spotd,
                         int *nref, int*nspat, int *ierr)
{
   *ierr = prl_ovlp1 (*mindx, *icrys, *imgnum, *spotd, nref, nspat);
   return;
}

/****************************************************************************/
/*               S E R V I C E       R O U T I N E S                        */
/****************************************************************************/


/*                                            
******************************
**    prl_del               **
******************************

Purpose: 'User supplied'  delete function for a given image from a 
         reflections list

Return:  =  0, Do not delete reflection
         =  1, Delete reflection

Author:  John W. Campbell, May 1995

**** SERVICE ROUTINE ONLY ****

*/

int prl_del (void * rec, void * client_data)
{
   PRLreflection *refl;
   PRLimgcrys *imgc;

   refl = (PRLreflection *) rec;
   imgc = (PRLimgcrys *) client_data;
   if (refl->icrys==imgc->icrys&&refl->imgnum==imgc->imgnum) return 1;
   return 0;
}

/*                                            
******************************
**    prl_sorxx1            **
******************************

Purpose: 'User supplied'  sort comparison function for 'hkl' based sort. 

Return:  (For as needed by dml_sort)
         
Author:  John W. Campbell, May 1995

**** SERVICE ROUTINE ONLY ****

*/

int prl_sorxx1 (void * rec1, void * rec2, int * i1, int * i2)
{
   PRLreflection *refl1, *refl2;

   refl1 = (PRLreflection *) rec1;
   refl2 = (PRLreflection *) rec2;
   if (refl1->hkl>refl2->hkl) return 1;
   if (refl1->hkl<refl2->hkl) return -1;
   if (*i1>*i2) return 1;
   if (*i1<*i2) return -1;
   return 0;
}

/*                                            
******************************
**    prl_sorxx2            **
******************************

Purpose: 'User supplied'  sort comparison function for 'ax1_coord' based sort. 

Return:  (For as needed by dml_sort)
         
Author:  John W. Campbell, May 1995

**** SERVICE ROUTINE ONLY ****

*/

int prl_sorxx2 (void * rec1, void * rec2, int * i1, int * i2)
{
   PRLreflection *refl1, *refl2;

   refl1 = (PRLreflection *) rec1;
   refl2 = (PRLreflection *) rec2;
   if (refl1->ax1_coord>refl2->ax1_coord) return 1;
   if (refl1->ax1_coord<refl2->ax1_coord) return -1;
   if (*i1>*i2) return 1;
   if (*i1<*i2) return -1;
   return 0;
}

/*                                            
******************************
**    prl_sorxx3            **
******************************

Purpose: 'User supplied'  sort comparison function for 'ax2_coord' based sort. 

Return:  (For as needed by dml_sort)
         
Author:  John W. Campbell, May 1995

**** SERVICE ROUTINE ONLY ****

*/

int prl_sorxx3 (void * rec1, void * rec2, int * i1, int * i2)
{
   PRLreflection *refl1, *refl2;

   refl1 = (PRLreflection *) rec1;
   refl2 = (PRLreflection *) rec2;
   if (refl1->ax2_coord>refl2->ax2_coord) return 1;
   if (refl1->ax2_coord<refl2->ax2_coord) return -1;
   if (*i1>*i2) return 1;
   if (*i1<*i2) return -1;
   return 0;
}
