/*===========================================*/
/*       UNQ Reflection list   routines      */
/*===========================================*/

/*-Title: UNIQUE REFLECTION LIST ROUTINES
-end*/

/*-Intro:
These are a series of routines concerned with the allocation
and management of memory for lists of reflection records to be 
used in conjunction with the set of unique data reflection prediction 
routines which are part of the Reflection Prediction Function (RPF) set 
routines. The routines make use of the Dynamic Memory List (DML) set of 
management routines for in-memory record lists. The basic purpose of such
lists is to enable predicted or observed data to be analysed in terms of
the coverage of the unique data and varoius counts may be accumulated
to facilitate this.
-end*/

/*Routines:
The following routines are available:

     unq_reflist_init         Allocate and initialise a uniquereflections list
     unq_reflist_addref       Add a reflection to a unique reflections list
     unq_reflist_numrefs      Get number of reflections in a list
     unq_reflist_clear_counts Clear counts in list
     unq_reflist_inc          Increment counts in a unique reflections list
     unq_reflist_hkl          Get hkl for a reflection
     unq_reflist_xyz          Get reciprocal lattice xyz for a reflection
     unq_reflist_get          Get counts details for a reflection
     unq_reflist_clear        Clear a unique reflections list
     unq_reflist_free         Clear and free a unique reflections list
     unq_reflist_sorthkl      Sort a unique reflections list on hkl
     unq_reflist_sortlhk      Sort a unique reflections list on lhk

     The equivalent Fortran interface routines are named unqf_reflist...
*/

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#include <stdio.h>

#include "dml.h"

#include "unq_systyp.h"
#include "unq.h"

/* Define Fortran binding names */

#if    LINKTYP == 1

#define unqf_reflist_init unqf_reflist_init_
#define unqf_reflist_addref unqf_reflist_addref_ 
#define unqf_reflist_numrefs unqf_reflist_numrefs_
#define unqf_reflist_clear_counts unqf_reflist_clear_counts_
#define unqf_reflist_inc unqf_reflist_inc_
#define unqf_reflist_hkl unqf_reflist_hkl_
#define unqf_reflist_xyz unqf_reflist_xyz_
#define unqf_reflist_get unqf_reflist_get_
#define unqf_reflist_clear unqf_reflist_clear_
#define unqf_reflist_free unqf_reflist_free_
#define unqf_reflist_sorthkl unqf_reflist_sorthkl_
#define unqf_reflist_sortlhk unqf_reflist_sortlhk_

#endif

#if  LINKTYP == 2

#define unqf_reflist_init UNQF_REFLIST_INIT
#define unqf_reflist_addref UNQF_REFLIST_ADDREF 
#define unqf_reflist_numrefs UNQF_REFLIST_NUMREFS
#define unqf_reflist_clear_counts UNQF_REFLIST_CLEAR_COUNTS
#define unqf_reflist_inc UNQF_REFLIST_INC
#define unqf_reflist_hkl UNQF_REFLIST_HKL
#define unqf_reflist_xyz UNQF_REFLIST_XYZ
#define unqf_reflist_get UNQF_REFLIST_GET
#define unqf_reflist_clear UNQF_REFLIST_CLEAR
#define unqf_reflist_free UNQF_REFLIST_FREE
#define unqf_reflist_sorthkl UNQF_REFLIST_SORTHKL
#define unqf_reflist_sortlhk UNQF_REFLIST_SORTLHK

#endif

void unqf_reflist_init();
void unqf_reflist_addref();
void unqf_reflist_numrefs();
void unqf_reflist_clear_counts();
void unqf_reflist_inc();
void unqf_reflist_hkl();
void unqf_reflist_xyz();
void unqf_reflist_get();
void unqf_reflist_clear();
void unqf_reflist_free();
void unqf_reflist_sorthkl();
void unqf_reflist_sortlhk();

typedef struct _UNQreflection
{
   short h;                    /* 'h' index */
   short k;                    /* 'k' index */
   short l;                    /* 'l' index */
   float xrl;                  /* Reciprocal lattice 'x' coordinate wrt 
                                  standard orthogonalisation */
   float yrl;                  /* Reciprocal lattice 'y' coordinate wrt 
                                  standardorthogonalisation */
   float zrl;                  /* Reciprocal lattice 'z' coordinate wrt 
                                  standard orthogonalisation */
   short xpix;                 /* x pixel position for xdl_unq_sim */
   short ypix;                 /* y pixel position for xdl_unq_sim */

                               /* Following counts are for NON-SPATIALLY
                                  OVERLAPPED reflections */
   unsigned short n_full_p;    /* Count number of full reflection measurements 
                                  +h+k+l */
   unsigned short n_full_m;    /* Count number of full reflection measurements 
                                 -h-k-l */
   unsigned short n_part2_p;   /* Count number of 2 image partial reflection 
                                  measurements +h+k+l */
   unsigned short n_part2_m;   /* Count number of 2 image partial reflection 
                                  measurements -h-k-l */
   unsigned short n_part3_p;   /* Count number of 3 image partial reflection 
                                  measurements +h+k+l */
   unsigned short n_part3_m;   /* Count number of 3 image partial reflection 
                                  measurements -h-k-l */
   unsigned short n_partn_p;   /* Count no. of partial reflection measurements
                                  (more than 3 images but less than a 
                                  requested maximum no. of images) +h+k+l */
   unsigned short n_partn_m;   /* Count no. of partial reflection measurements
                                  (more than 3 images but less that a 
                                  requested  maximum no. of images) -h-k-l */
   unsigned short n_unk_p;     /* Count number of unknown (e.g from merged
                                  MTZ file) reflection measurements 
                                  +h+k+l */
   unsigned short n_unk_m;     /* Count number of unknown (e.g from merged
                                  MTZ file) reflection measurements 
                                  -h-k-l */
   unsigned short n_unk_u;     /* Count number of unknown (e.g from merged
                                  MTZ file) reflection measurements of
                                  unknown sign */

   unsigned char n_full_sp_p;  /* Counts as above but for spatial overlaps */
   unsigned char n_full_sp_m;  /* Counts as above but for spatial overlaps */
   unsigned char n_part2_sp_p; /* Counts as above but for spatial overlaps */
   unsigned char n_part2_sp_m; /* Counts as above but for spatial overlaps */
   unsigned char n_part3_sp_p; /* Counts as above but for spatial overlaps */
   unsigned char n_part3_sp_m; /* Counts as above but for spatial overlaps */
   unsigned char n_partn_sp_p; /* Counts as above but for spatial overlaps */
   unsigned char n_partn_sp_m; /* Counts as above but for spatial overlaps */
}UNQreflection;


/*-Section: Creating and Writing Unique Data Lists
Routines are available to create new unique reflection lists, to add records
to such lists, to return the number of records in the list  and to increment 
counts held in such lists. The counts may be re-initialised when required so 
that data may be collected for another analysis.
-end*/


/*-Routine: Initialise a unique reflections list  - unq_reflist_init
The routine unq_reflist_init (unqf_reflist_init) checks to see if a spare 
slot is available and, if so, allocates the memory for a new unique 
reflections list with an initial allocation of records. The routine returns 
an index (handle) to be used in all other routines for accessing that list.
-end*/

/*                                            
******************************
**    unq_reflist_init      **
******************************

Purpose: Allocate and initialise a unique reflections list (For UNQ functions)

Author:  John W. Campbell, July 1995

CD-Fortran:
         CALL UNQF_REFLIST_INIT  (INIT_ALLOC, INT_ALLOC, MINDX)
CD-end
*/
/*-Parameters:
INIT_ALLOC    i  (R)  Initial allocation required (no. of reflection
                      records).
INC_ALLOC     i  (R)  No. of records by which to extend the reflections 
                      list each time an extension is required.
MINDX         i  (W)  Index for the allocated reflection list (>=0)
                      A negative return value indicates an error and
                      the value is the return code from the routine
                      unq_reflist_init.
-end*/

/*-C:*/
int unq_reflist_init (int init_alloc, int inc_alloc)
/*end*/

/*-Parameters:
int init_alloc;      No. of records to be allocated initially (R)
int inc_alloc;       No. of records to be added each time an extension
                     to the list is required (R)

Return:  >=0, the index of the allocated list
          -1, no spare lists available
          -2, cannot allocate initial memory
          -3, init_alloc<1 or inc_alloc<1
-end*/

{
   return dml_init (init_alloc, inc_alloc, sizeof(UNQreflection));
}

/* Fortran binding: unqf_reflist_init */

void unqf_reflist_init (int * init_alloc, int * inc_alloc, int *mindx)
{
   *mindx = unq_reflist_init (*init_alloc, *inc_alloc);
   return;
}


/*-Routine: Add a reflection to the list  - unq_reflist_addref
The routine unq_reflist_addref (unqf_reflist_addref) adds a new reflection to 
the end of the requested unique reflections list providing that the list has 
been initialised and that sufficient memory is available or can be allocated 
for it.
-end*/

/*                                            
*******************************
**    unq_reflist_addref     **
*******************************

Purpose: Add a reflection record to a reflections list

Author:  John W. Campbell, July 1995


CD-Fortran:
         CALL UNQF_REFLIST_ADDREF  (MINDX, IH, IK, IL, XRL, YRL, ZRL, 
        +                           IERR)
CD-end
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by UNQF_REFLIST_INIT
IH           i  (R)  Reflection 'h' index
IK           i  (R)  Reflection 'k' index
IL           i  (R)  Reflection 'l' index
XRL          r  (R)  Reciprocal lattice 'x' coordinate (wrt 
                     laboratory axes)
YRL          r  (R)  Reciprocal lattice 'y' coordinate (wrt 
                     laboratory axes)
ZRL          r  (R)  Reciprocal lattice 'z' coordinate (wrt 
                     laboratory axes)
IERR         r  (W)  Error return code from unq_reflist_addref
-end*/

/*-C:*/
int unq_reflist_addref (int mindx,  int h, int k, int l,  float xrl, 
                        float yrl, float zrl)
/*end*/

/*-Parameters:
int mindx;       Index as returned by unq_reflist_init (R)
int h;           Reflection 'h' index (R)
int k;           Reflection 'k' index (R)
int l;           Reflection 'l' index (R)
float xrl;       Reciprocal lattice 'x' coordinate (wrt laboratory 
                 axes) (R)
float yrl;       Reciprocal lattice 'y' coordinate (wrt laboratory 
                 axes) (R)
float zrl;       Reciprocal lattice 'z' coordinate (wrt laboratory 
                 axes) (R)

Return:  =  0, OK
         = -1, Invalid index given or list not initialised
         = -2, Cannot allocate required memory 
-end*/
{
   UNQreflection refl;

   refl.h = h;
   refl.k = k;
   refl.l = l;
   refl.xrl = xrl;
   refl.yrl = yrl;
   refl.zrl = zrl;
   refl.n_full_p = 0;
   refl.n_full_m = 0;
   refl.n_part2_p = 0;
   refl.n_part2_m = 0;
   refl.n_part3_p = 0;
   refl.n_part3_m = 0;
   refl.n_partn_p = 0;
   refl.n_partn_m = 0;
   refl.n_unk_p = 0;
   refl.n_unk_m = 0;
   refl.n_unk_u = 0;
   refl.n_full_sp_p = 0;
   refl.n_full_sp_m = 0;
   refl.n_part2_sp_p = 0;
   refl.n_part2_sp_m = 0;
   refl.n_part3_sp_p = 0;
   refl.n_part3_sp_m = 0;
   refl.n_partn_sp_p = 0;
   refl.n_partn_sp_m = 0;
   return dml_addrec (mindx, (void *) &refl);
}

/* Fortran binding: unqf_reflist_addref */

void unqf_reflist_addref (int *mindx, int *ih, int *ik, int *il,
                          float *xrl, float*yrl, float*zrl, int *ierr)
{
   *ierr = unq_reflist_addref (*mindx, *ih, *ik, *il, *xrl, *yrl, *zrl);
   return;
}

/*-Routine: Get number of reflections in a list  - unq_reflist_numrefs
The routine unq_reflist_numrefs (unqf_reflist_numrefs) returns the number of 
reflections currently stored in a unique reflections list.
-end*/

/*                                            
********************************
**    unq_reflist_numrefs     **
********************************

Purpose: Get no. of reflections in a unique reflections list

Author:  John W. Campbell, July 1995

CD-Fortran:
         CALL UNQF_REFLIST_NUMREFS  (MINDX, NUMREFS)
CD-end:
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by UNQF_REFLIST_INIT
NUMREFS      i  (W)  No. of reflections in the reflections list. If
                     there is an error a value of -1 will be returned
                     from the unq_reflist_numrefs call
-end*/

/*-C:*/
int unq_reflist_numrefs (int mindx)
/*end*/

/*-Parameters:
int mindx;       Index as returned by unq_reflist_init (R)

Return:  >= 0, The number of reflections
         = -1, Invalid index given or list not initialised
-end*/
{
   return dml_numrecs (mindx);
}

/* Fortran binding: unqf_reflist_numrefs */

void unqf_reflist_numrefs (int *mindx, int*numrefs)
{
   *numrefs = unq_reflist_numrefs (*mindx);
   return;
}

/*-Routine: Clear counts in a list  - unq_reflist_clear_counts
The routine unq_reflist_clear_counts (unqf_reflist_clear_counts) clears
(sets to zero) all the stored counts in a unique reflections list.
-end*/

/*                                            
************************************
**    unq_reflist_clear_counts    **
************************************

Purpose: Clear counts in a unique reflections list

Author:  John W. Campbell, July 1995

CD-Fortran:
         CALL UNQF_REFLIST_CLEAR_COUNTS (MINDX, IERR)
CD-end:
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by UNQF_REFLIST_INIT
IERR         i  (R)  Return code from unq_reflist_clear_counts
-end*/

/*-C:*/
int unq_reflist_clear_counts (int mindx)
/*end*/

/*-Parameters:
int mindx;       Index as returned by unq_reflist_init (R)

Return:  =  0, OK
         = -1, Invalid index given or list not initialised
-end*/
{
   UNQreflection *refl;
   int ir;
   int n;

   n = dml_numrecs (mindx);
   if (n<0) return n;
   if (n==0) return 0;
   for (ir=1;ir<=n;++ir)
   {
      refl = (UNQreflection *) dml_recpointer (mindx, ir);
      if (refl==0) return -1;
      refl->n_full_p = 0;
      refl->n_full_m = 0;
      refl->n_part2_p = 0;
      refl->n_part2_m = 0;
      refl->n_part3_p = 0;
      refl->n_part3_m = 0;
      refl->n_partn_p = 0;
      refl->n_partn_m = 0;
      refl->n_unk_p = 0;
      refl->n_unk_m = 0;
      refl->n_unk_u = 0;
      refl->n_full_sp_p = 0;
      refl->n_full_sp_m = 0;
      refl->n_part2_sp_p = 0;
      refl->n_part2_sp_m = 0;
      refl->n_part3_sp_p = 0;
      refl->n_part3_sp_m = 0;
      refl->n_partn_sp_p = 0;
      refl->n_partn_sp_m = 0;
   }
   return 0;
}

/* Fortran binding: unqf_reflist_clear_counts */

void unqf_reflist_clear_counts (int *mindx, int*ierr)
{
   *ierr = unq_reflist_clear_counts (*mindx);
   return;
}

/*-Routine: Update reflection counts in a unique list  - unq_reflist_inc
The routine unq_reflist_inc (unqf_reflist_inc) updates the counts data for a 
reflection in the requested unique reflections list. The parameters indicate
the classes to which the reflection belongs so that the appropriate counts
may be updated.
-end*/

/*                                            
*******************************
**    unq_reflist_inc        **
*******************************

Purpose: Increment counts is a unique reflections list

Author:  John W. Campbell, July 1995


CD-Fortran:
         CALL UNQF_REFLIST_INC  (MINDX, IREF, IPART, ISIGN, ISPAT, IERR)
CD-end:
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by UNQF_REFLIST_INIT
IREF         i  (R)  The reflection number (from 1 upwards)
IPART        i  (R)  Full/partial flag =1 full
                                       =2 partial over 2 images
                                       =3 partial over 3 images
                                       >3 partial over >3 images 
                                          probably up to some user
                                          defined maximum 
                                       =0 unknown (e.g. from merged 
                                          MTZ file)
ISIGN        i  (R)  +1 for +h+k+l, -1 for -h-k-l class 
                     (0=unknown - only allowed if ipart=0)
ISPAT        i  (R)  =0 not spatially overlapped spot
                     =1 spatially overlapped spot
                     (ignored for 'unknown' full/partial class of 
                      reflections)
IERR         i  (W)  Error return code from unq_reflist_inc
-end*/

/*-C:*/
int unq_reflist_inc (int mindx, int iref, int ipart, int isign, int ispat) 
/*end*/

/*-Parameters:
int mindx;       Index as returned by unq_reflist_init (R)
int iref;        The reflection number (from 1 upwards) (R)
int ipart;       Full/partial flag =1 full
                                   =2 partial over 2 images
                                   =3 partial over 3 images
                                   >3 partial over >3 images 
                                      probably up to some user
                                      defined maximum 
                                   =0 unknown (e.g. from merged MTZ file)
                                   (R)
int isign;       +1 for +h+k+l, -1 for -h-k-l class 
                 (0=unknown - only allowed if ipart=0) (R)
int ispat;       =0 not spatially overlapped spot
                 =1 spatially overlapped spot 
                 (ignored for 'unknown' full/partial class of 
                  reflections) (R)

Return:  =  0, OK
         = -1, Invalid index given, list not initialised or reflection  
               record not present
         =  1, Invalid parameters in call
-end*/
{
   UNQreflection *refl;

   refl = (UNQreflection *) dml_recpointer (mindx, iref);
   if (refl==0) return -1;
   if (ipart<0) return 1;
   if (isign!=1&&isign!=-1&&isign!=0) return 1;
   if (isign==0&&ipart!=0) return 1;
   if (ispat!=0&&ispat!=1) return 1;
   if (ispat==0||ipart==0)
   {
      if (ipart==0)
      {
         if (isign==0)
	 {
            refl->n_unk_u +=1;
	 }
         else if (isign>0)
	 {
            refl->n_unk_p +=1;
	 }
         else if (isign<0)
	 {
            refl->n_unk_m +=1;
	 }
      }
      if (ipart==1)
      {
         if (isign>0) {refl->n_full_p +=1;}else{refl->n_full_m +=1;}
      }
      else if (ipart==2)
      {
         if (isign>0) {refl->n_part2_p +=1;}else{refl->n_part2_m +=1;}
      }
      else if (ipart==3)
      {
         if (isign>0) {refl->n_part3_p +=1;}else{refl->n_part3_m +=1;}
      }
      else if (ipart>3)
      {
         if (isign>0) {refl->n_partn_p +=1;}else{refl->n_partn_m +=1;}
      }
   }
   else
   {
      if (ipart==1)
      {
         if (isign>0) {refl->n_full_sp_p +=1;}else{refl->n_full_sp_m +=1;}
      }
      else if (ipart==2)
      {
         if (isign>0) {refl->n_part2_sp_p +=1;}else{refl->n_part2_sp_m +=1;}
      }
      else if (ipart==3)
      {
         if (isign>0) {refl->n_part3_sp_p +=1;}else{refl->n_part3_sp_m +=1;}
      }
      else if (ipart>3)
      {
         if (isign>0) {refl->n_partn_sp_p +=1;}else{refl->n_partn_sp_m +=1;}
      }
   }
   return dml_putrec (mindx, iref, (void *) refl);
}

/* Fortran binding: unqf_reflist_inc */

void unqf_reflist_inc (int *mindx, int *iref, int *ipart, int *isign,
                       int * ispat, int * ierr) 
{
   *ierr = unq_reflist_inc (*mindx, *iref, *ipart, *isign, *ispat);
   return;
}
/*-Section: Get Reflection Details from a List
Various routines are available to extract data for reflections in a unique
data reflections list so that such data may be analysed
-end*/

/*-Routine: Get reflection indices  - unq_reflist_hkl
The routine unq_reflist_hkl (unqf_reflist_hkl) gets the indices for a given 
reflection in a unique reflections list.
-end*/

/*                                            
********************************
**    unq_reflist_hkl         **
********************************

Purpose: Get reflection indices

Author:  John W. Campbell, July 1995


CD-Fortran:
         CALL UNQF_REFLIST_HKL  (MINDX, IREF, IH, IK, IL, IERR)
CD-end:
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by UNQF_REFLIST_INIT
IREF         i  (R)  The reflection number (from 1 upwards)
IH           i  (W)  'h' index
IK           i  (W)  'k' index
IL           i  (W)  'l' index
IERR         i  (W)  Error return code from unq_reflist_hkl
-end*/

/*-C:*/
int unq_reflist_hkl (int mindx, int iref, int *h, int*k, int*l)
/*end*/

/*-Parameters:
int mindx;        Index as returned by unq_reflist_init (R)
int iref;         Reflection number (from 1 upwards) (R)
int *h;           'h' index (W)
int *k;           'k' index (W)
int *l;           'l' index (W)

Return:  =  0, OK
         = -1, Invalid index given, list not initialised or reflection
               record not present.
-end*/
{
   UNQreflection *refl;

   refl = (UNQreflection *) dml_recpointer (mindx, iref);
   if (refl==0) return -1;
   *h = refl->h;
   *k = refl->k;
   *l = refl->l;
   return 0;
}
/* Fortran binding: unqf_reflist_hkl */

void unqf_reflist_hkl (int *mindx, int * iref, int*h, int *k, int *l, 
                       int *ierr)
{
   *ierr = unq_reflist_hkl (*mindx, *iref, h, k, l);
   return;
}
/*-Routine: Get reciprocal lattice coordinates  - unq_reflist_xyz
The routine unq_reflist_xyz (unqf_reflist_xyz) gets the reciprocal lattice
coordinates for a given reflection in a unique reflections list.
-end*/

/*                                            
********************************
**    unq_reflist_xyz         **
********************************

Purpose: Get reflection reciprocal lattice coordinates

Author:  John W. Campbell, July 1995

CD-Fortran:
         CALL UNQF_REFLIST_XYZ  (MINDX, IREF, XR, YR, ZR, IERR)
CD-end:
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by UNQF_REFLIST_INIT
IREF         i  (R)  The reflection number (from 1 upwards)
XR           r  (W)  Reciprocal lattice 'xr' coordinate
YR           r  (W)  Reciprocal lattice 'yr' coordinate
ZR           r  (W)  Reciprocal lattice 'zr' coordinate
IERR         r  (W)  Error return code from unq_reflist_hkl
-end*/

/*-C:*/
int unq_reflist_xyz (int mindx, int iref, float *xr, float *yr, float *zr)
/*end*/

/*-Parameters:
int mindx;        Index as returned by unq_reflist_init (R)
int iref;         Reflection number (from 1 upwards) (R)
float *xr;        Reciprocal lattice 'xr' coordinate (W)
float *yr;        Reciprocal lattice 'yr' coordinate (W)
float *zr;        Reciprocal lattice 'zr' coordinate(W)

Return:  =  0, OK
         = -1, Invalid index given, list not initialised or reflection
               record not present.
-end*/
{
   UNQreflection *refl;

   refl = (UNQreflection *) dml_recpointer (mindx, iref);
   if (refl==0) return -1;
   *xr = refl->xrl;
   *yr = refl->yrl;
   *zr = refl->zrl;
   return 0;
}
/* Fortran binding: unqf_reflist_xyz */

void unqf_reflist_xyz (int *mindx, int * iref, float *xr, float *yr, 
                       float *zr, int *ierr)
{
   *ierr = unq_reflist_xyz (*mindx, *iref, xr, yr, zr);
   return;
}

/*-Routine: Get reflection counts - unq_reflist_get
The routine unq_reflist_get (unqf_reflist_get) gets the counts for a given 
reflection in a unique reflections list.
-end*/

/*                                            
********************************
**    unq_reflist_get         **
********************************

Purpose: Get reflection counts

Author:  John W. Campbell, July 1995


CD-Fortran:
         CALL UNQF_REFLIST_GET  (MINDX, IREF, NFULL_P, NFULL_M,
        +                        NPART2_P, NPART2_M, NPART3_P, NPART3_M,
        +                        NPARTN_P, NPARTN_M, NUNK_P, NUNK_M,
        +                        NUNK_U, NSFULL_P, NSFULL_M,
        +                        NSPART2_P, NSPART2_M, NSPART3_P, NSPART3_M,
        +                        NSPARTN_P, NSPARTN_M, IERR)
CD-end:
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by UNQF_REFLIST_INIT
IREF         i  (R)  The reflection number (from 1 upwards)
NFULL_P      i  (W)  Count: fulls +h+k+l (non-spatially overlapped)
NFULL_M      i  (W)  Count: fulls -h-k-l (non-spatially overlapped)
NPART2_P     i  (W)  Count: 2 image partials +h+k+l 
                     (non-spatially overlapped)
NPART2_M     i  (W)  Count: 2 image partials -h-k-l 
                     (non-spatially overlapped)
NPART3_P     i  (W)  Count: 3 image partials +h+k+l 
                     (non-spatially overlapped)
NPART3_M     i  (W)  Count: 3 image partials -h-k-l 
                     (non-spatially overlapped)
NPARTN_P     i  (W)  Count: >3 but OK image partials +h+k+l 
                     (non-spatially overlapped)
NPARTN_M     i  (W)  Count: >3 but OK image partials -h-k-l 
                     (non-spatially overlapped)
NUNK_P       i  (W)  Count: unknown (e.g. merged) +h+k+l 
NUNK_M       i  (W)  Count: unknown (e.g. merged) -h-k-l 
NUNK_U       i  (W)  Count: unknown (e.g. merged) and unknown sign 
NSFULL_P     i  (W)  Count: fulls +h+k+l (spatially overlapped)
NSFULL_M     i  (W)  Count: fulls -h-k-l (spatially overlapped)
NSPART2_P    i  (W)  Count: 2 image partials +h+k+l 
                     (spatially overlapped)
NSPART2_M    i  (W)  Count: 2 image partials -h-k-l 
                     (spatially overlapped)
NSPART3_P    i  (W)  Count: 3 image partials +h+k+l 
                     (spatially overlapped)
NSPART3_M    i  (W)  Count: 3 image partials -h-k-l 
                     (spatially overlapped)
NSPARTN_P    i  (W)  Count: >3 but OK image partials +h+k+l 
                     (spatially overlapped)
NSPARTN_M    i  (W)  Count: >3 but OK image partials -h-k-l 
                     (spatially overlapped)
IERR         i  (W)  Error return code from unq_reflist_get
-end*/

/*-C:*/
int unq_reflist_get (int mindx, int iref, 
                     int *nfull_p, int *nfull_m, 
                     int *npart2_p, int *npart2_m, 
                     int *npart3_p, int *npart3_m,
                     int *npartn_p, int *npartn_m,
                     int *nunk_p, int *nunk_m, int * nunk_u,
                     int *nsfull_p, int *nsfull_m, 
                     int *nspart2_p, int *nspart2_m, 
                     int *nspart3_p, int *nspart3_m,
                     int *nspartn_p, int *nspartn_m)
/*end*/

/*-Parameters:
int mindx;        Index as returned by unq_reflist_init (R)
int iref;         Reflection number (from 1 upwards) (R)
int *nfull_p;     Count: fulls +h+k+l (non-spatially overlapped) (W)
int *nfull_m;     Count: fulls -h-k-l (non-spatially overlapped) (W)
int *npart2_p;    Count: 2 image partials +h+k+l 
                  (non-spatially overlapped) (W)
int *npart2_m;    Count: 2 image partials -h-k-l 
                  (non-spatially overlapped) (W)
int *npart3_p;    Count: 3 image partials +h+k+l 
                  (non-spatially overlapped) (W)
int *npart3_m;    Count: 3 image partials -h-k-l 
                  (non-spatially overlapped) (W)
int *npartn_p;    Count: >3 but OK image partials +h+k+l 
                  (non-spatially overlapped) (W)
int *npartn_m;    Count: >3 but OK image partials -h-k-l 
                  (non-spatially overlapped) (W)
int *nunk_p;      Count: unknown (e.g. merged) +h+k+l
int *nunk_m;      Count: unknown (e.g. merged) -h-k-l
int *nunk_u;      Count: unknown (e.g. merged) and unknown sign
int *nsfull_p;    Count: fulls +h+k+l (spatially overlapped) (W)
int *nsfull_m;    Count: fulls -h-k-l (spatially overlapped) (W)
int *nspart2_p;   Count: 2 image partials +h+k+l 
                  (spatially overlapped) (W)
int *nspart2_m;   Count: 2 image partials -h-k-l 
                  (spatially overlapped) (W)
int *nspart3_p;   Count: 3 image partials +h+k+l 
                  (spatially overlapped) (W)
int *nspart3_m;   Count: 3 image partials -h-k-l 
                  (spatially overlapped) (W)
int *nspartn_p;   Count: >3 but OK image partials +h+k+l 
                  (spatially overlapped) (W)
int *nspartn_m;   Count: >3 but OK image partials -h-k-l 
                  (spatially overlapped) (W)

Return:  =  0, OK
         = -1, Invalid index given, list not initialised or reflection
               record not present.
-end*/
{
   UNQreflection *refl;

   refl = (UNQreflection *) dml_recpointer (mindx, iref);
   if (refl==0) return -1;
   *nfull_p = refl->n_full_p;
   *nfull_m = refl->n_full_m;
   *npart2_p = refl->n_part2_p;
   *npart2_m = refl->n_part2_m;
   *npart3_p = refl->n_part3_p;
   *npart3_m = refl->n_part3_m;
   *npartn_p = refl->n_partn_p;
   *npartn_m = refl->n_partn_m;
   *nunk_p = refl->n_unk_p;
   *nunk_m = refl->n_unk_m;
   *nunk_u = refl->n_unk_u;
   *nsfull_p = refl->n_full_sp_p;
   *nsfull_m = refl->n_full_sp_m;
   *nspart2_p = refl->n_part2_sp_p;
   *nspart2_m = refl->n_part2_sp_m;
   *nspart3_p = refl->n_part3_sp_p;
   *nspart3_m = refl->n_part3_sp_m;
   *nspartn_p = refl->n_partn_sp_p;
   *nspartn_m = refl->n_partn_sp_m;
    return 0;
}
/* Fortran binding: unqf_reflist_get */

void unqf_reflist_get (int *mindx, int * iref, 
                       int *nfull_p, int *nfull_m, 
                       int *npart2_p, int *npart2_m, 
                       int *npart3_p, int *npart3_m,
                       int *npartn_p, int *npartn_m,
                       int *nunk_p, int *nunk_m, int *nunk_u,
                       int *nsfull_p, int *nsfull_m, 
                       int *nspart2_p, int *nspart2_m, 
                       int *nspart3_p, int *nspart3_m,
                       int *nspartn_p, int *nspartn_m,
                       int *ierr)
{
   *ierr = unq_reflist_get (*mindx, *iref, nfull_p, nfull_m, 
                            npart2_p, npart2_m, npart3_p, npart3_m,
                            npartn_p, npartn_m, nunk_p, nunk_m,
                            nunk_u, nsfull_p, nsfull_m, 
                            nspart2_p, nspart2_m, nspart3_p, nspart3_m,
                            nspartn_p, nspartn_m);
   return;
}

/*-Section: Clear a Unique Reflections List
Two routines are available for clearing a unique reflections list. In one
the data records are removed but the list remains initialised and in the
other the list is removed completely.
-end*/

/*-Routine: Clear a unique reflections list - unq_reflist_clear
The routine unq_reflist_clear (unqf_reflist_clear) clears a unique reflections
list and frees unused space. The index returned from the 
unq_reflist_init (unqf_reflist_init) call can still be used and new 
reflections may be added etc. as required. (cf unq_reflist_free
(unqf_reflist_free).
-end*/

/*                                            
******************************
**    unq_reflist_clear     **
******************************

Purpose: Clear a unique reflections list

Author:  John W. Campbell, July 1995

CD-Fortran:
         CALL UNQF_REFLIST_CLEAR  (MINDX, IERR)
CD-end:
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by UNQF_REFLIST_INIT
IERR         i  (W)  Error return code from unq_reflist_free
-end*/

/*-C:*/
int unq_reflist_clear (int mindx)
/*end*/

/*-Parameters:
int mindx;       Index as returned by unq_reflist_init (R)

Return:  =  0, OK
         = -1, Invalid index given or list not initialised
-end*/
{
   int i;

   i = dml_clear (mindx);
   if (i<0) return i;
   return dml_freespace(mindx);
}

/* Fortran binding: unqf_reflist_clear */

void unqf_reflist_clear (int *mindx, int *ierr)
{
   *ierr = unq_reflist_clear (*mindx);
   return;
}

/*-Routine: Clear and free a unique reflections list - unq_reflist_free
The routine unq_reflist_free (unqf_reflist_free) clears a unique reflections 
list and makes it available for re-use via another initialisation call. All the
memory that was allocated for the list is freed.
-end*/

/*                                            
******************************
**    unq_reflist_free      **
******************************

Purpose: Clear a reflections list and free it for re-use

Author:  John W. Campbell, July 1995

CD-Fortran:
         CALL UNQF_REFLIST_FREE  (MINDX, IERR)
CD-end:
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by UNQF_REFLIST_INIT
IERR         i  (W)  Error return code from unq_reflist_free
-end*/

/*-C:*/
int unq_reflist_free (int mindx)
/*end*/

/*-Parameters:
int mindx;       Index as returned by unq_reflist_init (R)

Return:  =  0, OK
         = -1, Invalid index given or list not initialised
-end*/
{
   return dml_free (mindx);
}

/* Fortran binding: unqf_reflist_free */

void unqf_reflist_free (int *mindx, int *ierr)
{
   *ierr = unq_reflist_free (*mindx);
   return;
}

/*-Section: Sort Unique Reflections Lists
Routines are available to sort a unique reflections list by 'h k l' or
'l h k'.
-end*/

/*-Routine: Sort a unique reflections list on hkl - unq_reflist_sorthkl
The routine unq_reflist_sorthkl (unqf_reflist_sorthkl) sorts a unique 
reflections list on h, k, l (ascending). The order of any records, with the 
same sort key value, is preserved.
-end*/

/*                                            
******************************
**    unq_reflist_sorthkl   **
******************************

Purpose: Sort a unique reflections list on h k l.

Author:  John W. Campbell, July 1995

CD-Fortran:
         CALL UNQF_REFLIST_SORTHKL  (MINDX, IERR)
CD-end:
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by UNQF_REFLIST_INIT
IERR         i  (W)  Error return code from unq_reflist_sorthkl
-end*/

/*-C:*/
int unq_reflist_sorthkl (int mindx)
/*end*/

/*-Parameters:
int mindx;       Index as returned by unq_reflist_init (R)

Return:  =  0, OK
         = -1, Invalid index given or list not initialised
-end*/
{
   return dml_sort (mindx, unq_reflist_sor_hkl);
}

/* Fortran binding: unqf_reflist_sorthkl */

void unqf_reflist_sorthkl (int *mindx, int *ierr)
{
   *ierr = unq_reflist_sorthkl (*mindx);
   return;
}

/*-Routine: Sort a unique reflections list on lhk - unq_reflist_sortlhk
The routine unq_reflist_sortlhk (unqf_reflist_sortlhk) sorts a unique 
reflections list on l, h, k (ascending). The order of any records, with the 
same sort key value, is preserved.
-end*/

/*                                            
******************************
**    unq_reflist_sortlhk   **
******************************

Purpose: Sort a unique reflections list on l h k

Author:  John W. Campbell, July 1995

CD-Fortran:
         CALL UNQF_REFLIST_SORTLHK  (MINDX, IERR)
CD-end:
*/
/*-Parameters:
MINDX        i  (R)  Index as returned by UNQF_REFLIST_INIT
IERR         i  (W)  Error return code from unq_reflist_sorthkl
-end*/

/*-C:*/
int unq_reflist_sortlhk (int mindx)
/*end*/

/*-Parameters:
int mindx;       Index as returned by unq_reflist_init (R)

Return:  =  0, OK
         = -1, Invalid index given or list not initialised
-end*/
{
   return dml_sort (mindx, unq_reflist_sor_lhk);
}

/* Fortran binding: unqf_reflist_sortlhk */

void unqf_reflist_sortlhk (int *mindx, int *ierr)
{
   *ierr = unq_reflist_sortlhk (*mindx);
   return;
}


/****************************************************************************/
/*               S E R V I C E       R O U T I N E S                        */
/****************************************************************************/


/*                                            
******************************
**    unq_reflist_sor_hkl   **
******************************

Purpose: 'User supplied'  sort comparison function for 'hkl' based sort. 

Return:  (For as needed by dml_sort)
         
Author:  John W. Campbell, July 1995

**** SERVICE ROUTINE ONLY ****

*/

int unq_reflist_sor_hkl (void * rec1, void * rec2, int i1, int i2)
{
   UNQreflection *refl1, *refl2;

   refl1 = (UNQreflection *) rec1;
   refl2 = (UNQreflection *) rec2;
   if (refl1->h>refl2->h) return 1;
   if (refl1->h<refl2->h) return -1;
   if (refl1->k>refl2->k) return 1;
   if (refl1->k<refl2->k) return -1;
   if (refl1->l>refl2->l) return 1;
   if (refl1->l<refl2->l) return -1;
   if (i1>i2) return 1;
   if (i1<i2) return -1;
   return 0;
}

/*                                            
******************************
**    unq_reflist_sor_lhk   **
******************************

Purpose: 'User supplied'  sort comparison function for 'lhk' based sort. 

Return:  (For as needed by dml_sort)
         
Author:  John W. Campbell, July 1995

**** SERVICE ROUTINE ONLY ****

*/

int unq_reflist_sor_lhk (void * rec1, void * rec2, int i1, int i2)
{
   UNQreflection *refl1, *refl2;

   refl1 = (UNQreflection *) rec1;
   refl2 = (UNQreflection *) rec2;
   if (refl1->l>refl2->l) return 1;
   if (refl1->l<refl2->l) return -1;
   if (refl1->h>refl2->h) return 1;
   if (refl1->h<refl2->h) return -1;
   if (refl1->k>refl2->k) return 1;
   if (refl1->k<refl2->k) return -1;
   if (i1>i2) return 1;
   if (i1<i2) return -1;
   return 0;
}

